/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include "bucket.h"

#include <float.h>
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

#include "err.h"
#include "math_common.h"

#define Q1 0xff
#define B1 8
#define P 4
#define PS 256

DOUBLE MAXD(DOUBLE,DOUBLE);
DOUBLE MIND(DOUBLE,DOUBLE);
#if USE_MPI
INT g_buck_len[DIMS0];
INT g_max_x;
INT g_max_xy;
#endif

void radixsort( key_pid* tab, int size, key_pid* tmp)
{
    int cs[P][PS];
    int i,j;
    int mv[] = {0,8,16,24};
    memset(cs,0,sizeof(cs));
    for ( i = 0; i < size; ++i )
    {
        int key = tab[i].key;
        for ( j = 0; j < P; ++j )
        {
            cs[j][ key&Q1 ]++;
            key >>= B1;
        }
    }
    for ( j = 0; j < P; ++j )
    {
        int prev = 0;
        for ( i = 0; i < PS; ++i )
        {
            int c = cs[j][i];
            cs[j][i] = prev;
            prev += c;
        }
    }
    for ( j = 0; j < P; ++j )
    {
        key_pid* src = (j&1) ? tmp : tab;
        key_pid* dst = (j&1) ? tab : tmp;
        for ( i = 0; i < size; ++i )
        {
            int key = (src[i].key>>mv[j])&Q1;
            dst[ cs[j][key] ] = src[ i ];
            cs[j][key]++;
        }
    }
}

INT in_phan_reg_down( DOUBLE x, INT dim, DOUBLE egde_len )
{
    switch(dim)
    {
        case 0: return x <= -xbox/2 + egde_len + 0.01f;
        case 1: return x <= -ybox/2 + egde_len + 0.01f;
        case 2: return x <= -zbox/2 + egde_len + 0.01f;
        default: UNERR("Wrong dim"); return 0;
    }
}

INT in_phan_reg_up( DOUBLE x, INT dim, DOUBLE egde_len )
{
    switch(dim)
    {
        case 0: return x >= xbox/2 - egde_len - 0.01f;
        case 1: return x >= ybox/2 - egde_len - 0.01f;
        case 2: return x >= zbox/2 - egde_len - 0.01f;
        default: UNERR("Wrong dim"); return 0;
    }
}

VOID correct_xyz( DOUBLE* v )
{
    v[0] += v[0] < -xbox/2 ? xbox : 0;
    v[1] += v[1] < -ybox/2 ? ybox : 0;
    v[2] += v[2] < -zbox/2 ? zbox : 0;
}

INT get_h_l( INT size, const PTR_DOUBLE coords, DOUBLE h[3], DOUBLE l[3], INT LD, PTR_DOUBLE blen )
{
    INT i, j;
    for( j=0; j<3; ++j ) h[j] = -FLT_MAX;
    for( j=0; j<3; ++j ) l[j] = FLT_MAX;
    for( i = 0; i < size; ++i )
    {
        DOUBLE v[3]; 
        v[0] = coords[i*LD  ];
        v[1] = coords[i*LD+1];
        v[2] = coords[i*LD+2];
        trans_to_box( v );
#if USE_FLOAT
        correct_xyz(v);
#endif
        for ( j=0; j<3; ++j )
        {
            if( v[j] > h[j] )
            {
                h[j]=v[j];
            }
            if( v[j] < l[j] )
            {
                l[j]=v[j];
            }
        }
    }
    if ( bc == BC_BOX )
    {
        if ( l[0] < -xbox/2 || l[1] < -ybox/2 || l[2] < -zbox/2 ) WARNV(("Under box %g %g %g",l[0],l[1],l[2]))
        if ( h[0] >  xbox/2 || h[1] >  ybox/2 || h[2] >  zbox/2 ) WARNV(("Above box %g %g %g",h[0],h[1],h[2]))
        h[0] = MAXD( blen[0] + xbox/2, h[0] );
        h[1] = MAXD( blen[1] + ybox/2, h[1] );
        h[2] = MAXD( blen[2] + zbox/2, h[2] );
        l[0] = MIND( xbox/2 - 3*blen[0], l[0]);
        l[1] = MIND( ybox/2 - 3*blen[1], l[1]);
        l[2] = MIND( zbox/2 - 3*blen[2], l[2]);
    }
    blen[0] = MAXD( (h[0]-l[0])/BUCKET_SPATIAL_M, blen[0]);
    blen[1] = MAXD( (h[1]-l[1])/BUCKET_SPATIAL_M, blen[1]);
    blen[2] = MAXD( (h[2]-l[2])/BUCKET_SPATIAL_M, blen[2]);
    if ( bc == BC_BOX )
    {
        if ( xbox / blen[0] <= 3.01 ||
             ybox / blen[1] <= 3.01 ||
             zbox / blen[2] <= 3.01 ||  
             l[0] < -xbox/2 || l[1] < -ybox/2 || l[2] < -zbox/2 )
        {
            return 0;
        }
    }
    return 1;
}

INT correct_indx( int indx, int last )
{
    if ( indx < 0 )
    {
        if ( indx == - 1 )
        {
            WARN("Bead on edge detected beg")
            return last - 1;
        }
        UNERR("Wrong indx, too small");
    }
    if ( indx >= last )
    {
        if ( indx == last )
        {
            WARN("Bead on edge detected end")
            return 0;
        }
        UNERR("Wrong indx, too large");
    }
    /*if( indx < 0 ) return last + indx;
    return (indx)%last;*/
    return indx;
}

int cmp_keys( const void* a, const void* b )
{
    const key_pid* aa = (const key_pid*)a;
    const key_pid* bb = (const key_pid*)b;
    if( aa->key - bb->key == 0 )
        return aa->pid - bb->pid;
    else
        return aa->key - bb->key;
}

INT can_interact( INT pidn, INT pidm )
{
    INT m1 = pidn & MVP_MASK;
    INT m2 = pidm;
    INT id1 = pidn >> MVP_BITS;
    INT id2 = pidm >> MVP_BITS;
    if ( id1 == id2 ) return 0;
    return !(m1&m2);
}

INT make_keyb(INT x,INT y,INT z)
{
    INT cx = correct_indx( x, BUCKET_SPATIAL_M + 1 );
    INT cy = correct_indx( y, BUCKET_SPATIAL_M + 1 );
    INT cz = correct_indx( z, BUCKET_SPATIAL_M + 1 );
    INT key = ( cx << BUCKET_SPATIAL_X) |
              ( cy << BUCKET_SPATIAL_Y) |
              ( cz << BUCKET_SPATIAL_Z);
    return key;
}

#ifdef USE_MPI
INT g_addall = 0;
UINT gen_hash_key( INT key )
{
    if ( !g_max_x )
    {
        return (key >> BUCKET_SPATIAL_Z) ^ (key >> BUCKET_SPATIAL_Y) ^ key;
    }
    else
    {
        INT ix = key & BUCKET_SPATIAL_MASK;
        INT iy = (key >> BUCKET_SPATIAL_Y) & BUCKET_SPATIAL_MASK;
        INT iz = (key >> BUCKET_SPATIAL_Z) & BUCKET_SPATIAL_MASK;
        ix /= g_buck_len[0];
        iy /= g_buck_len[1];
        iz /= g_buck_len[2];
        return ((ix%g_max_x) + iy * g_max_x)%g_max_xy + iz * g_max_xy;
    }
}
INT can_add( INT key )
{
    key = gen_hash_key(key);
    return g_addall || ((key%g_numprocs) == g_id);
}
#endif

void add_key( key_pid* keys, INT* key_size, INT key, INT pid )
{
#ifdef USE_MPI
    if( can_add( key ) )
    {
#endif
    keys[ *key_size ].key = key;
    keys[ *key_size ].pid = pid;
    (*key_size)++;
#ifdef USE_MPI
    }
#endif
}

void gen_keys( bucket_s* br, const PTR_DOUBLE _coords, INT id, INT (*ignore)(const DOUBLE*) )
{
    key_pid* keys = br->keys;
    INT* key_size = &br->size_keys;
    const PTR_DOUBLE blen = br->blen;
    const PTR_DOUBLE l = br->l;
    INT x,y,z,k;
    DOUBLE v[DIMS0];
    if ( ignore && (*ignore)(_coords ) ) return;

    v[0] = _coords[0];
    v[1] = _coords[1];
    v[2] = _coords[2];

    x = (INT)FLOORD((v[0]-l[0])/blen[0]);
    y = (INT)FLOORD((v[1]-l[1])/blen[1]);
    z = (INT)FLOORD((v[2]-l[2])/blen[2]);

    for ( k = 0; k < 8; ++k )
    {
        INT nkey = make_keyb( x+((k&MVP_X)>>MVP_Xb),
                              y+((k&MVP_Y)>>MVP_Yb),
                              z+((k&MVP_Z)>>MVP_Zb) );
        add_key( keys, key_size, nkey, (id << MVP_BITS) | k );
    }
}

void gen_keys_p( bucket_s* br, const PTR_DOUBLE _coords, INT id, INT (*ignore)(const DOUBLE*) )
{
    key_pid* keys = br->keys;
    INT* key_size = &br->size_keys;
    const PTR_DOUBLE blen = br->blen;
    const PTR_DOUBLE l = br->l;
    INT k, n, m;
    INT x,y,z;
    DOUBLE vv[DIMS0];
    vv[0] = _coords[0];
    vv[1] = _coords[1];
    vv[2] = _coords[2];
    trans_to_box(vv);
#if USE_FLOAT
    correct_xyz(vv);
#endif
    if ( ignore && (*ignore)(_coords) ) return;

    x = (INT)FLOORD((vv[0]-l[0])/blen[0]);
    y = (INT)FLOORD((vv[1]-l[1])/blen[1]);
    z = (INT)FLOORD((vv[2]-l[2])/blen[2]);

    for ( k = 0; k < 8; ++k )
    {
        INT nkey = make_keyb( x+((k&MVP_X)>>MVP_Xb),
                              y+((k&MVP_Y)>>MVP_Yb),
                              z+((k&MVP_Z)>>MVP_Zb) );
        add_key( keys, key_size, nkey, (id << MVP_BITS) | k );
    }

    for ( k = 0; k <= in_phan_reg_down(vv[0],0,blen[0]); ++k )
        for ( n = 0; n <= in_phan_reg_down(vv[1],1,blen[1]); ++n )
            for ( m = 0; m <= in_phan_reg_down(vv[2],2,blen[2]); ++m )
            {
                if( k | n | m )
                {
                    INT ix,iy,iz;
                    INT a,b,c;
                    INT d,e,f;
                    DOUBLE dx,dy,dz;
                    dx = vv[0] + k * xbox;
                    dy = vv[1] + n * ybox;
                    dz = vv[2] + m * zbox;

                    ix = (INT)FLOORD((dx-l[0])/blen[0]);
                    iy = (INT)FLOORD((dy-l[1])/blen[1]);
                    iz = (INT)FLOORD((dz-l[2])/blen[2]);

                    for ( a = 0; a <= 1-k; ++a )
                        for ( b = 0; b <= 1-n; ++b )
                            for ( c = 0; c <= 1-m; ++c )
                            {
                                INT nkey = make_keyb( ix+a, iy+b, iz+c );
                                add_key( keys, key_size, nkey,
                                         (id << MVP_BITS) | (a << MVP_Xb ) | (b << MVP_Yb ) | (c << MVP_Zb ) | MVP_PH );
                            }
                    for ( d = 0; d <= in_phan_reg_up(vv[0],0,blen[0]); ++d )
                        for ( e = 0; e <= in_phan_reg_up(vv[1],1,blen[1]); ++e )
                            for ( f = 0; f <= in_phan_reg_up(vv[2],2,blen[2]); ++f )
                            {
                                if( (d | e | f) && (( (!k & !d) + 3*(!n & !e) + 9*(!m & !f) + d*2 + e*6 + f*18 ) <= 13) )
                                {
                                    INT cx, cy, cz;
                                    cx = (INT)FLOORD((dx-d*xbox-l[0])/blen[0]);
                                    cy = (INT)FLOORD((dy-e*ybox-l[1])/blen[1]);
                                    cz = (INT)FLOORD((dz-f*zbox-l[2])/blen[2]);
                                    if ( cx>=-1 && cy >=-1 && cz>=-1)
                                    {
                                        cx = d ? 0 : ix;
                                        cy = e ? 0 : iy;
                                        cz = f ? 0 : iz;

                                        for ( a = 0; a <= (1-k)*(1-d); ++a )
                                            for ( b = 0; b <= (1-n)*(1-e); ++b )
                                                for ( c = 0; c <= (1-m)*(1-f); ++c )
                                                {
                                                    INT nkey = make_keyb(cx+a,cy+b,cz+c);
                                                    add_key( keys, key_size, nkey,
                                                             (id << MVP_BITS) | (a << MVP_Xb ) | (b << MVP_Yb ) | (c << MVP_Zb ) | MVP_PH );
                                                }
                                    }
                                }
                            }
                }
            }
}

INT init_keys( bucket_s* ret, INT size, const PTR_DOUBLE _coords, DOUBLE blen, INT LD, INT (*ignore)(const DOUBLE*) )
{
    DOUBLE h[3];
    INT i;
    DOUBLE* l = ret->l;
    CHMEM(ret);
    ret->blen[0] = ret->blen[1] = ret->blen[2] = blen;
    if ( ! get_h_l( size, _coords, h, l, LD, ret->blen ) ) return 0;
    ret->size_keys = 0;
    if ( bc != BC_BOX )
    {
        for ( i = 0; i < size; ++i )
        {
            gen_keys( ret, _coords + i*LD, i, ignore );
        }
    }
    else
    {
        for ( i = 0; i < size; ++i )
        {
            gen_keys_p( ret, _coords + i*LD, i, ignore );
        }
    }

    return init_keys_end(ret);
}

INT init_keys_end( bucket_s* ret )
{
    INT i, prev;
    if ( ret->size_keys >= ret->max_beads )
    {
        WARN("Too many phantoms, bug?");
        return 0;
    }
    if ( ret->size_keys < 2048 )
    {
        qsort( ret->keys, ret->size_keys, sizeof(key_pid), cmp_keys );
    }
    else
    {
        radixsort( ret->keys, ret->size_keys, ret->tmp );
    }
    prev = ret->keys[0].key;
    ret->size_ids = 1;
    ret->max_group = 1;
    ret->ids[0] = 0;
    for( i = 1; i < ret->size_keys; ++i )
    {
        if( prev != ret->keys[i].key )
        {
            ret->ids[2*ret->size_ids] = i;
            ret->ids[2*(ret->size_ids-1)+1] = i;
            if ( (ret->max_group < i - ret->ids[2*(ret->size_ids-1)]) )
                ret->max_group = i - ret->ids[2*(ret->size_ids-1)];
            ret->size_ids++;
            prev = ret->keys[i].key;
        }
    }
    ret->ids[2*(ret->size_ids-1)+1] = ret->size_keys;
    if ( ret->max_group < ret->size_keys - ret->ids[2*(ret->size_ids-1)] )
        ret->max_group = ret->size_keys - ret->ids[2*(ret->size_ids-1)];
    /*if ( ret->max_group >= size )
    {
        return NULL; //better to use optimized brute force method
    }*/
    return 1;
}

void init_bucket( bucket_s** pbr, INT size, INT LD_COORD, INT LD_F )
{
    INT i;
    bucket_s* br = (bucket_s*) malloc( sizeof(bucket_s) ); CHMEM(br);
    br->ids = (INT*) malloc( sizeof(INT) * size* (MAX_PHANTOMS+8) * 2 ); CHMEM(br->ids);
    br->keys = (key_pid*) malloc( sizeof(key_pid)*(size*8 + size*MAX_PHANTOMS+1) ); CHMEM(br->keys);
    br->tmp = (key_pid*) malloc( sizeof(key_pid)*(size*8 + size*MAX_PHANTOMS+1) ); CHMEM(br->tmp);
    br->max_beads = size*8 + size*MAX_PHANTOMS;
    br->capacity = (INT*) malloc( sizeof(INT)*g_threads ); CHMEM(br->capacity);
    for ( i = 0; i < g_threads; ++i ) br->capacity[i] = 16;
    br->coords = (DOUBLE**) malloc( sizeof(DOUBLE*) * g_threads); CHMEM(br->coords);
    for ( i = 0; i < g_threads; ++i )
    {
        br->coords[i] = (DOUBLE*) malloc( sizeof(DOUBLE) * LD_COORD * br->capacity[i] );
        CHMEM(br->coords[i]);
    }
    if ( LD_F != 0 )
    {
        br->F = (DOUBLE**) malloc( sizeof(DOUBLE*) * g_threads ); CHMEM(br->coords);
        for ( i = 0; i < g_threads; ++i )
        {
            br->F[i] = (DOUBLE*) malloc( sizeof(DOUBLE) * LD_F * br->capacity[i] );
            CHMEM(br->F[i]);
        }   
    }
    else
    {
        br->F = NULL;
    }
    *pbr = br;
}

DOUBLE* get_tmp_coords( bucket_s* bs, INT tid, INT LD_COORD, INT LD_F )
{
    if ( bs->capacity[tid] < bs->max_group )
    {
        bs->capacity[tid] = bs->max_group;
        bs->coords[tid] = (DOUBLE*) realloc( bs->coords[tid], bs->max_group*sizeof(DOUBLE)*LD_COORD );
        CHMEM( bs->coords[tid] );
        if ( bs->F )
        {
            bs->F[tid] = (DOUBLE*) realloc( bs->F[tid], bs->max_group*DIMS0*sizeof(DOUBLE)*LD_F );
            CHMEM( bs->F[tid] );
        }
    }
    return bs->coords[tid];
}

DOUBLE* get_tmp_forces( bucket_s* bs, INT tid, INT LD_COORD, INT LD_F )
{
    if ( bs->capacity[tid] < bs->max_group )
    {
        bs->capacity[tid] = bs->max_group;
        bs->coords[tid] = (DOUBLE*) realloc( bs->coords[tid], bs->max_group*sizeof(DOUBLE)*LD_COORD );
        CHMEM( bs->coords[tid] );
        if ( bs->F )
        {
            bs->F[tid] = (DOUBLE*) realloc( bs->F[tid], bs->max_group*DIMS0*sizeof(DOUBLE)*LD_F );
            CHMEM( bs->F[tid] );
        }
    }
    return bs->F[tid];
}

void free_bucket( bucket_s* br )
{
    INT i;
    if ( br )
    {
        free( br->ids );
        free( br->keys );
        for ( i = 0; i < g_threads; ++i )
        {
            free( br->coords[i] );
            if ( br->F ) free( br->F[i] );
        }
        if ( br->F ) free( br->F );
        free( br->coords );
        free( br->capacity );
        free( br->tmp );
        free( br );
    }
}
