/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2012: University of Warsaw                             *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@cent.uw.edu.pl>        *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include "verlet.cuh"

#include <cuda_runtime_api.h>
#include <float.h>
#include <stdio.h>

#include "sorter.cuh"

#include "../bucket.h"
#include "../cuda_common.h"
#include "../err.h"
#include "../math_common.h"

__constant__ DOUBLE dev_box0;
__constant__ DOUBLE dev_box1;
__constant__ DOUBLE dev_box2;
__constant__ DOUBLE dev_invL0;
__constant__ DOUBLE dev_invL1;
__constant__ DOUBLE dev_invL2;

__device__ UINT d_keys_size;
__constant__ DOUBLE dev_offset_u[3];
__constant__ DOUBLE dev_blen_u[3];

//#define EPS 0.001f
#define MAX_MULT 28
#define PATTERN_BITS 4
//#define PATTERN_MASK 15
#define PATTERN_PH 8
#define PH_COUNT 8

__device__ void red_single( INT tid, INT curr_size, DOUBLE* sh )
{
    sh[tid             ] = max( sh[tid             ], sh[tid+curr_size             ] );
    sh[tid+  blockDim.x] = max( sh[tid+  blockDim.x], sh[tid+curr_size+  blockDim.x] );
    sh[tid+2*blockDim.x] = max( sh[tid+2*blockDim.x], sh[tid+curr_size+2*blockDim.x] );
    sh[tid+3*blockDim.x] = min( sh[tid+3*blockDim.x], sh[tid+curr_size+3*blockDim.x] );
    sh[tid+4*blockDim.x] = min( sh[tid+4*blockDim.x], sh[tid+curr_size+4*blockDim.x] );
    sh[tid+5*blockDim.x] = min( sh[tid+5*blockDim.x], sh[tid+curr_size+5*blockDim.x] );
}

__device__ void red_single_vol( INT tid, volatile INT curr_size, volatile DOUBLE* sh )
{
    sh[tid             ] = max( sh[tid             ], sh[tid+curr_size             ] );
    sh[tid+  blockDim.x] = max( sh[tid+  blockDim.x], sh[tid+curr_size+  blockDim.x] );
    sh[tid+2*blockDim.x] = max( sh[tid+2*blockDim.x], sh[tid+curr_size+2*blockDim.x] );
    sh[tid+3*blockDim.x] = min( sh[tid+3*blockDim.x], sh[tid+curr_size+3*blockDim.x] );
    sh[tid+4*blockDim.x] = min( sh[tid+4*blockDim.x], sh[tid+curr_size+4*blockDim.x] );
    sh[tid+5*blockDim.x] = min( sh[tid+5*blockDim.x], sh[tid+curr_size+5*blockDim.x] );
}

template<INT cuda_block, INT curr_size>
__device__ void red6( int tid )
{
    extern __shared__ DOUBLE sh[];
    if ( cuda_block >= 2*curr_size )
    {
        if ( tid < curr_size )
        {
            red_single( tid, curr_size, sh );
        }
    }
}

template<INT cuda_block>
__global__ void comp_h_l( DOUBLE* coords, DOUBLE* fout, INT LD, INT u_size )
{
    int tid = threadIdx.x;
    int id = blockDim.x*blockIdx.x + tid;
    extern __shared__ DOUBLE sh[];
    DOUBLE hx, hy, hz, lx, ly, lz;
    hx = hy = hz = -FLT_MAX;
    lx = ly = lz = FLT_MAX;
    while ( id < u_size )
    {
        DOUBLE x, y, z;
        DOUBLE* cds = coords + id*LD;
        x = cds[0];
        y = cds[1];
        z = cds[2];
        hx = max(x,hx); hy = max(y,hy); hz = max(z,hz);
        lx = min(x,lx); ly = min(y,ly); lz = min(z,lz);
        id += blockDim.x*gridDim.x;
    }

    sh[tid             ] = hx;
    sh[tid+  blockDim.x] = hy;
    sh[tid+2*blockDim.x] = hz;
    sh[tid+3*blockDim.x] = lx;
    sh[tid+4*blockDim.x] = ly;
    sh[tid+5*blockDim.x] = lz;

    __syncthreads();
    red6<cuda_block,512>( tid ); __syncthreads();
    red6<cuda_block,256>( tid ); __syncthreads();
    red6<cuda_block,128>( tid ); __syncthreads();
    red6<cuda_block, 64>( tid ); __syncthreads();

    if ( cuda_block >= 64 )
    {
        if( tid < 32 )
        {
            red_single_vol( tid, 32, sh );
            red_single_vol( tid, 16, sh );
            red_single_vol( tid,  8, sh );
            red_single_vol( tid,  4, sh );
            red_single_vol( tid,  2, sh );
            red_single_vol( tid,  1, sh );
        }
    }
    else if( cuda_block == 32 )
    {
        if( tid < 16 )
        {
            red_single_vol( tid, 16, sh );
            red_single_vol( tid,  8, sh );
            red_single_vol( tid,  4, sh );
            red_single_vol( tid,  2, sh );
            red_single_vol( tid,  1, sh );
        }
    }// no else
    if ( tid == 0 )
    {
#pragma unroll 6
        for ( int i = 0; i < 6; ++i )
            fout[6*blockIdx.x+i]=sh[i*blockDim.x];
    }
}

__global__ void trans_to_box_all( DOUBLE* coord, INT LD, INT u_size  )
{
    int id = blockDim.x*blockIdx.x + threadIdx.x;
    if ( id < u_size )
    {
        DOUBLE* bc = coord + LD*id;
        bc[0] -= dev_box0 * FLOORD( bc[0] * dev_invL0 + 0.5f );
        bc[1] -= dev_box1 * FLOORD( bc[1] * dev_invL1 + 0.5f );
        bc[2] -= dev_box2 * FLOORD( bc[2] * dev_invL2 + 0.5f );
#if USE_FLOAT
        bc[0] += bc[0] < -dev_box0/2 ? dev_box0 : 0;
        bc[1] += bc[1] < -dev_box1/2 ? dev_box1 : 0;
        bc[2] += bc[2] < -dev_box2/2 ? dev_box2 : 0;
#endif
    }
}

template<int last>
__device__ INT correct_indx( INT indx )
{
    return (indx)%last;
}

__device__ INT make_key( INT x, INT y, INT z )
{
    INT cx = correct_indx<BUCKET_SPATIAL_M + 1>( x );
    INT cy = correct_indx<BUCKET_SPATIAL_M + 1>( y );
    INT cz = correct_indx<BUCKET_SPATIAL_M + 1>( z );
    INT key = ( cx << BUCKET_SPATIAL_X) |
              ( cy << BUCKET_SPATIAL_Y) |
              ( cz << BUCKET_SPATIAL_Z);
    return key;
}

template<int i>
__device__ void add_key( INT* kkeys, INT* kvalues, INT id, INT kx, INT ky, INT kz )
{
    INT key = make_key( kx+(i&1), ky+((i>>1)&1), kz+(i>>2) );
    kkeys[ PH_COUNT*id + i ] = key;
    kvalues[ PH_COUNT*id + i ] = (id<<PATTERN_BITS) + i;
}

__global__ void compute_keys_value( DOUBLE* coords, INT* kkeys, INT* kvalues, INT LD, INT u_size )
{
    INT id = blockDim.x*blockIdx.x + threadIdx.x;
    if ( id < u_size )
    {
        int kx, ky, kz;
        kx = (INT)FLOORD( (coords[id*LD  ] + dev_offset_u[0]) / dev_blen_u[0] );
        ky = (INT)FLOORD( (coords[id*LD+1] + dev_offset_u[1]) / dev_blen_u[1] );
        kz = (INT)FLOORD( (coords[id*LD+2] + dev_offset_u[2]) / dev_blen_u[2] );
        add_key<0>( kkeys, kvalues, id, kx, ky, kz);
        add_key<1>( kkeys, kvalues, id, kx, ky, kz);
        add_key<2>( kkeys, kvalues, id, kx, ky, kz);
        add_key<3>( kkeys, kvalues, id, kx, ky, kz);
        add_key<4>( kkeys, kvalues, id, kx, ky, kz);
        add_key<5>( kkeys, kvalues, id, kx, ky, kz);
        add_key<6>( kkeys, kvalues, id, kx, ky, kz);
        add_key<7>( kkeys, kvalues, id, kx, ky, kz);
    }
}

__device__ INT in_phan_reg_down_gpu( DOUBLE x, INT dim, DOUBLE cutoff )
{
    switch(dim)
    {
        case 0: return x <= -dev_box0/2 + cutoff + 0.01f;
        case 1: return x <= -dev_box1/2 + cutoff + 0.01f;
        case 2: return x <= -dev_box2/2 + cutoff + 0.01f;
    }
    return 0;
}

__device__ void add_key_periodic_ph( INT* kkeys, INT* kvals, INT key, INT id, INT i, INT u_size )
{
    unsigned old = atomicInc( &d_keys_size, 1<<30 );
    kkeys[ PH_COUNT*u_size + old ] = key;
    kvals[ PH_COUNT*u_size + old ] = (id<<PATTERN_BITS) | i | PATTERN_PH;
}

__device__ INT in_phan_reg_up_gpu( DOUBLE x, INT dim, DOUBLE cutoff )
{
    switch(dim)
    {
        case 0: return x >= dev_box0/2 - cutoff - 0.01f;
        case 1: return x >= dev_box1/2 - cutoff - 0.01f;
        case 2: return x >= dev_box2/2 - cutoff - 0.01f;
    }
    return 0;
}

__global__ void compute_keys_value_periodic( DOUBLE* coords, INT* kkeys, INT* kvals, INT COORDLD, INT u_size, INT u_cutoff )
{
    INT id = blockDim.x*blockIdx.x + threadIdx.x;
    if ( id < u_size )
    {
        INT kx, ky, kz;
        INT k, n, m;
        DOUBLE xx = coords[id*COORDLD];
        DOUBLE yy = coords[id*COORDLD+1];
        DOUBLE zz = coords[id*COORDLD+2];
        kx = (INT)floor( (xx + dev_offset_u[0]) / dev_blen_u[0] );
        ky = (INT)floor( (yy + dev_offset_u[1]) / dev_blen_u[1] );
        kz = (INT)floor( (zz + dev_offset_u[2]) / dev_blen_u[2] );
        add_key<0>( kkeys, kvals, id, kx, ky, kz);
        add_key<1>( kkeys, kvals, id, kx, ky, kz);
        add_key<2>( kkeys, kvals, id, kx, ky, kz);
        add_key<3>( kkeys, kvals, id, kx, ky, kz);
        add_key<4>( kkeys, kvals, id, kx, ky, kz);
        add_key<5>( kkeys, kvals, id, kx, ky, kz);
        add_key<6>( kkeys, kvals, id, kx, ky, kz);
        add_key<7>( kkeys, kvals, id, kx, ky, kz);
        for ( k = 0; k <= in_phan_reg_down_gpu(xx,0,u_cutoff); ++k )
                for ( n = 0; n <= in_phan_reg_down_gpu(yy,1,u_cutoff); ++n )
                    for ( m = 0; m <= in_phan_reg_down_gpu(zz,2,u_cutoff); ++m )
                    {
                        if( k | n | m )
                        {
                            INT ix,iy,iz;
                            INT a,b,c;
                            INT d,e,f;
                            DOUBLE dx,dy,dz;
                            dx = xx + k * dev_box0;
                            dy = yy + n * dev_box1;
                            dz = zz + m * dev_box2;

                            ix = (INT)FLOORD((dx+dev_offset_u[0])/dev_blen_u[0]);
                            iy = (INT)FLOORD((dy+dev_offset_u[1])/dev_blen_u[1]);
                            iz = (INT)FLOORD((dz+dev_offset_u[2])/dev_blen_u[2]);

                            for ( a = 0; a <= 1-k; ++a )
                                for ( b = 0; b <= 1-n; ++b )
                                    for ( c = 0; c <= 1-m; ++c )
                                    {
                                        INT nkey = make_key( ix+a, iy+b, iz+c );
                                        add_key_periodic_ph( kkeys, kvals, nkey, id, (a)|(b<<1)|(c<<2), u_size );
                                    }
                            for ( d = 0; d <= in_phan_reg_up_gpu(xx,0,u_cutoff); ++d )
                                for ( e = 0; e <= in_phan_reg_up_gpu(yy,0,u_cutoff); ++e )
                                    for ( f = 0; f <= in_phan_reg_up_gpu(zz,0,u_cutoff); ++f )
                                    {
                                        if( (d | e | f) && (( (!k & !d) + 3*(!n & !e) + 9*(!m & !f) + d*2 + e*6 + f*18 ) <= 13) )
                                        {
                                            INT cx, cy, cz;
                                            cx = (INT)FLOORD((dx-d*dev_box0+dev_offset_u[0])/dev_blen_u[0]);
                                            cy = (INT)FLOORD((dy-e*dev_box1+dev_offset_u[1])/dev_blen_u[1]);
                                            cz = (INT)FLOORD((dz-f*dev_box2+dev_offset_u[2])/dev_blen_u[2]);
                                            if ( cx>=-1 && cy >=-1 && cz>=-1)
                                            {
                                                cx = d ? 0 : ix;
                                                cy = e ? 0 : iy;
                                                cz = f ? 0 : iz;

                                                for ( a = 0; a <= (1-k)*(1-d); ++a )
                                                    for ( b = 0; b <= (1-n)*(1-e); ++b )
                                                        for ( c = 0; c <= (1-m)*(1-f); ++c )
                                                        {
                                                            INT nkey = make_key(cx+a,cy+b,cz+c);
                                                            add_key_periodic_ph( kkeys, kvals, nkey, id, (a)|(b<<1)|(c<<2), u_size );
                                                        }
                                            }
                                        }
                                    }
                        }
                    }
    }
}

__global__ void add_dummy_beads( INT* kkeys, INT* kvals, INT u_size )
{
    INT tid = threadIdx.x;
    INT last_indx = PH_COUNT * u_size + d_keys_size;
    kkeys[ last_indx + tid ]  = (BUCKET_SPATIAL_M + 1 + (tid & 0xF) ) << BUCKET_SPATIAL_Z;
    kkeys[ last_indx + tid ] |= (BUCKET_SPATIAL_M + 1 + ((tid & 0xF0)>>4) ) << BUCKET_SPATIAL_Y;
    kkeys[ last_indx + tid ] |= (BUCKET_SPATIAL_M + 1 + ((tid & 0xF00)>>8) ) << BUCKET_SPATIAL_Y;
    kvals[ last_indx + tid ] = PATTERN_PH;
}

#define MASK (~15)
#define COUNT 16

__global__ void get_indx_spatial( INT* kkeys, INT* indxs, INT u_size )
{
    INT tid = threadIdx.x;
    INT last_indx = PH_COUNT*u_size;
    last_indx += d_keys_size;
    if( (blockDim.x*blockIdx.x + tid)*COUNT < last_indx )
    {
        INT id_end = (blockDim.x*blockIdx.x + tid) * COUNT;/*first in warp*/
        INT last_warp_key = kkeys[ id_end + COUNT - 1 ];/*last in wrap*/
        while(1)
        {
            id_end += COUNT;
            if ( id_end < last_indx )
            {
                int ckey = kkeys[ id_end ];
                if ( ckey != last_warp_key ) break;
            }
            else
            {
                break;
            }
        }
        INT id_beg = (blockDim.x*blockIdx.x + tid) * COUNT;
        INT first_warp_key = kkeys[ id_beg ];//first in wrap
        id_beg += COUNT-1; //move to last in warp
        while(1)
        {
            id_beg -= COUNT;
            if ( id_beg >= 0 )
            {
                int ckey = kkeys[ id_beg ];
                if ( ckey != first_warp_key ) break;
            }
            else
            {
                break;
            }
        }
        INT id = blockDim.x*blockIdx.x + tid;
        indxs[ id*2   ] = (id_beg & MASK)+COUNT;// >=
        indxs[ id*2 +1] = id_end;// <
    }
    else
    {
        INT id = blockDim.x*blockIdx.x + tid;
        //no computation
        indxs[ id*2   ] = 0;// >=
        indxs[ id*2 +1] = 0;// <
    }
}

class Verlet_impl
{
protected:
    Verlet_impl( DOUBLE* coords, INT UDIM, INT u_size, DOUBLE u_cutoff, bool comp_indxs ) {
        coords_ = coords;
        UDIM_ = UDIM;
        u_size_ = u_size;
        u_cutoff_ = u_cutoff;
        comp_indxs_ = comp_indxs;
        brute_force_ = true;
        hl_ = (DOUBLE*) malloc( sizeof(DOUBLE) * sm_red * 6 ); CHMEM(hl_)
        u_sorter_ = NULL;
        kkeys_ = kvals_ = indxs_ = NULL;
    }
    DOUBLE* coords_;
    INT UDIM_;
    INT u_size_;
    DOUBLE u_cutoff_;
    bool comp_indxs_;
    bool brute_force_;
    DOUBLE* hl_;
    StaticSorter* u_sorter_;
    INT* kkeys_;
    INT* kvals_;
    INT* indxs_;
public:
    virtual INT buildlist( DOUBLE* d_out ) =0;
    virtual ~Verlet_impl() { free(hl_); }
    virtual void setLists( INT* kkeys, INT* kvals, INT* indxs ) =0;
    INT* getkkeys() { return kkeys_; }
    INT* getkvals() { return kvals_; }
    INT* getkindxs() { return indxs_; }
    bool useBruteForce() { return brute_force_; }
};

template <bool PBC>
class Verlet_impl_pbc: public Verlet_impl
{
protected:
    Verlet_impl_pbc( DOUBLE* coords, INT UDIM, INT u_size, DOUBLE u_cutoff, bool comp_indxs ):
        Verlet_impl( coords, UDIM, u_size, u_cutoff, comp_indxs ) {}
public:
    static Verlet_impl* createVerlet( DOUBLE* coords, INT UDIM, INT u_size, DOUBLE u_cutoff, bool comp_indxs );
    INT spatial_phase_one( DOUBLE* d_out );
    INT buildlist( DOUBLE* d_out )
    {
        if ( !u_sorter_ )
        {
            WARN("List not set")
            brute_force_ = true;
            return 0;
        }
        return spatial_phase_one( d_out );
    }
    void setLists( INT* kkeys, INT* kvals, INT* indxs )
    {
        kkeys_ = kkeys;
        kvals_ = kvals;
        indxs_ = indxs;
        INT mult = PBC ? MAX_MULT : PH_COUNT;
        u_sorter_ = new StaticSorter( kkeys_, kvals_, mult*u_size_ );
    }
};

template <bool PBC>
Verlet_impl* Verlet_impl_pbc<PBC>::createVerlet( DOUBLE* coords, INT UDIM, INT u_size, DOUBLE u_cutoff, bool comp_indxs )
{
    if ( PBC )
    {
        cudaMemcpyToSymbol( dev_box0, &xbox, sizeof(DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_box1, &ybox, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_box2, &zbox, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL0, &inv_L[0], sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL1, &inv_L[1], sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL2, &inv_L[2], sizeof (DOUBLE) ); CCERR
        if ( xbox <= 0 || ybox <= 0 || zbox <= 0 || isnan(xbox) || isinf(xbox) 
             || isnan(ybox) || isinf(ybox) || isnan(zbox) || isinf(zbox) )
        {
            WARN("Wrong values for box")
            return NULL;
        }
        DOUBLE hl[6];
        hl[0] = xbox/2;
        hl[1] = ybox/2;
        hl[2] = zbox/2;
        hl[3] = -xbox/2;
        hl[4] = -ybox/2;
        hl[5] = -zbox/2;

        hl[0] = MAXD( u_cutoff + xbox/2, hl[0] );
        hl[1] = MAXD( u_cutoff + ybox/2, hl[1] );
        hl[2] = MAXD( u_cutoff + zbox/2, hl[2] );
        hl[3] = MIND( xbox/2 - 3*u_cutoff, hl[3] );
        hl[4] = MIND( ybox/2 - 3*u_cutoff, hl[4] );
        hl[5] = MIND( zbox/2 - 3*u_cutoff, hl[5] );
        
        hl[0] = MAXD( (hl[0]-hl[3])/BUCKET_SPATIAL_M, u_cutoff );
        hl[1] = MAXD( (hl[1]-hl[4])/BUCKET_SPATIAL_M, u_cutoff );
        hl[2] = MAXD( (hl[2]-hl[5])/BUCKET_SPATIAL_M, u_cutoff );
        hl[3]=-hl[3];
        hl[4]=-hl[4];
        hl[5]=-hl[5];//revert sign
        if ( bc == BC_BOX )
        {
            if ( xbox / hl[0] <= 3.01 ||
                ybox / hl[1] <= 3.01 || 
                zbox / hl[2] <= 3.01 ||
                hl[3] > xbox/2 || hl[4] > ybox/2 || hl[5] > zbox/2 )
            {
                if ( hl[3] > xbox/2 || hl[4] > ybox/2 || hl[5] > zbox/2 ) WARN("box to small")
                return NULL;
            }
        }
    }
    Verlet_impl* ret = new Verlet_impl_pbc<PBC>( coords, UDIM, u_size, u_cutoff, comp_indxs );
    return ret;
}

template <bool PBC>
INT Verlet_impl_pbc<PBC>::spatial_phase_one( DOUBLE* d_out )
{
    INT cblocks = (u_size_+cuda_block-1)/cuda_block;
    INT i = 0;
    brute_force_ = false;
    if ( PBC )
    {
        trans_to_box_all<<<cblocks,cuda_block>>>( coords_, UDIM_, u_size_ );
    }
    switch(cuda_block)
    {
        case(1024): comp_h_l<1024><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        case( 512): comp_h_l< 512><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        case( 256): comp_h_l< 256><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        case( 128): comp_h_l< 128><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        case(  64): comp_h_l<  64><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        case(  32): comp_h_l<  32><<<sm_red,cuda_block,cuda_block*sizeof(DOUBLE)*6>>>( coords_, d_out, UDIM_, u_size_ ); break;
        default: UNERR("Unimplemented");
    }
    cudaThreadSynchronize(); CCERR
    cudaMemcpy( hl_, d_out, sm_red*sizeof(DOUBLE)*6, cudaMemcpyDeviceToHost ); CCERR
    for ( i = 1; i < sm_red; ++i )
    {
        hl_[0]=MAXD(hl_[6*i  ],hl_[0]);
        hl_[1]=MAXD(hl_[6*i+1],hl_[1]);
        hl_[2]=MAXD(hl_[6*i+2],hl_[2]);
        hl_[3]=MIND(hl_[6*i+3],hl_[3]);
        hl_[4]=MIND(hl_[6*i+4],hl_[4]);
        hl_[5]=MIND(hl_[6*i+5],hl_[5]);
    }

    if ( PBC )
    {
        hl_[0] = MAXD(   u_cutoff_ + xbox/2, hl_[0] );
        hl_[1] = MAXD(   u_cutoff_ + ybox/2, hl_[1] );
        hl_[2] = MAXD(   u_cutoff_ + zbox/2, hl_[2] );
        hl_[3] = MIND( xbox/2 - 3*u_cutoff_, hl_[3] );
        hl_[4] = MIND( ybox/2 - 3*u_cutoff_, hl_[4] );
        hl_[5] = MIND( zbox/2 - 3*u_cutoff_, hl_[5] );
    }

    hl_[0] = MAXD( (hl_[0]-hl_[3])/BUCKET_SPATIAL_M, u_cutoff_ );
    hl_[1] = MAXD( (hl_[1]-hl_[4])/BUCKET_SPATIAL_M, u_cutoff_ );
    hl_[2] = MAXD( (hl_[2]-hl_[5])/BUCKET_SPATIAL_M, u_cutoff_ );
    hl_[3]=-hl_[3];
    hl_[4]=-hl_[4];
    hl_[5]=-hl_[5];//revert sign
    if ( PBC )
    {
        if ( xbox / hl_[0] <= 3.01 || ybox / hl_[1] <= 3.01 || 
            zbox / hl_[2] <= 3.01 ||
            hl_[3] > xbox/2 || hl_[4] > ybox/2 || hl_[5] > zbox/2 )
        {
            if ( hl_[3] > xbox/2 || hl_[4] > ybox/2 || hl_[5] > zbox/2 ) UNERR("Wrong correction")
            brute_force_ = true;
        }
    }
    INT keys_size = 0;
    if ( !brute_force_ )
    {
        cudaMemcpyToSymbol( dev_offset_u, hl_+3, sizeof(DOUBLE)*3 ); CCERR
        cudaMemcpyToSymbol( dev_blen_u, hl_, sizeof(DOUBLE)*3 ); CCERR

        cudaMemcpyToSymbol( d_keys_size, &keys_size, sizeof(UINT) ); CCERR
        if ( PBC )
        {
            compute_keys_value_periodic<<<cblocks,cuda_block>>>( coords_, kkeys_, kvals_, UDIM_, u_size_, u_cutoff_ );
        }
        else
        {
            compute_keys_value<<<cblocks,cuda_block>>>( coords_, kkeys_, kvals_, UDIM_, u_size_ );
        }
        cudaThreadSynchronize(); CCERR
        keys_size = 0;
        if ( PBC ) { cudaMemcpyFromSymbol( &keys_size, d_keys_size, sizeof(UINT) ); CCERR }
        u_sorter_->sort( PH_COUNT*u_size_ + keys_size );
        cudaThreadSynchronize(); CCERR
        add_dummy_beads<<<1,cuda_block>>>( kkeys_, kvals_, u_size_ );
        cudaThreadSynchronize(); CCERR

        if ( comp_indxs_ )
        {
            INT indx_blocks = ((PH_COUNT*u_size_+keys_size+COUNT-1)/COUNT + cuda_block - 1)/cuda_block;
            get_indx_spatial<<<indx_blocks,cuda_block>>>( kkeys_, indxs_, u_size_ );
            cudaThreadSynchronize(); CCERR
        }
    }
    return keys_size;
}

Verlet::Verlet() {
    o_ = NULL;
    iter_ = 0;
    Rcount_ = 0;
    Roff_ = 0;
    retsize_ = 0;
}

Verlet::~Verlet() {
    if ( o_ ) delete o_;
}

bool Verlet::init( bool pbc, DOUBLE* coords, INT UDIM, INT u_size, DOUBLE u_cutoff, bool comp_indxs )
{
    if ( pbc )
    {
        o_ = Verlet_impl_pbc<true>::createVerlet( coords, UDIM, u_size, u_cutoff + Roff_, comp_indxs );
    }
    else
    {
        o_ = Verlet_impl_pbc<false>::createVerlet( coords, UDIM, u_size, u_cutoff + Roff_, comp_indxs );
    }
    return o_;
}

INT Verlet::buildlist( DOUBLE* d_out )
{
    if ( o_ )
    {
        if ( !iter_ ) retsize_ = o_->buildlist( d_out );
        return retsize_;
    }
    return 0;
}

bool Verlet::useBruteForce()
{
    if ( o_ ) return o_->useBruteForce();
    return true;
}

void Verlet::setLists( INT* kkeys, INT* kvals, INT* indxs )
{
    if ( o_ )
    {
        o_->setLists(kkeys,kvals,indxs);
    }
}

INT* Verlet::getkkeys()
{
    if ( o_ ) return o_->getkkeys();
    return NULL;
}

INT* Verlet::getkvals()
{
    if ( o_ ) return o_->getkvals();
    return NULL;
}

INT* Verlet::getkindxs()
{
    if ( o_ ) return o_->getkindxs();
    return NULL;
}

void Verlet::setParams( INT vcount, DOUBLE vroff )
{
    iter_ = 0;
    Rcount_ = vcount;
    Roff_ = vroff;
    retsize_ = 0;
}
