/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2012: University of Warsaw                             *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@cent.uw.edu.pl>        *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include "verlet.h"

#include <float.h>
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "err.h"
#include "bucket.h"
#include "math_common.h"

MAKE_STR_IN(INT,verlet_count,1,"frequency of nonbonded neighbor list regeneration")
MAKE_STR_IN(DOUBLE,verlet_roff,10,"outer cutoff in the double-cutoff scheme")

INT verlet_dash[][3] = {{0,0,1},{0,1,0},{0,1,1},{1,0,0},{1,0,1},{1,1,0},{1,1,1},{-1,-1,1},{0,-1,1},{1,-1,1},{1,-1,0},{-1,0,1},{-1,1,1}};
INT verlet_dash_size = 13;

INT get_h_l_v( INT size, const PTR_DOUBLE coords, DOUBLE h[DIMS0], DOUBLE l[DIMS0], INT MI[DIMS0], INT LD, PTR_DOUBLE blen )
{
    INT i, j;
    for( j=0; j<3; ++j ) h[j] = -FLT_MAX;
    for( j=0; j<3; ++j ) l[j] = FLT_MAX;
    
    if ( bc == BC_BOX )
    {
        if ( l[0] < -xbox/2 || l[1] < -ybox/2 || l[2] < -zbox/2 ) WARNV(("Under box %g %g %g",l[0],l[1],l[2]))
        if ( h[0] >  xbox/2 || h[1] >  ybox/2 || h[2] >  zbox/2 ) WARNV(("Above box %g %g %g",h[0],h[1],h[2]))
        h[0] = xbox/2;
        h[1] = ybox/2;
        h[2] = zbox/2;
        l[0] = -xbox/2;
        l[1] = -ybox/2;
        l[2] = -zbox/2;
    }
    else
    {
        for( i = 0; i < size; ++i )
        {
            DOUBLE v[3]; 
            v[0] = coords[i*LD  ];
            v[1] = coords[i*LD+1];
            v[2] = coords[i*LD+2];
            trans_to_box( v );
#if USE_FLOAT
            correct_xyz(v);
#endif
            for ( j=0; j<3; ++j )
            {
                if( v[j] > h[j] )
                {
                    h[j]=v[j];
                }
                if( v[j] < l[j] )
                {
                    l[j]=v[j];
                }
            }
        }
    }
    blen[0] = MAXD( (h[0]-l[0])/BUCKET_SPATIAL_M, blen[0]);
    blen[1] = MAXD( (h[1]-l[1])/BUCKET_SPATIAL_M, blen[1]);
    blen[2] = MAXD( (h[2]-l[2])/BUCKET_SPATIAL_M, blen[2]);
    
    if ( bc == BC_BOX )
    {
        blen[0] = (h[0]-l[0]) / FLOORD( (h[0]-l[0])/blen[0] );
        blen[1] = (h[1]-l[1]) / FLOORD( (h[1]-l[1])/blen[1] );
        blen[2] = (h[2]-l[2]) / FLOORD( (h[2]-l[2])/blen[2] );
    }

    MI[0] = (INT)FLOORD((h[0]-l[0])/blen[0] + 0.001f);
    MI[1] = (INT)FLOORD((h[1]-l[1])/blen[1] + 0.001f);
    MI[2] = (INT)FLOORD((h[2]-l[2])/blen[2] + 0.001f);

    if ( bc == BC_BOX )
    {
        if ( xbox / blen[0] <= 3.01 ||
             ybox / blen[1] <= 3.01 ||
             zbox / blen[2] <= 3.01 )
        {
            return 0;
        }
    }
    else
    {
        MI[0]++;
        MI[1]++;
        MI[2]++;
    }
    return 1;
}

INT make_key_v(INT x,INT y,INT z, verlet_s* vs )
{
    INT cx = correct_indx( x, vs->mh[0] );
    INT cy = correct_indx( y, vs->mh[1] );
    INT cz = correct_indx( z, vs->mh[2] );
    INT key = ( cx << BUCKET_SPATIAL_X) |
              ( cy << BUCKET_SPATIAL_Y) |
              ( cz << BUCKET_SPATIAL_Z);
    return key;
}

int cmp_keys_v( const void* a, const void* b )
{
    const key_pid* aa = (const key_pid*)a;
    const key_pid* bb = (const key_pid*)b;
    return aa->key - bb->key;
}

int cmp_INT( const void* a, const void* b )
{
    const INT* aa = (const INT*)a;
    const INT* bb = (const INT*)b;
    return (*aa) - (*bb);
}

INT gen_keys_v( verlet_s* br, const PTR_DOUBLE _coords, INT id, INT tid, INT (*ignore)(const DOUBLE*) )
{
    const PTR_DOUBLE blen = br->blen;
    const PTR_DOUBLE l = br->l;
    INT x,y,z;
    key_pid kp;
    DOUBLE v[DIMS0];
    if ( ignore && (*ignore)(_coords ) ) return 1;

    v[0] = _coords[0];
    v[1] = _coords[1];
    v[2] = _coords[2];
    trans_to_box(v);
    x = (INT)FLOORD((v[0]-l[0])/blen[0]);
    y = (INT)FLOORD((v[1]-l[1])/blen[1]);
    z = (INT)FLOORD((v[2]-l[2])/blen[2]);

    kp.pid = id;
    kp.key = make_key_v( x, y, z, br );
    add_to_hash_table_kp( br->ht[tid], kp );
    return 1;
}

INT init_keys_end_v( verlet_s* ret, const PTR_DOUBLE _coords, INT LD, INT (*ignore_pair)(key_pid*,key_pid*), INT (*can_compute)(INT) )
{
    if ( !ret->iter )
    {
        INT i;
        /* join hash tables*/
#ifdef _OPENMP
#pragma omp parallel for
#endif
        for ( i = 0; i < ret->ht[0]->size; ++i )
        {
            INT ns = ret->ht[0]->entrs[i].end;
            INT j;
            INT tid = 0;
#ifdef _OPENMP
            tid = omp_get_thread_num();
#endif
            for ( j = 1; j < g_threads; ++j ) ns += ret->ht[j]->entrs[i].end;
            if ( ret->ht[0]->entrs[i].cap < ns )
            {
                ret->ht[0]->entrs[i].cap = ns;
                ret->ht[0]->entrs[i].vals = (key_pid*) realloc( ret->ht[0]->entrs[i].vals, sizeof(key_pid)*ns );
            }
            for ( j = 1; j < g_threads; ++j )
            {
                memcpy( ret->ht[0]->entrs[i].vals + ret->ht[0]->entrs[i].end,
                        ret->ht[j]->entrs[i].vals, ret->ht[j]->entrs[i].end * sizeof(key_pid) );
                ret->ht[0]->entrs[i].end += ret->ht[j]->entrs[i].end;
                ret->ht[j]->entrs[i].end = 0;
            }
            if ( ns != ret->ht[0]->entrs[i].end ) UNERR("Assertion fault" );
            if ( ignore_pair || ret->ht[0]->entrs[i].end < 2048 ) qsort( ret->ht[0]->entrs[i].vals, ret->ht[0]->entrs[i].end, sizeof(key_pid), cmp_keys );
            else radixsort( ret->ht[0]->entrs[i].vals, ret->ht[0]->entrs[i].end, ret->tmp[tid] );
        }
        /*for ( i = 0; i < ret->ht[0]->size; ++i )
        {
            INT j;
            for ( j = 0; j < ret->ht[0]->entrs[i].end; ++j )
            {
                printf( "kp[%d]: %d %d\n", i, ret->ht[0]->entrs[i].vals[j].key, ret->ht[0]->entrs[i].vals[j].pid );
            }
        }*/
#ifdef _OPENMP
#pragma omp parallel for
#endif
        for ( i = 0; i < ret->ht[0]->size; ++i )
        {
            if ( ret->ht[0]->entrs[i].end )
            {
                hash_entry_kp* kp = ret->ht[0]->entrs + i;
                INT tid = 0, j, beg = 0, end, k, t, x, y, z, dx, dy, dz;
                DOUBLE* crds;
#ifdef _OPENMP
                tid = omp_get_thread_num();
#endif
                crds = ret->coords[tid];
                while ( beg < kp->end )
                {
                    j = beg;
                    while( j < kp->end && kp->vals[j].key == kp->vals[beg].key )
                    {
                        crds[ (j-beg)*DIMS0  ] = _coords[ kp->vals[j].pid*LD  ];
                        crds[ (j-beg)*DIMS0+1] = _coords[ kp->vals[j].pid*LD+1];
                        crds[ (j-beg)*DIMS0+2] = _coords[ kp->vals[j].pid*LD+2];
                        ++j;
                    }
                    end = j;
                    if ( !can_compute || can_compute( kp->vals[beg].key ) )
                    {
                        if ( ignore_pair )
                        {
                            for ( k = 0; k < end-beg; ++k )
                            {
                                for ( t = 0; t < k; ++t )
                                {
                                    INT ip = ignore_pair( kp->vals + beg + k, kp->vals + beg + t );
                                    if ( !( ip ) )
                                    {
                                        DOUBLE r_sq;
                                        get_v2_sq( crds + DIMS0*k, crds + DIMS0*t, &r_sq );
                                        if ( r_sq <= ret->Rcutoff )
                                            add_pair_g( ret->g + tid, kp->vals[beg + k].pid, kp->vals[beg+t].pid );
                                    }
                                    else if ( ip == VS_IG ) break;
                                }
                            }
                        }
                        else
                        {
                            for ( k = 0; k < end-beg; ++k )
                            {
                                for ( t = 0; t < k; ++t )
                                {
                                    DOUBLE r_sq;
                                    get_v2_sq( crds + DIMS0*k, crds + DIMS0*t, &r_sq );
                                    if ( r_sq <= ret->Rcutoff )
                                        add_pair_g( ret->g + tid, kp->vals[beg + k].pid, kp->vals[beg+t].pid );
                                }
                            }
                        }
                        x = kp->vals[beg].key & BUCKET_SPATIAL_MASK;
                        y = (kp->vals[beg].key >> BUCKET_SPATIAL_Y) & BUCKET_SPATIAL_MASK;
                        z = (kp->vals[beg].key >> BUCKET_SPATIAL_Z) & BUCKET_SPATIAL_MASK;
                        for ( k = 0; k < verlet_dash_size; ++k )
                        {
                            dx = verlet_dash[k][0];
                            dy = verlet_dash[k][1];
                            dz = verlet_dash[k][2];
                            check_box( ret, ret->g + tid, kp->vals + beg, crds, end-beg,
                                       x+dx, y+dy, z+dz, crds + (end-beg)*DIMS0, _coords, LD, ignore_pair );
                        }
                    }
                    beg = j;
                }
            }
        }
#ifdef _OPENMP
#pragma omp parallel
        {
            INT i, j;
            INT* off = (INT*) malloc( sizeof(INT) * g_threads );
            INT th;
#pragma omp for
            for ( i = 0; i < ret->size; ++i )
            {
                th = ret->g[0].size[i];
                off[0] = 0;
                for ( j = 1; j < g_threads; ++j )
                {
                    off[j] = th;
                    th += ret->g[j].size[i];
                }
                if ( ret->g[0].cap[i] < th )
                {
                    ret->g[0].cap[i] = th;
                    ret->g[0].conns[i] = (INT*) realloc( ret->g[0].conns[i], sizeof(INT) * th ); CHMEM(ret->g[0].conns[i]);
                }
                for ( j = 1; j < g_threads; ++j )
                {
                    memcpy( ret->g[0].conns[i] + off[j], ret->g[j].conns[i], ret->g[j].size[i]*sizeof(INT) );
                    ret->g[j].size[i] = 0;
                }
                ret->g[0].size[i] = th;
                /*qsort( ret->g[0].conns[i], th, sizeof(INT), cmp_INT );*/
            }
            free( off );
        }
#endif
    /* testowanie czy wygenerowano wszystkie pary
        {
            bucket_s* br, *bs;
            init_bucket( &br, size, DIMS0, 0 );
            init_keys( br, size, _coords, blen, LD, ignore );
            bs = br;
#pragma omp parallel
    {
    DOUBLE* tc;
    DOUBLE* ptr_lj = _coords;
    INT k;
    INT tid = omp_get_thread_num();
    tc = (DOUBLE*)get_tmp_coords( bs, tid, DIMS0, 0 );
#ifdef _OPENMP
#pragma omp for schedule( guided )
#endif
    for ( k = 0; k < bs->size_ids; k++ )
    {
        INT b = bs->ids[2 * k];
        INT e = bs->ids[2 * k + 1];
        INT i;
        if ( 1 )
        {
            INT e_prev, molid_prev;
            for ( i = b; i < e; ++i )
            {
                INT id = bs->keys[i].pid >> MVP_BITS;
                tc[(i-b)*DIMS0  ] = ptr_lj[ LD*(id)    ];
                tc[(i-b)*DIMS0+1] = ptr_lj[ LD*(id) +1 ];
                tc[(i-b)*DIMS0+2] = ptr_lj[ LD*(id) +2 ];
            }
            e_prev = -1;
            molid_prev = -1;
            for ( i = b; i < e; ++i )
            {
                INT j;
                INT n = bs->keys[i].pid;
                if ( molid_prev == -1 || !ignore_pair( molid_prev, n>>MVP_BITS) )
                {
                    e_prev = i;
                    molid_prev = n>>MVP_BITS;
                }
                for ( j = b; j < e_prev; ++j )
                {
                    INT m = bs->keys[j].pid;
                    if ( !ignore_pair( n>>MVP_BITS, m>>MVP_BITS) && can_interact( n, m ) )
                    {
                        DOUBLE r_sq;
                        get_v2_sq( tc+(i-b)*DIMS0, tc+(j-b)*DIMS0, &r_sq );
                        if ( r_sq < ret->Rcutoff )
                        {
                            INT find = 0;
                            INT s = m>>MVP_BITS;
                            INT* p1 = bsearch( &s, ret->g[0].conns[n>>MVP_BITS], ret->g[0].size[n>>MVP_BITS], sizeof(INT), cmp_INT );
                            find = p1 != 0;
                            s = n>>MVP_BITS;
                            p1 = bsearch( &s, ret->g[0].conns[m>>MVP_BITS], ret->g[0].size[m>>MVP_BITS], sizeof(INT), cmp_INT );
                            if( find && p1 ) UNERR("pair added two times")
                            if( !find && !p1 ) UNERR("pair not find")
                        }
                    }
                }
            }
        }
    }
}
            free_bucket( br );
        }*/
        /*for ( i = 0; i < th; ++i )
            printf("%d %d %d\n",i,ret->ps[0].prs[i].id, ret->ps[0].prs[i].jd );*/
        /*ret->iter = (1 % verlet_count );*/
    }
/*{
extern INT step;
printf("%d %d %d|%d\n", step, !ret->iter, ret->iter, verlet_count);
}*/
    return 1;
}

/*void show_key( INT key )
{
    INT x = key & BUCKET_SPATIAL_MASK;
    INT y = (key >> BUCKET_SPATIAL_Y) & BUCKET_SPATIAL_MASK;
    INT z = (key >> BUCKET_SPATIAL_Z) & BUCKET_SPATIAL_MASK;
    fprintf( stderr, "%d - %d %d %d ", key, x, y, z );
}*/

VOID fix_cells( verlet_s* vs, INT* px, INT* py, INT* pz )
{
    if ( *px < 0 ) *px += vs->mh[0];
    if ( *py < 0 ) *py += vs->mh[1];
    if ( *pz < 0 ) *pz += vs->mh[2];
    if ( *px >= vs->mh[0] ) *px -= vs->mh[0];
    if ( *py >= vs->mh[1] ) *py -= vs->mh[1];
    if ( *pz >= vs->mh[2] ) *pz -= vs->mh[2];
}

VOID check_box( verlet_s* vs, graph* g, key_pid* kp1, DOUBLE* crds1, INT s1, INT x, INT y, INT z, DOUBLE* crds2, DOUBLE* _coords, INT LD, INT (*ignore_pair)(key_pid*,key_pid*) )
{
    key_pid nkey;
    hash_entry_kp* kp2;
    key_pid* ptr, *end;
    INT s2 = 0, i, j;

    fix_cells( vs, &x, &y, &z );

    nkey.key = make_key_v( x, y, z, vs );
    nkey.pid = -1;

    kp2 = find_elem_ht_kp( vs->ht[0], nkey.key );
    ptr = bsearch( &nkey, kp2->vals, kp2->end, sizeof(key_pid), cmp_keys_v );
    if ( !ptr ) return; /* no key */
    while ( ptr - 1 >= kp2->vals && (ptr-1)->key == nkey.key ) ptr--;
    end = ptr;
    while ( end < kp2->vals + kp2->end && end->key == nkey.key )
    {
        crds2[ s2*DIMS0  ] = _coords[ end->pid*LD  ];
        crds2[ s2*DIMS0+1] = _coords[ end->pid*LD+1];
        crds2[ s2*DIMS0+2] = _coords[ end->pid*LD+2];
        ++end;
        ++s2;
    }
    if ( ignore_pair )
    {
        INT ob, oe;
        INT olast = 0;
        key_pid prev_kp;
        prev_kp.pid = -1;
        prev_kp.key= 0;
        for ( i = 0; i < s1; ++i )
        {
            INT ip = 0;
            if ( prev_kp.pid != -1 ) ip = ignore_pair( kp1 + i, &prev_kp );
            if ( !ip || ip == VS_IS )
            {
                ob = oe = -1;
                for ( ; olast < s2; ++olast )
                {
                    ip = ignore_pair( kp1 + i, ptr + olast );
                    if ( !ip || ip == VS_IS )
                    {
                        if ( ptr[olast].pid > kp1[i].pid ) break;
                    }
                    else
                    {
                        if ( ob == -1 ) ob = olast;
                    }
                }
                if ( ob != -1 ) oe = olast;
                else oe = ob = s2;
                prev_kp.pid = kp1[i].pid;
            }
            for ( j = 0; j < ob; ++j )
            {
                DOUBLE r_sq;
                get_v2_sq( crds1 + DIMS0*i, crds2 + DIMS0*j, &r_sq );
                if ( r_sq <= vs->Rcutoff )
                {
                    INT ip = ignore_pair( kp1 + i, ptr + j );
                    if ( !ip ) add_pair_g( g, kp1[i].pid, ptr[j].pid );
                    else if ( ip == VS_IG ) UNERR("ASSERTION")
                }
            }
            for ( j = oe; j < s2; ++j )
            {
                DOUBLE r_sq;
                get_v2_sq( crds1 + DIMS0*i, crds2 + DIMS0*j, &r_sq );
                if ( r_sq <= vs->Rcutoff )
                {
                    INT ip = ignore_pair( kp1 + i, ptr + j );
                    if ( !ip ) add_pair_g( g, kp1[i].pid, ptr[j].pid );
                    else if ( ip == VS_IG ) UNERR("ASSERTION")
                }
            }
        }
    }
    else
    {
        for ( i = 0; i < s1; ++i )
            for ( j = 0; j < s2; ++j )
            {
                DOUBLE r_sq;
                get_v2_sq( crds1 + DIMS0*i, crds2 + DIMS0*j, &r_sq );
                if ( r_sq <= vs->Rcutoff )
                    add_pair_g( g, kp1[i].pid, ptr[j].pid );
            }
    }
}

/*#include "bucket.h"*/

INT init_keys_v( verlet_s* ret, INT size, const PTR_DOUBLE _coords, DOUBLE blen, INT LD, INT (*ignore)(const DOUBLE*), INT (*ignore_pair)(key_pid*,key_pid*), INT (*can_compute)(INT) )
{
    if ( !ret->iter )
    {
        DOUBLE h[3];
        INT i;

        ret->Rcutoff = blen + verlet_roff;
        ret->Rcutoff = ret->Rcutoff * ret->Rcutoff;
        ret->blen[0] = ret->blen[1] = ret->blen[2] = blen + verlet_roff;
        if ( ! get_h_l_v( size, _coords, h, ret->l, ret->mh, LD, ret->blen ) ) return 0;
        /*fprintf( stderr, "%g blen %g %g %g\n", blen, ret->blen[0], ret->blen[1], ret->blen[2] );*/
        ret->size = size;
#ifdef _OPENMP
#pragma omp parallel for
#endif
        for ( i = 0; i < ret->ht[0]->size; ++i )
        {
            ret->ht[0]->entrs[i].end = 0;/* reset hash table entries */
            ret->g[0].size[i] = 0;
        }

#ifdef _OPENMP
#pragma omp parallel for
#endif
        for ( i = 0; i < size; ++i )
        {
            INT tid = 0;
#ifdef _OPENMP
            tid = omp_get_thread_num();
#endif
            gen_keys_v( ret, _coords + i*LD, i, tid, ignore );
        }
        /*for ( i = 0; i < ret->ht[0]->size; ++i )
        {
            INT j;
            for ( j = 0; j < ret->ht[0]->entrs[i].end; ++j )
            {
                if ( ret->ht[0]->entrs[i].vals[j].pid == 34704 ||
                     ret->ht[0]->entrs[i].vals[j].pid == 60017 )
                    fprintf( stderr, "kp: %d %d\n", ret->ht[0]->entrs[i].vals[j].key, ret->ht[0]->entrs[i].vals[j].pid );
            }
        }*/
    }
    return init_keys_end_v( ret, _coords, LD, ignore_pair, can_compute );
}

VOID add_pair_g( graph* g, INT id, INT jd )
{
    if ( jd < id )
    {
        INT tmp = id;
        id = jd;
        jd = tmp;
    }
    if ( g->size[id] >= g->cap[id] )
    {
        INT cap = g->cap[id] = g->cap[id]*2 + 1;
        g->conns[id] = (INT*) realloc( g->conns[id], sizeof(INT) * cap ); CHMEM(g->conns[id]);
    }
    g->conns[ id ][ g->size[id] ] = jd;
    g->size[id]++;
}

void init_verlet( verlet_s** pbr, INT size )
{
    INT i;
    verlet_s* br = (verlet_s*) malloc( sizeof(verlet_s) ); CHMEM(br);
    br->iter = 0;
    br->size = size;
    br->tmp = (key_pid**) malloc( sizeof(key_pid*)*g_threads ); CHMEM(br->tmp);
    br->coords = (DOUBLE**) malloc( sizeof(DOUBLE*) * g_threads); CHMEM(br->coords);
    for ( i = 0; i < g_threads; ++i )
    {
        br->coords[i] = (DOUBLE*) malloc( sizeof(DOUBLE) * DIMS0 * size );
        CHMEM(br->coords[i]);
        br->tmp[i] = (key_pid*) malloc( sizeof(key_pid)*size ); CHMEM(br->tmp);
        CHMEM(br->tmp[i]);
    }
    br->ht = (hash_table_kp**) malloc( sizeof(hash_table_kp*) * g_threads ); CHMEM( br->ht );
    for ( i = 0; i < g_threads; ++i )
    {
        br->ht[i] = create_hash_table_kp( size, 2 );
    }
    br->g = (graph*) malloc( sizeof(graph) * g_threads );
    for ( i = 0; i < g_threads; ++i )
    {
        INT j = 0;
        br->g[i].conns = (INT**) malloc( sizeof(INT*) * size );
        br->g[i].cap = (INT*) malloc( sizeof(INT) * size );
        br->g[i].size = (INT*) malloc( sizeof(INT) * size );
        for ( j = 0; j < size; ++j )
        {
            br->g[i].cap[j] = 4;
            br->g[i].size[j] = 0;
            br->g[i].conns[j] = (INT*) malloc( sizeof(INT) * br->g[i].cap[j] );
        }
    }
    *pbr = br;
}

void free_verlet( verlet_s* br )
{
    INT i, j;
    if ( br )
    {
        for ( i = 0; i < g_threads; ++i )
        {
            for ( j = 0; j < br->size; ++j )
            {
                free( br->g[i].conns[j] );
            }
            free( br->coords[i] );
            free_hash_table_kp( br->ht[i] );
            free( br->g[i].conns );
            free( br->g[i].size );
            free( br->g[i].cap );
            free( br->tmp[i] );
        }
        free( br->coords );
        free( br->tmp );
        free( br->ht );
        free( br->g );
        free( br );
    }
}
