/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cuda_runtime_api.h>
#include <cublas.h>

#include "diff_tensor.cuh"

#include "../diff_tensor.h"
#include "../../cuda.h"
#include "../../trans.h"

#include "../../../common/err.h"
#include "../../../common/myblas.h"
#include "../../../common/rand_move.h"

DOUBLE* d_eps = NULL;
DOUBLE* h_eps = NULL;
DOUBLE* d_eff = NULL;
DOUBLE* d_C   = NULL;
__constant__ INT dev_size;

template<int cuda_block>
__device__ void eps_reduction( DOUBLE e_curr, DOUBLE* out )
{
    extern __shared__ DOUBLE sh[];
    int tid = threadIdx.x;

    __syncthreads( );
    sh[tid] = e_curr;
    __syncthreads( );
    if ( cuda_block >= 512 )
    {
        if ( tid < 256 ) sh[tid] += sh[tid + 256];
        __syncthreads( );
    }
    if ( cuda_block >= 256 )
    {
        if ( tid < 128 ) sh[tid] += sh[tid + 128];
        __syncthreads( );
    }
    if ( cuda_block >= 128 )
    {
        if ( tid < 64 ) sh[tid] += sh[tid + 64];
        __syncthreads( );
    }
    volatile DOUBLE *redE = sh;
    if ( tid < 32 )
    {
        if ( cuda_block >= 64 ) redE[tid] += redE[tid + 32];
        if ( cuda_block > 32 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
        else if ( tid < 16 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
    }
    if ( tid == 0 )
    {
        out[blockIdx.x] = redE[0];
    }
}

template<INT cuda_block >
__global__ void comp_geyer_beg( DOUBLE* ptr_d, DOUBLE* ptr_x, DOUBLE* ptr_eff, DOUBLE* ptr_c, DOUBLE* ptr_eps )
{
    DOUBLE eps_curr = 0;
    DOUBLE Feff = 0;
    DOUBLE C = 0;
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    INT LD = (gridDim.x * blockDim.x/DIMS0)*TENSORM;

    for ( INT k = 0; k < DIMS0*dev_size; ++k )
    {
        DOUBLE dk = ptr_d[LD*k+k];
        DOUBLE dik = 0;
        if ( id<k ) dik = ptr_d[LD*id + k];
        else dik = ptr_d[LD*k + id];
        if ( k != id )
        {
            Feff += ptr_x[k] * dik;
            C += dik*dik/dk;
            eps_curr+=dik;
        }
    }
    ptr_eff[id] = Feff;
    ptr_c[id] = C;
    if ( id >= DIMS0*dev_size ) eps_curr = 0;
    eps_reduction<cuda_block>( eps_curr/2, ptr_eps );
}

template<int cuda_block >
__global__ void comp_geyer_end( DOUBLE* ptr_d, DOUBLE* ptr_x, DOUBLE* ptr_eff, DOUBLE* ptr_c, DOUBLE beta, DOUBLE* out)
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    DOUBLE Feff = ptr_eff[id];
    DOUBLE C = ptr_c[id];
    INT LD = (gridDim.x * blockDim.x/DIMS0)*TENSORM;
    DOUBLE dii = ptr_d[LD*id+id];
    C *= beta*beta/dii;
    C = rsqrtf( 1 + C );
    out[id] += C*(beta*Feff + dii*ptr_x[id]);
}

extern "C"
void compute_r( DOUBLE* _D )
{
    INT i;
    DOUBLE a, eps;

    gauss_rand( X, size * DIMS0 );
    switch ( hydro )
    {
        case DIFF_NONE:
        {
            for ( i = 0; i < size * DIMS0; ++i )
            {
                X[ i ] = X[ i ] * SQRTD( diag_D[i] * 2 * curr_dt );
            }
            cudaMemcpy( d_X, X, sizeof(DOUBLE)*size*DIMS0, cudaMemcpyHostToDevice ); CCERR
            CUBLAS_AXPYD( size*DIMS0, 1.0f, d_X, 1, d_Rs0, 1 ); cublasCheckError(__FILE__,__LINE__);
            break;
        }
        case DIFF_CHOLESKY:
        {
            /* computation on CPU */
            a = SQRTD( 2 * curr_dt );
            for ( i = 0; i < size * DIMS0; ++i )
            {
                X[i] *= a;
            }
            cudaMemcpy( d_X, X, sizeof(DOUBLE)*size*DIMS0, cudaMemcpyHostToDevice ); CCERR
            if ( curr_iter != save_iter )
            {
                if ( my_dpotrf( _D, size*DIMS0, csize*TENSORM ) )
                    UNERR("Cholesky decomposition error")
            }
            my_dtrmv( _D, d_X, d_Rs0, size*DIMS0, csize*TENSORM );
            break;
        }
        case DIFF_GEYER:
        {
            for ( i = 0; i < size * DIMS0; ++i )
            {
                X[ i ] *= SQRTD( 2 * curr_dt * inv_diag_D[i] );
            }
            cudaMemcpy( d_X, X, sizeof(DOUBLE)*size*DIMS0, cudaMemcpyHostToDevice ); CCERR

            int to_get = DIMS0*cblocks;
            if ( geyer_on_the_fly )
            {
                comp_geyer_beg_in_fly( d_coord, d_X, d_eff, d_C, d_eps );
                to_get = cblocks;
            }
            else
            {
                INT shSize = sizeof(DOUBLE)*cuda_block;
                switch ( cuda_block )
                {
                    case(1024):comp_geyer_beg<1024><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                    case( 512):comp_geyer_beg< 512><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                    case( 256):comp_geyer_beg< 256><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                    case( 128):comp_geyer_beg< 128><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                    case(  64):comp_geyer_beg<  64><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                    case(  32):comp_geyer_beg<  32><<< DIMS0*cblocks, cuda_block,shSize>>>( d_D, d_X, d_eff, d_C, d_eps );break;
                }
            }
            cudaThreadSynchronize(); CCERR
            cudaMemcpy( h_eps, d_eps, to_get*sizeof(DOUBLE),cudaMemcpyDeviceToHost); CCERR
            eps=0;
            for( INT i = 0; i < to_get; ++i )
            {
                eps += h_eps[i];
            }
            eps /= size * DIMS0 / ((DOUBLE)2);
            eps /= size * DIMS0 - 1;
            if ( eps > 1.0 || isnan(eps) || isinf(eps) )
            {
                WARNV(("epsilon(%g) truncated to 1.0",eps))
                eps = 1.0;
            }
            a = (3 * size - 1) * eps * eps-(3 * size - 2) * eps;
            DOUBLE beta = 0.5;
            if ( abs( a ) < 0.0000001f )/*we assume no correlation*/
            {
                WARN("Epsilon == 0.0")
            }
            else
            {
                beta = (1 - SQRTD( 1 - a )) / a;
            }
            if ( hydro == DIFF_GEYER && algorithm < DIFF_ALG_IGT_CONST )
            {
                comp_geyer_end_in_fly( d_coord, d_X, d_eff, d_C, beta, d_Rs0 );
            }
            else
            {
                switch(cuda_block)
                {
                    case(1024):comp_geyer_end<1024><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                    case( 512):comp_geyer_end< 512><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                    case( 256):comp_geyer_end< 256><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                    case( 128):comp_geyer_end< 128><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                    case(  64):comp_geyer_end<  64><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                    case(  32):comp_geyer_end<  32><<< DIMS0*cblocks, cuda_block >>>( d_D, d_X, d_eff, d_C, beta, d_Rs0 );break;
                }
            }
            cudaThreadSynchronize(); CCERR
            break;
        }
        case DIFF_CHEBYSHEV:
        {
            UNERR( "Unimplemented CHEBYSHEV method")
        }
    }
    save_iter = curr_iter;
}

extern "C"
void init_cuda_rand()
{
    if ( hydro == DIFF_GEYER )
    {
        cudaMalloc( (void**) &d_eps, sizeof(DOUBLE)*DIMS0*csize); CCERR
        cudaMalloc( (void**) &d_eff, sizeof(DOUBLE)*DIMS0*csize); CCERR
        cudaMalloc( (void**) &d_C, sizeof(DOUBLE)*DIMS0*csize);   CCERR
        h_eps = (DOUBLE*) malloc( sizeof(DOUBLE)*DIMS0*csize );   CHMEM(h_eps)
    }
    cudaMemcpyToSymbol( dev_size, &size, sizeof(INT) ); CCERR
}

extern "C"
void free_cuda_rand()
{
    cudaFree( d_eps ); CCERR
    cudaFree( d_eff ); CCERR
    cudaFree( d_C );   CCERR
    free( h_eps );
}
