/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cuda_runtime_api.h>

#include "../diff_tensor.h"
#include "../../cuda.h"
#include "../../trans.h"

#include "../../../common/err.h"
#include "../../../common/math_common.h"

__constant__ DOUBLE dev_preDii;
__constant__ DOUBLE dev_preDij;
__constant__ DOUBLE dev_ewald_alpha;
__constant__ DOUBLE dev_ewald_alpha2;
__constant__ DOUBLE dev_ewald_alpha3;
__constant__ DOUBLE dev_M_PI_ALPHA_SQ;
__constant__ DOUBLE dev_max_sq;
__constant__ DOUBLE dev_box0;
__constant__ DOUBLE dev_box1;
__constant__ DOUBLE dev_box2;
__constant__ DOUBLE dev_invL0;
__constant__ DOUBLE dev_invL1;
__constant__ DOUBLE dev_invL2;
__constant__ INT dev_size;

__device__ static void tensor_calc_pos_diag( DOUBLE_4 a, DOUBLE* D, int LD )
{
    DOUBLE_3 tmp = {0.0f, 0.0f, 0.0f};
    tmp.x = dev_preDii / a.w;
    *((DOUBLE_3*) D) = tmp;
    tmp.y = tmp.x;
    tmp.x = 0;
    *((DOUBLE_3*) (D + LD)) = tmp;
    tmp.z = tmp.y;
    tmp.y = 0;
    *((DOUBLE_3*) (D + 2 * LD)) = tmp;
}

__device__ DOUBLE_3 add3(DOUBLE_3 a, DOUBLE_3 b)
{
    a.x += b.x;
    a.y += b.y;
    a.z += b.z;
    return a;
}

__device__ DOUBLE_3 sub3(DOUBLE_3 a, DOUBLE_3 b)
{
    a.x -= b.x;
    a.y -= b.y;
    a.z -= b.z;
    return a;
}

__device__ static void tensor_calc_pos( DOUBLE_4 a, DOUBLE_4 b, DOUBLE* D, int LD )
{
    DOUBLE_4 r_ij;
    r_ij.x = a.x - b.x;
    r_ij.y = a.y - b.y;
    r_ij.z = a.z - b.z;

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    DOUBLE r_norm = sqrtf( r_sq );
    DOUBLE_3 d1,d2,d3;

    DOUBLE sigma = (a.w + b.w) / 2;
    DOUBLE pre = (r_norm < a.w + b.w) ? (dev_preDii / sigma) : (dev_preDij / r_norm);
    DOUBLE sigm_r = (a.w * a.w + b.w * b.w) / r_sq;

    DOUBLE divr = (r_norm < a.w + b.w) ? (sigma * r_norm * 32) : (r_sq);
    DOUBLE mult = (r_norm < a.w + b.w) ? (3) : (1 - sigm_r);
    DOUBLE part1 = (r_norm < a.w + b.w) ? (1 - (9 * r_norm / (32 * sigma))) : (1 + (sigm_r / 3));

    d3.z = d2.y = d1.x = part1;

    d1.x += mult * r_ij.x * r_ij.x / divr;
    d1.x *= pre;

    d2.y += mult * r_ij.y * r_ij.y / divr;
    d2.y *= pre;

    d3.z += mult * r_ij.z * r_ij.z / divr;
    d3.z *= pre;

    d2.x = d1.y = pre * mult * r_ij.x * r_ij.y / divr;
    d3.x = d1.z = pre * mult * r_ij.x * r_ij.z / divr;
    d2.z = d3.y = pre * mult * r_ij.z * r_ij.y / divr;

    *((DOUBLE_3*) D) = d1;
    *((DOUBLE_3*) (D + LD)) = d2;
    *((DOUBLE_3*) (D + 2 * LD)) = d3;
}

__device__ void get_ev2_norms( DOUBLE x1, DOUBLE y1, DOUBLE z1, DOUBLE_3& e, DOUBLE& inv_r, DOUBLE& r_norm, DOUBLE& r_sq )
{
    inv_r = 0.0;
    e.x = x1;
    e.y = y1;
    e.z = z1;

    r_sq = (x1)*(x1) + (y1)*(y1) + (z1)*(z1);
    r_norm = sqrtf( r_sq );
    inv_r = 1 / r_norm;

    e.x *= inv_r;
    e.y *= inv_r;
    e.z *= inv_r;
}

__device__ void comp_OQ( int f, int g, int h, DOUBLE_4 v, DOUBLE_3* O1, DOUBLE_3* O2, DOUBLE_3* Q1, DOUBLE_3* Q2 )
{
    DOUBLE r_norm, r_sq, inv_r;
    DOUBLE_3 e;

    get_ev2_norms( v.x + h, v.y + f, v.z + g, e, inv_r, r_norm, r_sq );
    if ( r_norm != 0.0f )
    {
        DOUBLE inv_r3 = 1 / (r_norm * r_sq);
        DOUBLE sc = erfc( r_norm * dev_ewald_alpha ) * inv_r;
        DOUBLE sc2 = 2 * dev_ewald_alpha * M1_SQRTPIF * expf( -dev_ewald_alpha2 * r_sq );
        DOUBLE sc_ek = e.x * (sc + sc2);

        (*O1).x += sc + sc_ek * e.x;
        sc_ek = e.y * (sc + sc2);
        (*O2).y += sc + sc_ek * e.y;
        (*O1).y += e.x * sc_ek;
        sc_ek = e.z * (sc + sc2);
        (*O2).x += sc + sc_ek * e.z;
        (*O1).z += e.x * sc_ek;
        (*O2).z += e.y * sc_ek;

        sc = inv_r3 * (erfc( r_norm * dev_ewald_alpha ) + 2 * dev_ewald_alpha * r_norm * M1_SQRTPIF * expf( -dev_ewald_alpha2 * r_sq ));
        sc2 = -4 * dev_ewald_alpha3 * M1_SQRTPIF * expf( -dev_ewald_alpha2 * r_sq );
        sc_ek = e.x * (-3 * sc + sc2);
        (*Q1).x += sc + sc_ek * e.x;
        sc_ek = e.y * (-3 * sc + sc2);
        (*Q2).y += sc + sc_ek * e.y;
        (*Q1).y += e.x * sc_ek;
        sc_ek = e.z * (-3 * sc + sc2);
        (*Q2).x += sc + sc_ek * e.z;
        (*Q1).z += e.x * sc_ek;
        (*Q2).z += e.y * sc_ek;
    }

    get_ev2_norms( h, f, g, e, inv_r, r_norm, r_sq );
    if ( r_norm != 0.0f )
    {
        DOUBLE sc = 2 * expf( dev_M_PI_ALPHA_SQ * r_sq ) / (M_PIF * r_sq);
        DOUBLE sc2 = -1 + dev_M_PI_ALPHA_SQ * r_sq;
        sc *= cos( 2 * M_PIF * (v.x * h + v.y * f + v.z * g) );
        sc2 *= sc;
        DOUBLE sc_ek = sc2 * e.x;
        (*O1).x += sc + sc_ek * e.x;
        sc_ek = e.y * sc2;
        (*O2).y += sc + sc_ek * e.y;
        (*O1).y += e.x * sc_ek;
        sc_ek = e.z * sc2;
        (*O2).x += sc + sc_ek * e.z;
        (*O1).z += e.x * sc_ek;
        (*O2).z += e.y * sc_ek;

        sc = 4 * M_PIF * expf( dev_M_PI_ALPHA_SQ * r_sq );
        sc *= cosf( 2 * M_PIF * (v.x * h + v.y * f + v.z * g) );
        sc_ek = e.x * sc;
        (*Q1).x += sc_ek * e.x;
        sc_ek = e.y * sc;
        (*Q2).y += sc_ek * e.y;
        (*Q1).y += e.x * sc_ek;
        sc_ek = e.z * sc;
        (*Q2).x += sc_ek * e.z;
        (*Q1).z += e.x * sc_ek;
        (*Q2).z += e.y * sc_ek;
    }
}

template<int ewald_max>
__device__ void tensor_calc_pos_ewald( DOUBLE_4 a, DOUBLE_4 b, DOUBLE* D, int LD )
{
    DOUBLE_4 v;
    v.x = a.x - b.x;
    v.y = a.y - b.y;
    v.z = a.z - b.z;

    v.x *= dev_invL0;
    v.y *= dev_invL1;
    v.z *= dev_invL2;

    DOUBLE_3 O1, O2, Q1, Q2 = {0.0f, 0.0f, 0.0f};
    O1 = O2 = Q1 = Q2;
    int f, g, h;
    comp_OQ( ewald_max, 0, 0, v, &O1, &O2, &Q1, &Q2 );
    if ( ewald_max )comp_OQ( -ewald_max, 0, 0, v, &O1, &O2, &Q1, &Q2 );
    for ( h = -ewald_max + 1; h < ewald_max; ++h )
    {
        int next_f = sqrtf( dev_max_sq - h * h );
        for ( f = -next_f; f <= next_f; ++f )
        {
            int next_g = sqrtf( dev_max_sq - h * h - f * f );
            for ( g = -next_g; g <= next_g; ++g )
            {
                comp_OQ( h, f, g, v, &O1, &O2, &Q1, &Q2 );
            }
        }
    }
    const DOUBLE preO = 0.75f * dev_invL0;
    const DOUBLE sigma_sq = 0.5f * (a.w * a.w + b.w * b.w);
    const DOUBLE preQ = 0.5f * sigma_sq * dev_invL0 * dev_invL0 * dev_invL0;
    DOUBLE_3 d1, d2, d3;
    d1.x = dev_preDii * (preO * O1.x + preQ * Q1.x);
    d1.y = dev_preDii * (preO * O1.y + preQ * Q1.y);
    d1.z = dev_preDii * (preO * O1.z + preQ * Q1.z);

    d2.x = d1.y;
    d2.y = dev_preDii * (preO * O2.y + preQ * Q2.y);
    d2.z = dev_preDii * (preO * O2.z + preQ * Q2.z);

    d3.x = d1.z;
    d3.y = d2.z;
    d3.z = dev_preDii * (preO * O2.x + preQ * Q2.x);

    *((DOUBLE_3*) D) = d1;
    *((DOUBLE_3*) (D + LD)) = d2;
    *((DOUBLE_3*) (D + 2 * LD)) = d3;
}

__device__ void tensor_calc_pos_correction( DOUBLE_4 a, DOUBLE_4 b, DOUBLE* D, int LD )
{
    DOUBLE_4 r_ij;
    r_ij.x = a.x - b.x;
    r_ij.y = a.y - b.y;
    r_ij.z = a.z - b.z;

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    if( r_sq < (a.w + b.w)*(a.w + b.w) )
    {
        DOUBLE_3 bd1,bd2,bd3;
        {
            DOUBLE r_norm = sqrtf( r_sq );

            DOUBLE sigma = (a.w + b.w) / 2;
            DOUBLE pre = dev_preDii / sigma;

            DOUBLE divr = sigma * r_norm * 32;

            bd3.z = bd2.y = bd1.x = 1 - (9 * r_norm / (32 * sigma));

            bd1.x += 3 * r_ij.x * r_ij.x / divr;
            bd1.x *= pre;

            bd2.y += 3 * r_ij.y * r_ij.y / divr;
            bd2.y *= pre;

            bd3.z += 3 * r_ij.z * r_ij.z / divr;
            bd3.z *= pre;

            bd2.x = bd1.y = pre * 3 * r_ij.x * r_ij.y / divr;
            bd3.x = bd1.z = pre * 3 * r_ij.x * r_ij.z / divr;
            bd2.z = bd3.y = pre * 3 * r_ij.z * r_ij.y / divr;
        }
        {
            DOUBLE r_norm = sqrtf( r_sq );
            DOUBLE_3 d1,d2,d3;

            DOUBLE pre = dev_preDij / r_norm;
            DOUBLE sigm_r = (a.w * a.w + b.w * b.w) / r_sq;

            DOUBLE divr = r_sq;
            DOUBLE mult = 1 - sigm_r;
            DOUBLE part1 = 1 + (sigm_r / 3);

            d3.z = d2.y = d1.x = part1;

            d1.x += mult * r_ij.x * r_ij.x / divr;
            d1.x *= pre;

            d2.y += mult * r_ij.y * r_ij.y / divr;
            d2.y *= pre;

            d3.z += mult * r_ij.z * r_ij.z / divr;
            d3.z *= pre;

            d2.x = d1.y = pre * mult * r_ij.x * r_ij.y / divr;
            d3.x = d1.z = pre * mult * r_ij.x * r_ij.z / divr;
            d2.z = d3.y = pre * mult * r_ij.z * r_ij.y / divr;

            bd1 = sub3( bd1, d1 );
            bd2 = sub3( bd2, d2 );
            bd3 = sub3( bd3, d3 );
        }

        *((DOUBLE_3*) D) = add3( bd1, *((DOUBLE_3*) D) );
        *((DOUBLE_3*) (D + LD)) = add3( bd2, *((DOUBLE_3*) (D + LD)) );
        *((DOUBLE_3*) (D + 2 * LD)) = add3( bd3, *((DOUBLE_3*) (D + 2 * LD)) ) ;
    }
}

template<int ewald_real>
__device__ void tensor_calc_pos_ewald_diag( DOUBLE_4 a, DOUBLE* D, int LD )
{
    DOUBLE_3 tab;
    tensor_calc_pos_ewald<ewald_real > (a, a, (DOUBLE*) &tab, 0);
    DOUBLE sc3 = dev_ewald_alpha * a.w * dev_invL0;
    sc3 = sc3 * sc3 * sc3 * M1_SQRTPIF / 3;
    sc3 = 0.75f * dev_invL0 * 1.5f * dev_ewald_alpha * a.w * M1_SQRTPIF * dev_invL0 +
        0.5f * a.w * a.w * dev_invL0 * dev_invL0 * dev_invL0 * sc3;
    sc3 *= dev_preDii;
    DOUBLE_3 tmp = {0.0f, 0.0f, 0.0f};
    tmp.x = dev_preDii / a.w - sc3 + tab.z; // z is only correct due to 0 in LD
    *((DOUBLE_3*) D) = tmp;
    tmp.y = tmp.x;
    tmp.x = 0;
    *((DOUBLE_3*) (D + LD)) = tmp;
    tmp.z = tmp.y;
    tmp.y = 0;
    *((DOUBLE_3*) (D + 2 * LD)) = tmp;
}

__device__ void comp_Dt( void* ptr_coord, void* D )
{
    const INT idx = blockIdx.x * blockDim.x + threadIdx.x;
    const INT idy = blockIdx.y * blockDim.y + threadIdx.y;

    DOUBLE* Dtab = (DOUBLE*) D;
    DOUBLE_4* ctab = (DOUBLE_4*) ptr_coord;
    const INT LD = gridDim.x * blockDim.x * TENSORM;

    if ( idx <= idy )
    {
        DOUBLE_4 curr_x = ctab[ idx ];
        DOUBLE_4 curr_y = ctab[ idy ];
        //__syncthreads( );
        if ( idx != idy )
            tensor_calc_pos( curr_y, curr_x, Dtab + 3 * idx * LD + 3 * idy, LD );
        else
            tensor_calc_pos_diag( curr_x, Dtab + idx * LD * 3 + idx * 3, LD );
    }
}

template<int ewald_real>
__device__ void comp_Dt_ewald_sq( void* ptr_coord, void* D )
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int idy = blockIdx.y * blockDim.y + threadIdx.y;

    DOUBLE* Dtab = (DOUBLE*) D;
    DOUBLE_4* ctab = (DOUBLE_4*) ptr_coord;
    const int LD = gridDim.x * blockDim.x * TENSORM;

    if ( idx <= idy )
    {
        DOUBLE_4 curr_x = ctab[ idx ];
        DOUBLE_4 curr_y = ctab[ idy ];
        //__syncthreads( );
        if ( idx != idy )
            tensor_calc_pos_ewald<ewald_real > (curr_y, curr_x, Dtab + 3 * idx * LD + 3 * idy, LD);
        else
            tensor_calc_pos_ewald_diag<ewald_real > (curr_x, Dtab + idx * LD * 3 + idx * 3, LD);
    }
}

__device__ void comp_Dt_ewald_correction( void* ptr_coord, void* D )
{
    INT idx = blockIdx.x * blockDim.x + threadIdx.x;
    INT idy = blockIdx.y * blockDim.y + threadIdx.y;

    DOUBLE* Dtab = (DOUBLE*) D;
    DOUBLE_4* ctab = (DOUBLE_4*) ptr_coord;
    const int LD = gridDim.x * blockDim.x * TENSORM;

    if ( idx < idy )
    {
        DOUBLE_4 curr_x = ctab[ idx ];
        DOUBLE_4 curr_y = ctab[ idy ];
        tensor_calc_pos_correction(curr_y, curr_x, Dtab + 3 * idx * LD + 3 * idy, LD);
    }
}

template<bool ewald, int ewald_real >
__global__ void choose_comp_D( void* coord_ptr, void* _D )
{
    if ( !ewald )
    {
        comp_Dt( coord_ptr, _D );
    }
    else
    {
        comp_Dt_ewald_sq< ewald_real >( coord_ptr, _D );
    }
}

__global__ void choose_comp_D_correction( void* coord_ptr, void* _D )
{
    comp_Dt_ewald_correction( coord_ptr, _D );
}

extern "C"
void compute_D( DOUBLE* _D )
{
    if ( !hydro ) return; /* End if no hydro */

    /*double mintime = 111111;
    int mx,my;
    for ( int xtdim = 1; xtdim <= 256; xtdim*=2)
    for ( int ytdim = 1; ytdim <= 256; ytdim*=2)
    if(xtdim*ytdim<=128)
    {
    printf("%d %d\n",xtdim,ytdim);
    mstart();*/
    if ( !ewald )
    {
        const int xtdim = 1;
        const int ytdim = (128 < csize ) ? 128 : csize;
        dim3 dim_block( xtdim, ytdim );
        dim3 dim_grid( csize / xtdim, csize / ytdim );
        if ( dim_block.x*dim_grid.x != (UINT)csize )
        {
            dim_block.x = cuda_block;
            dim_grid.x = csize / dim_block.x;
        }
        if ( dim_block.y*dim_grid.y != (UINT)csize )
        {
            dim_block.y = cuda_block;
            dim_grid.y = csize / dim_block.y;
        }
        if ( dim_block.x*dim_grid.x != (UINT)csize ||
             dim_block.y*dim_grid.y != (UINT)csize )
        {
            WARNV(("Wrong sizes %d %d",dim_block.x*dim_grid.x,dim_block.y*dim_grid.y))
            UNERRDW
        }
        if ( bc != BC_BOX )
        {
            choose_comp_D< false, 0 > <<< dim_grid, dim_block >>>(d_coord, _D);
        }
        else
        {
            UNERR("Assertion failure")
        }
    }
    else
    {
        int xtdim = 2;
        int ytdim = 64;
        if ( xtdim*ytdim > csize )
        {
            xtdim = 1;
            ytdim = csize;
        }
        dim3 dim_block( xtdim, ytdim );
        dim3 dim_grid( csize / xtdim, csize / ytdim );
        if ( dim_block.x*dim_grid.x != (UINT)csize )
        {
            dim_block.x = cuda_block;
            dim_grid.x = csize / dim_block.x;
        }
        if ( dim_block.y*dim_grid.y != (UINT)csize )
        {
            dim_block.y = cuda_block;
            dim_grid.y = csize / dim_block.y;
        }
        if ( dim_block.x*dim_grid.x != (UINT)csize ||
             dim_block.y*dim_grid.y != (UINT)csize )
        {
            WARNV(("Wrong sizes %d %d",dim_block.x*dim_grid.x,dim_block.y*dim_grid.y))
            UNERRDW
        }

        if( ewald_method == EWALD_METHOD_SMITH )
        {
            if ( ewald_real == 0 && ewald_recip == 0 )
                choose_comp_D<true, 0 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else if ( ewald_real == 1 && ewald_recip == 1 )
                choose_comp_D<true, 1 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else if ( ewald_real == 2 && ewald_recip == 2 )
                choose_comp_D<true, 2 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else if ( ewald_real == 3 && ewald_recip == 3 )
                choose_comp_D<true, 3 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else if ( ewald_real == 4 && ewald_recip == 4 )
                choose_comp_D<true, 4 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else if ( ewald_real == 5 && ewald_recip == 5 )
                choose_comp_D<true, 5 > <<< dim_grid, dim_block >>>(d_coord, _D);
            else
                UNERR( "Wrong ewald values")
        }
        else
        {
            UNERR( "Unimplemented ewald method beenakker")
        }
        choose_comp_D_correction <<< dim_grid, dim_block >>>( d_coord, _D );//
    }
    cudaThreadSynchronize( ); CCERR
    /*mstop();
    extern double g_start;
    if( g_start < mintime )
    {
        mintime = g_start;
        my = ytdim;
        mx = xtdim;
    }
    }
    printf("Winner %g, %d %d\n",mintime, mx,my);*/
}

extern "C"
void init_tensor()
{
    DOUBLE fDii = preDii, fDij = preDij, ewa = ewald_alpha;
    cudaMemcpyToSymbol( dev_preDii, &fDii, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_preDij, &fDij, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_ewald_alpha, &ewa, sizeof(DOUBLE) ); CCERR

    DOUBLE ewa2 = ewa * ewa;
    cudaMemcpyToSymbol( dev_ewald_alpha2, &ewa2, sizeof(DOUBLE) ); CCERR
    DOUBLE ewa3 = ewa * ewa2;
    cudaMemcpyToSymbol( dev_ewald_alpha3, &ewa3, sizeof(DOUBLE) ); CCERR
    DOUBLE M_PI_ALPHA_SQ = -M_PIF * M_PIF / ewa2;
    cudaMemcpyToSymbol( dev_M_PI_ALPHA_SQ, &M_PI_ALPHA_SQ, sizeof(DOUBLE) ); CCERR
    DOUBLE max_sq = (DOUBLE)ewald_real*ewald_real;
    cudaMemcpyToSymbol( dev_max_sq, &max_sq, sizeof(DOUBLE) ); CCERR

    DOUBLE bx[3];
    DOUBLE iL[3];
    bx[0] = box[0], bx[1] = box[1], bx[2] = box[2];
    iL[0] = inv_L[0], iL[1] = inv_L[1], iL[2] = inv_L[2];
    cudaMemcpyToSymbol( dev_box0, bx, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_box1, bx + 1, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_box2, bx + 2, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_invL0, iL, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_invL1, iL + 1, sizeof(DOUBLE) ); CCERR
    cudaMemcpyToSymbol( dev_invL2, iL + 2, sizeof(DOUBLE) ); CCERR
    if( hydro && !( hydro == DIFF_GEYER && algorithm < DIFF_ALG_IGT_CONST ) )
    {
        cudaMemset( d_D, 0, csize * TENSORM * csize * DIMS0 * sizeof (DOUBLE) ); CCERR
    }
    cudaMemcpyToSymbol( dev_size, &size, sizeof(INT) ); CCERR
}

void cm_dh_tensor( DOUBLE* _D )
{
    INT i, j;
    DOUBLE* h_D = (DOUBLE*) malloc( size_D ); CHMEM(h_D)
    cudaMemcpy( h_D, d_D, size_D, cudaMemcpyDeviceToHost ); CCERR
    memset( _D, 0, DIMS0 * size * DIMS0 * size * sizeof (_D[0]) );
    for ( i = 0; i < DIMS0 * size; ++i )
    {
        for ( j = 0; j < DIMS0 * size; ++j )
        {
            if ( i <= j )
            {
                _D[ j * DIMS0 * size + i ] = h_D[ i * TENSORM * csize + j ];
            }
        }
    }
    free( h_D );
}

void compute_single_D( DOUBLE* _D )
{
    compute_D( d_D );
    cm_dh_tensor( _D );
}

template<bool ewald, int ewald_real>
__global__ void rinfly( DOUBLE_4* coords, DOUBLE_3* forces, DOUBLE_3* r )
{
    INT tid = threadIdx.x;
    INT id = blockIdx.x*blockDim.x + tid;
    extern __shared__ DOUBLE sh[];
    DOUBLE_4* sh_coords = (DOUBLE_4*) sh;
    DOUBLE_3* sh_forces = (DOUBLE_3*)( sh_coords + blockDim.x );

    DOUBLE_4 curr_coords = coords[id];
    DOUBLE_3 out_r = {0.0f, 0.0f, 0.0f};

    for ( int i = 0; i < gridDim.x; ++i )
    {
        sh_coords[tid] = coords[ i*blockDim.x + tid ];
        sh_forces[tid] = forces[ i*blockDim.x + tid ];
        __syncthreads();
        for ( int k = 0; k < blockDim.x; ++k )
        {
            INT jd = i * blockDim.x + k;
            if ( jd < dev_size )
            {
                if ( jd == id )//branch divergent on diagonal but only n-times
                {
                    DOUBLE_3 tmp;
                    if ( ewald ) tensor_calc_pos_ewald_diag<ewald_real>( curr_coords, (DOUBLE*) &tmp, 0 );
                    else tensor_calc_pos_diag( curr_coords, (DOUBLE*) &tmp, 0 );
                    out_r.x += tmp.z * sh_forces[k].x;
                    out_r.y += tmp.z * sh_forces[k].y;
                    out_r.z += tmp.z * sh_forces[k].z;
                }
                else
                {
                    DOUBLE t[9];
                    if ( ewald ) tensor_calc_pos_ewald<ewald_real>( curr_coords, sh_coords[k], t, 3 );
                    else tensor_calc_pos( curr_coords, sh_coords[k], t, 3 );
                    DOUBLE_3 f = sh_forces[k];
                    out_r.x += f.x * t[0] + f.y * t[1] + f.z * t[2];
                    out_r.y += f.x * t[3] + f.y * t[4] + f.z * t[5];
                    out_r.z += f.x * t[6] + f.y * t[7] + f.z * t[8];
                }
            }
        }
        __syncthreads();
    }
    r[id] = out_r;
}

void comp_rinfly( DOUBLE* coord, DOUBLE* forces, DOUBLE* r )
{
    INT shMemSize = cuda_block*sizeof(DOUBLE)*7;
    if( !ewald )
    {
        rinfly<false,0><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
    }
    else
    {
        if ( ewald_real == 0 && ewald_recip == 0 )
            rinfly<true,0><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else if ( ewald_real == 1 && ewald_recip == 1 )
            rinfly<true,1><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else if ( ewald_real == 2 && ewald_recip == 2 )
            rinfly<true,2><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else if ( ewald_real == 3 && ewald_recip == 3 )
            rinfly<true,3><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else if ( ewald_real == 4 && ewald_recip == 4 )
            rinfly<true,4><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else if ( ewald_real == 5 && ewald_recip == 5 )
            rinfly<true,5><<<cblocks, cuda_block, shMemSize>>>
                    ( (DOUBLE_4*)coord, (DOUBLE_3*)forces, (DOUBLE_3*)r );
        else
            UNERR( "Wrong ewald values")
    }
}

template<int cuda_block>
__device__ void eps_reduction( DOUBLE e_curr, DOUBLE* out )
{
    extern __shared__ DOUBLE sh[];
    int tid = threadIdx.x;

    __syncthreads( );
    sh[tid] = e_curr;
    __syncthreads( );
    if ( cuda_block >= 512 )
    {
        if ( tid < 256 ) sh[tid] += sh[tid + 256];
        __syncthreads( );
    }
    if ( cuda_block >= 256 )
    {
        if ( tid < 128 ) sh[tid] += sh[tid + 128];
        __syncthreads( );
    }
    if ( cuda_block >= 128 )
    {
        if ( tid < 64 ) sh[tid] += sh[tid + 64];
        __syncthreads( );
    }
    volatile DOUBLE *redE = sh;
    if ( tid < 32 )
    {
        if ( cuda_block >= 64 ) redE[tid] += redE[tid + 32];
        if ( cuda_block > 32 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
        else if ( tid < 16 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
    }
    if ( tid == 0 )
    {
        out[blockIdx.x] = redE[0];
    }
}

__device__ DOUBLE_3 add( DOUBLE_3 a, DOUBLE_3 b) { a.x += b.x; a.y += b.y; a.z += b.z; return a; }

template<int cuda_block, bool ewald, int ewald_real>
__global__ void comp_geyer_beg_in_fly_kernel( DOUBLE_4* coords, DOUBLE_3* ptr_x,
                                              DOUBLE_3* ptr_eff, DOUBLE_3* ptr_c,
                                              DOUBLE* ptr_eps )
{
    DOUBLE eps_curr = 0;
    DOUBLE_3 Feff = {0,0,0};
    DOUBLE_3 C = {0,0,0};
    int tid = threadIdx.x;
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    extern __shared__ DOUBLE sh[];
    DOUBLE_4* sh_coords = (DOUBLE_4*)sh;
    DOUBLE_3* sh_X = (DOUBLE_3*)( sh + 4*blockDim.x );
    DOUBLE_4 curr_coords = coords[ id ];
    for ( int k = 0; k < gridDim.x; ++k )
    {
        sh_coords[tid] = coords[k*blockDim.x+tid];
        sh_X[tid] = ptr_x[k*blockDim.x+tid];
        __syncthreads();
        for ( int j = 0; j < blockDim.x; ++j )
        {
            int jd = k*blockDim.x + j;
            if( jd < dev_size && jd != id )
            {
                DOUBLE_4 k_coord = sh_coords[j];
                DOUBLE_3 tmp;
                if ( ewald ) tensor_calc_pos_ewald_diag<ewald_real>( k_coord, (DOUBLE*) &tmp, 0 );
                else tensor_calc_pos_diag( k_coord, (DOUBLE*) &tmp, 0 );
                DOUBLE dk = tmp.z;
                DOUBLE t[9]; 
                if ( ewald ) tensor_calc_pos_ewald<ewald_real>( curr_coords, k_coord, t, 3 );
                else tensor_calc_pos( curr_coords, k_coord, t, 3 );
                DOUBLE_3 X = sh_X[j];
                Feff.x += X.x*t[0] + X.y*t[3] + X.z*t[6];
                Feff.y += X.x*t[1] + X.y*t[4] + X.z*t[7];
                Feff.z += X.x*t[2] + X.y*t[5] + X.z*t[8];
                //Feff += ptr_x[k] * dik;
                C.x += (t[0]*t[0] + t[3]*t[3] + t[6]*t[6])/dk;
                C.y += (t[1]*t[1] + t[4]*t[4] + t[7]*t[7])/dk;
                C.z += (t[2]*t[2] + t[5]*t[5] + t[8]*t[8])/dk;
                //C += dik*dik/dk;
                eps_curr += t[0]+t[1]+t[2]+t[3]+t[4]+t[5]+t[6]+t[7]+t[8];
            }
        }
        __syncthreads();
    }
    if ( id >= dev_size )
    {
        Feff.x = Feff.y =Feff.z = 0;
        C.x = C.y = C.z = 0;
        eps_curr = 0;
    }
    ptr_eff[id] = Feff;
    ptr_c[id] = C;
    eps_reduction<cuda_block>( eps_curr/2, ptr_eps );
}

template<int cuda_block, bool ewald, int ewald_real>
__global__ void comp_geyer_end_in_fly( DOUBLE_4* coords, DOUBLE* ptr_x, DOUBLE* ptr_eff, DOUBLE* ptr_c, DOUBLE beta, DOUBLE* out)
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    DOUBLE Feff = ptr_eff[id];
    DOUBLE C = ptr_c[id];

    DOUBLE dii;
    DOUBLE_3 tmp;
    if ( ewald ) tensor_calc_pos_ewald_diag<ewald_real>( coords[id/3], (DOUBLE*) &tmp, 0 );
    else tensor_calc_pos_diag( coords[id/3], (DOUBLE*) &tmp, 0 );
    dii = tmp.z;
    C *= beta*beta/dii;
    C = rsqrtf( 1 + C );
    out[id] += C*(beta*Feff + dii*ptr_x[id]);
}

template<int cuda_block>
void comp_geyer_beg_in_fly_ewald( DOUBLE* coord, DOUBLE* x, DOUBLE* eff, DOUBLE* c, DOUBLE* eps )
{
    if ( ewald_real == ewald_recip )
    {
        INT shSize = sizeof(DOUBLE)*7*cuda_block;
        switch( ewald_real )
        {
            case 0: comp_geyer_beg_in_fly_kernel<cuda_block,true,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
            case 1: comp_geyer_beg_in_fly_kernel<cuda_block,true,1><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
            case 2: comp_geyer_beg_in_fly_kernel<cuda_block,true,2><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
            case 3: comp_geyer_beg_in_fly_kernel<cuda_block,true,3><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
            case 4: comp_geyer_beg_in_fly_kernel<cuda_block,true,4><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
            case 5: comp_geyer_beg_in_fly_kernel<cuda_block,true,5><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );return;
        }
    }
    UNERR("Unimplemented")
}

void comp_geyer_beg_in_fly( DOUBLE* coord, DOUBLE* x, DOUBLE* eff, DOUBLE* c, DOUBLE* eps )
{
    if( !ewald )
    {
        INT shSize = sizeof(DOUBLE)*7*cuda_block;
        switch(cuda_block)
        {
            case(1024):comp_geyer_beg_in_fly_kernel<1024,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            case( 512):comp_geyer_beg_in_fly_kernel< 512,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            case( 256):comp_geyer_beg_in_fly_kernel< 256,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            case( 128):comp_geyer_beg_in_fly_kernel< 128,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            case(  64):comp_geyer_beg_in_fly_kernel<  64,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            case(  32):comp_geyer_beg_in_fly_kernel<  32,false,0><<<cblocks,cuda_block,shSize>>>( (DOUBLE_4*)coord, (DOUBLE_3*)x, (DOUBLE_3*)eff, (DOUBLE_3*)c, eps );break;
            default: UNERR("Unimplemented")
        }
    }
    else
    {
        switch(cuda_block)
        {
            case(1024): comp_geyer_beg_in_fly_ewald<1024>( coord, x, eff, c, eps );break;
            case( 512): comp_geyer_beg_in_fly_ewald< 512>( coord, x, eff, c, eps );break;
            case( 256): comp_geyer_beg_in_fly_ewald< 256>( coord, x, eff, c, eps );break;
            case( 128): comp_geyer_beg_in_fly_ewald< 128>( coord, x, eff, c, eps );break;
            case(  64): comp_geyer_beg_in_fly_ewald<  64>( coord, x, eff, c, eps );break;
            case(  32): comp_geyer_beg_in_fly_ewald<  32>( coord, x, eff, c, eps );break;
            default: UNERR("Unimplemented")
        }
    }
}

template<int cuda_block>
void comp_geyer_end_in_fly_ewald( DOUBLE* coord, DOUBLE* ptr_x, DOUBLE* ptr_eff,
                                       DOUBLE* ptr_c, DOUBLE beta, DOUBLE* out)
{
    if ( ewald_real == ewald_recip )
    {
        switch( ewald_real )
        {
            case 0: comp_geyer_end_in_fly<cuda_block,true,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case 1: comp_geyer_end_in_fly<cuda_block,true,1><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case 2: comp_geyer_end_in_fly<cuda_block,true,2><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case 3: comp_geyer_end_in_fly<cuda_block,true,3><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case 4: comp_geyer_end_in_fly<cuda_block,true,4><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case 5: comp_geyer_end_in_fly<cuda_block,true,5><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
        }
    }
    UNERR("Unimplemented ewald")
}

void comp_geyer_end_in_fly( DOUBLE* coord, DOUBLE* ptr_x, DOUBLE* ptr_eff,
                                       DOUBLE* ptr_c, DOUBLE beta, DOUBLE* out)
{
    if( !ewald )
    {
        switch(cuda_block)
        {
            case(1024):comp_geyer_end_in_fly<1024,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 512):comp_geyer_end_in_fly< 512,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 256):comp_geyer_end_in_fly< 256,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 128):comp_geyer_end_in_fly< 128,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case(  64):comp_geyer_end_in_fly<  64,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case(  32):comp_geyer_end_in_fly<  32,false,0><<< DIMS0*cblocks, cuda_block >>>( (DOUBLE_4*)coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
        }
    }
    else
    {
        switch(cuda_block)
        {
            case(1024):comp_geyer_end_in_fly_ewald<1024>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 512):comp_geyer_end_in_fly_ewald< 512>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 256):comp_geyer_end_in_fly_ewald< 256>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case( 128):comp_geyer_end_in_fly_ewald< 128>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case(  64):comp_geyer_end_in_fly_ewald<  64>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
            case(  32):comp_geyer_end_in_fly_ewald<  32>( coord, ptr_x, ptr_eff, ptr_c, beta, out );return;
        }
    }
    UNERR("Unimplemented in fly")
}
