/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include "diff_tensor.h"

#include <float.h>
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "newton.h"

#include "../input.h"
#include "../main_loop.h"
#include "../math_help.h"
#include "../trans.h"

#include "../potentials/calc_func.h"
#include "../potentials/LJ.h"

#include "../../common/err.h"
#include "../../common/rand_move.h"
#include "../../common/myblas.h"

#if USE_MPI
#include "mpi.h"
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

double g_start = -1;

void mstart( )
{
#ifdef _OPENMP
    g_start = omp_get_wtime();
#else
    g_start = 1.*clock()/CLOCKS_PER_SEC;
#endif
}

double mstop( )
{
#ifdef _OPENMP
    return omp_get_wtime() - g_start;
#else
    return (1.*clock()/CLOCKS_PER_SEC) - g_start;
#endif
}

INT coll_count;

DOUBLE comp_step_IGT_const( DOUBLE* _E )
{
    curr_dt = dt;
    curr_iter++;
    coll_count = 0;
    for(;;)
    {
        INT res = compute_hydro_IGT( _E );
        if ( res )
        {
            if ( res == 2 )/* if first bas been accepted, revoke only second */
            {
                load_coord();
                while ( compute_hydro_IGT_second_phase( _E ) )
                {
                    load_coord();
                    ++coll_count;
                    if ( coll_count > move_attempts )
                        UNERR( "Wrong configuration, can't move")
                }
                break;
            }
            else
            {
                /* revoke predictor step, only forces */
                load_forces();
            }
        }
        else
        {
            break;
        }
        ++coll_count;
        if ( coll_count > move_attempts ) UNERR( "Wrong configuration, can't move")
    }
    return curr_dt;
}

DOUBLE comp_step_IGT_var( DOUBLE* _E )
{
    curr_dt = dt;
    curr_iter++;
    coll_count = 0;
    for(;;)
    {
        INT res = compute_hydro_IGT( _E );
        if ( res )
        {
            if ( coll_count < 16 )
            {
                curr_dt /= 2;
            }
            load_coord();
            load_forces();
            if ( res == 2 )/* revoke both steps */
            {
                save_iter--;
                curr_iter--;
            }
        }
        else
        {
            break;
        }
        ++coll_count;
        if ( coll_count > move_attempts ) UNERR( "Wrong configuration, can't move")
    }
    return curr_dt;
}

DOUBLE comp_step_rev_IGT( DOUBLE* _E )
{
    UNERR("Unimplemented method: comp_step_rev_IGT")
    return _E[0];
}

DOUBLE comp_step_ermak( DOUBLE* _E )
{
    curr_dt = dt;
    curr_iter++;
    coll_count = 0;
    for(;;)
    {
        INT ret;
        compute_hydro_ermak( coord );
        ret = compute_forces( _E, curr_time + curr_dt );
        if ( ret )
        {
            if ( algorithm == DIFF_ALG_ERMAK_VAR && coll_count < 16 )
            {
                curr_dt /= 2;
            }
            load_coord();
            load_forces();
        }
        else
        {
            break;
        }
        ++coll_count;
        if ( coll_count > move_attempts ) UNERR("Wrong configuration, can't move")
    }
    return curr_dt;
}

DOUBLE comp_step_ermak_newton( DOUBLE* _E )
{
    INT ret = 1;

    coll_count = 0;
    curr_iter++;
    while ( ret )
    {
        curr_dt = dt;
        compute_hydro_ermak( coord );
        ret = compute_forces( _E, curr_time + curr_dt );
        if ( ret )
        {
            ret = 0;/* clear ret */
            newton_init_vel();
            load_coord();
            ret = newton_moves( &coll_count );
            ret = ret || compute_forces( _E, curr_time + curr_dt );
            if( ret )
            {
                load_coord();
                load_forces();
                if( bc != BC_SPHERE ) WARN("Revert move in no bounding sphere case");
            }
        }
    }
    return dt;
}

void compute_hydro( DOUBLE* _E )
{
    switch ( algorithm )
    {
        case DIFF_ALG_ERMAK_CONST:
        case DIFF_ALG_ERMAK_VAR:
            curr_time += comp_step_ermak( _E );
            break;
        case DIFF_ALG_ERMAK_NEWTON:
            curr_time += comp_step_ermak_newton( _E );
            break;
        case DIFF_ALG_IGT_CONST:
            curr_time += comp_step_IGT_const( _E );
            break;
        case DIFF_ALG_IGT_VAR:
            curr_time += comp_step_IGT_var( _E );
            break;
        case DIFF_ALG_IGT_VAR_REV:
            curr_time += comp_step_rev_IGT( _E );
            break;
        default: UNERR( "Unimplemented algorithm")
    }
}
