/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include "bucketf.h"

#include <float.h>
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

#include "calc_func.h"
#include "electro.h"
#include "LJ.h"

#include "../input.h"
#include "../math_help.h"
#include "../trans.h"

#include "../../common/err.h"
#include "../../common/math_common.h"

MAKE_STR_IN(BUCKET_NONE_SPATIAL,nb_list,0,"algorithm for generation of nonbonded interactions lists")
DOUBLE max_force_cutoff;

void get_h_l_f( INT size, DOUBLE4* coords, DOUBLE* radii, DOUBLE blen, DOUBLE h[3], DOUBLE l[3] )
{
    INT i, j;
    for(j=0;j<3;++j) h[j] = -FLT_MAX;
    for(j=0;j<3;++j) l[j] = FLT_MAX;
    for( i = 0; i < size; ++i )
    {
        for(j=0;j<3;++j)
        {
            if( coords[i].t[j] > h[j] )
            {
                h[j]=coords[i].t[j];
            }
            if( coords[i].t[j] < l[j] )
            {
                l[j]=coords[i].t[j];
            }
        }
        if( radii && 2*radii[2*i] > blen ) blen = 2*radii[2*i] + 0.01f;
    }
}

INT correct_indx_f( int indx, int last )
{
    if( indx < 0 ) return last + indx;
    return (indx)%last;
}

void get_neighbours( bucket_f* br, DOUBLE4* bead, int sizes[64], int* ids[64] )
{
    int count = 0;
    int i,j,k,m;
    int bp[3];
    int p[3];
    int blocal[3];
    int local[3];
    list_reg* plr;
    int li, lj, lk, hi, hj, hk;
    int to_corr = br->mlocal[0] >= 0;

    get_bid_f( br, bead, bp, blocal );
    li = lj = lk = -1;
    hi = hj = hk = 1;
    if ( to_corr )
    {
        if ( br->mlocal[0] > 2 && blocal[0] == 0 ) li = -2;
        if ( br->mlocal[1] > 2 && blocal[1] == 0 ) lj = -2;
        if ( br->mlocal[2] > 2 && blocal[2] == 0 ) lk = -2;
        
        if ( br->mlocal[0] > 2 && blocal[0] == br->mlocal[0]-1 ) hi = 2;
        if ( br->mlocal[1] > 2 && blocal[1] == br->mlocal[1]-1 ) hj = 2;
        if ( br->mlocal[2] > 2 && blocal[2] == br->mlocal[2]-1 ) hk = 2;
        if ( br->mlocal[0] == 0 ) li = hi = 0;
        if ( br->mlocal[1] == 0 ) lj = hj = 0;
        if ( br->mlocal[2] == 0 ) lk = hk = 0;

        if ( br->mlocal[0] == 1 ) li = 0;
        if ( br->mlocal[1] == 1 ) lj = 0;
        if ( br->mlocal[2] == 1 ) lk = 0;
    }
    
    for( i = li; i <= hi; ++i)
    {
        for( j = lj; j <= hj; ++j)
        {
            for( k = lk; k <= hk; ++k)
            {
                if( to_corr )
                {
                    local[0] = correct_indx_f(blocal[0]+i,br->mlocal[0]+1);
                    local[1] = correct_indx_f(blocal[1]+j,br->mlocal[1]+1);
                    local[2] = correct_indx_f(blocal[2]+k,br->mlocal[2]+1);
                    p[0] = correct_indx_f( local[0], MIN(BUCKET_N,br->mlocal[0]+1) );
                    p[1] = correct_indx_f( local[1], MIN(BUCKET_N,br->mlocal[1]+1) );
                    p[2] = correct_indx_f( local[2], MIN(BUCKET_N,br->mlocal[2]+1) );
                }
                else
                {
                    p[0] = correct_indx_f( bp[0]+i, BUCKET_N );
                    p[1] = correct_indx_f( bp[1]+j, BUCKET_N );
                    p[2] = correct_indx_f( bp[2]+k, BUCKET_N );
                    local[0] = blocal[0]+i;
                    local[1] = blocal[1]+j;
                    local[2] = blocal[2]+k;
                }            
                if( local[0] >= 0 && local[1] >= 0 && local[2] >= 0 )
                {
                    plr = &br->lreg[p[0]][p[1]][p[2]];
                    sizes[count]=0;
                    ids[count]=NULL;
                    for( m = 0; m < plr->size; ++m )
                    {
                        if( plr->regs[m].x == local[0] &&
                            plr->regs[m].y == local[1] &&
                            plr->regs[m].z == local[2])
                        {
                            sizes[count] = plr->regs[m].size;
                            ids[count] = plr->regs[m].ids;
                            count ++;
                            break;
                        }
                    }
                }
            }
        }
    }
    if( count < 64 ) sizes[count] = 0;
}

bucket_f* init_bucket_f( int size, DOUBLE4* coords, DOUBLE* radii, DOUBLE blen )
{
    int i,j;
    DOUBLE h[3], l[3];
    int count = 0;
    bucket_f* br = (bucket_f*) malloc(sizeof(bucket_f)); CHMEM(br)
    memset( br, 0, sizeof(bucket_f) );

    get_h_l_f( size, coords, radii, blen, h, l );

    for(j=0;j<3;++j)
    {
        count += (INT)((h[j]-l[j])/blen);
    }
    if( count <= 6 )/*no sense*/
    {
        free( br );
        return NULL;
    }
    for(j=0;j<3;++j) br->offset[j] = -l[j];
    br->blen[0] = br->blen[1] = br->blen[2] = blen;
    br->mlocal[0] = br->mlocal[1] = br->mlocal[2] = -1;
    if ( bc == BC_BOX )
    {
        for ( i = 0; i < size; ++i )
        {
            for ( j = 0; j < DIMS0; ++j )
            {
                INT tmp = (INT)((coords[i].t[j]+br->offset[j])/br->blen[j]);
                if ( tmp > br->mlocal[j] ) br->mlocal[j] = tmp;
            }
        }
    }
    for ( i = 0; i < size; ++i )
    {
        add_coord_f( br, i, coords+i);
    }
    return br;
}

void get_bid_f( bucket_f* br, DOUBLE4* bead, int* p, int* local )
{
    int j;
    for(j=0;j<3;++j)
    {
        local[j] = (INT)((bead->t[j]+br->offset[j])/br->blen[j]);
        p[j] = local[j] % BUCKET_N;
    }
}

void add_coord_f( bucket_f* br, int id, DOUBLE4* pbead )
{
    INT p[3], j, local[3];
    list_reg* plr;
    get_bid_f( br, pbead, p, local);
    plr = &br->lreg[p[0]][p[1]][p[2]];
    if( plr->size )
    {
        for( j = 0; j < plr->size; ++j )
        {
            if( plr->regs[j].x == local[0] && plr->regs[j].y == local[1] && plr->regs[j].z == local[2] )
                break;
        }
        if( j < plr->size )
        {
            if( plr->regs[j].size + 1 < plr->regs[j].capacity )
            {
                plr->regs[j].ids[plr->regs[j].size++]=id;
            }
            else
            {
                plr->regs[j].capacity = plr->regs[j].capacity*BUCKET_CAP_MLT + BUCKET_CAP_ADD;
                plr->regs[j].ids = (int*)realloc( plr->regs[j].ids, sizeof(int)*plr->regs[j].capacity );
                plr->regs[j].ids[plr->regs[j].size++]=id;
            }
            j = 0;
        }
        else
        {
            j = 1;
        }
    }
    else
    {
        j = 1;
    }
    if( j )
    {
        plr->size++;
        plr->regs = (reg_bucket*)realloc( plr->regs, sizeof(reg_bucket)*plr->size );
        plr->regs[plr->size-1].x = local[0];
        plr->regs[plr->size-1].y = local[1];
        plr->regs[plr->size-1].z = local[2];
        plr->regs[plr->size-1].size=1;
        plr->regs[plr->size-1].capacity=2;
        plr->regs[plr->size-1].ids= (int*) malloc(sizeof(int)*2);
        plr->regs[plr->size-1].ids[0] = id;
    }
}

void free_bucket_f( bucket_f* br, int size, DOUBLE4* coord )
{
    int i;
    if( size > 0 && coord )
    {
        int p[3], local[3];
        for( i = 0; i < size; ++i )
        {
            list_reg* plr;
            get_bid_f(br,coord+i,p,local);
            plr = &br->lreg[p[0]][p[1]][p[2]];
            if( plr->size )
            {
                int j;
                for( j = 0; j < plr->size; ++j )
                {
                    free(plr->regs[j].ids);
                }
                free( plr->regs );
                plr->size = 0;
            }
        }
    }
    else
    {
        int j, k, l;
        for( i = 0; i < BUCKET_N; ++i )
        {
            for( j = 0; j < BUCKET_N; ++j )
            {
                for( k = 0; k < BUCKET_N; ++k )
                {
                    list_reg* plr;
                    plr = &br->lreg[i][j][k];
                    if( plr->size )
                    {
                        for( l = 0; l < plr->size; ++l )
                        {
                            free(plr->regs[l].ids);
                        }
                        free( plr->regs );
                        plr->size = 0;
                    }
                }
            }
        }
    }
    free(br);
}

void print_statistic_f( bucket_f* br )
{
    int i, j, k, l;
    int count = 0;
    int max_reg = 0;
    int max_ids = 0;
    for( i = 0; i < BUCKET_N; ++i )
    {
        for( j = 0; j < BUCKET_N; ++j )
        {
            for( k = 0; k < BUCKET_N; ++k )
            {
                list_reg* plr;
                plr = &br->lreg[i][j][k];
                if( plr->size )
                {
                    for( l = 0; l < plr->size; ++l )
                    {
                        count += plr->regs[l].size;
                        if( plr->regs[l].size > max_ids )
                            max_ids = plr->regs[l].size;
                    }
                    if( plr->size > max_reg )
                        max_reg = plr->size;
                }
            }
        }
    }
    printf("C: %d, mreg: %d, mi: %d\n",count,max_reg,max_ids);
}
