/**************************************************************************
 *                                                                        *
 *   Copyright (C) 2010 - 2012: University of Warsaw                      *
 *   Authors:            Pawel Zielinski <pzielins@icm.edu.pl>            *
 *                       Maciej Dlugosz  <mdlugosz@icm.edu.pl>            *
 *                                                                        *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 3 of the License, or    *
 *   (at your option) any later version.                                  *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, see http://www.gnu.org/licenses     *
 *   or write to the Free Software Foundation,Inc., 51 Franklin Street,   *
 *   Fifth Floor, Boston, MA 02110-1301  USA                              *
 *                                                                        *
 **************************************************************************/

#include <cuda_runtime_api.h>
#include <cublas.h>

#include <cfloat>
#include <cstdio>
#include <cstdlib>
#include <memory.h>

#include "../calc_func.h"

#include "../angle.h"
#include "../angle_cos.h"
#include "../bond.h"
#include "../bucketf.h"
#include "../dihe.h"
#include "../dihe_angle.h"
#include "../electro.h"
#include "../electro_ext.h"
#include "../LJ.h"

#include "../../cuda.h"
#include "../../input.h"
#include "../../math_help.h"
#include "../../trans.h"

#include "../../netdecl.cuh"

#include "../../../common/err.h"
#include "../../../common/bucket.h"
#include "../../../common/math_common.h"
#include "../../../common/cuda/sorter.cuh"

#if USE_MPI
#include <mpi.h>
#else
#ifdef _OPENMP
#include <omp.h>
#endif
#endif

using namespace std;

#define EPS 0.001f
#define MAX_MULT 28
#define PATTERN_BITS 4
#define PATTERN_MASK 15
#define PATTERN_PH 8
#define PH_COUNT 8

__constant__ DOUBLE dev_offset[3];
__constant__ DOUBLE dev_blen[3];
__constant__ DOUBLE dev_kappa;
__constant__ DOUBLE dev_gamma_eps;
__constant__ DOUBLE dev_cutoffc;
__constant__ DOUBLE dev_cutoffLJ;
__constant__ DOUBLE dev_alpha_lj;
__constant__ DOUBLE dev_lj_6_term;
__constant__ DOUBLE dev_box0;
__constant__ DOUBLE dev_box1;
__constant__ DOUBLE dev_box2;
__constant__ DOUBLE dev_invL0;
__constant__ DOUBLE dev_invL1;
__constant__ DOUBLE dev_invL2;
__constant__ DOUBLE dev_bond_lj_scale;
__constant__ DOUBLE dev_bond_c_scale;
__constant__ DOUBLE dev_max_cutoff;
__constant__ INT* dev_gconns;
__constant__ INT dev_size;
__constant__ INT dev_bonds;

__device__ UINT d_keys_size;

texture< DOUBLE_4, 1, cudaReadModeElementType> tr_coords;
texture< DOUBLE, 1, cudaReadModeElementType> tr_q;
texture< DOUBLE_2, 1, cudaReadModeElementType> tr_lj;

INT max_gs = 0;
INT* graph = NULL;
INT* d_gconns = NULL;

INT* pkkeys = NULL;
INT* pkvals = NULL;
UINT* pkpos = NULL;

DOUBLE* d_out = NULL;
DOUBLE* d_bforces = NULL;
DOUBLE* d_help = NULL;
INT* d_indxs = NULL;
DOUBLE* d_ljcoord;

bool forceBrute;
INT forces_size;

StaticSorter* sorter = NULL;

DOUBLE* E;

MAKE_STR_IN( DOUBLE, bond_lj_scale, 1.0, "scaling factor for Lennard–Jones interactions between bonded partners 1–2" )
MAKE_STR_IN( DOUBLE, bond_c_scale, 1.0, "scaling factor for electrostatic interactions between bonded partners 1–2" )
MAKE_STR_IN( YESNO, check_overlap, 1, "checking overlap" )

INT* d_overlap = NULL;
DOUBLE* d_lj = NULL;
DOUBLE* d_Q = NULL;

void mstart( );
double mstop( );

template<bool pbc>
__device__ DOUBLE_4 get_dist( DOUBLE_4 ci, DOUBLE_4 cj )
{
    DOUBLE_4 r_ij;
    r_ij.x = ci.x - cj.x;
    r_ij.y = ci.y - cj.y;
    r_ij.z = ci.z - cj.z;

    if ( pbc )
    {
        r_ij.x -= dev_box0 * FLOORD( r_ij.x * dev_invL0 + 0.5f );
        r_ij.y -= dev_box1 * FLOORD( r_ij.y * dev_invL1 + 0.5f );
        r_ij.z -= dev_box2 * FLOORD( r_ij.z * dev_invL2 + 0.5f );
    }
    return r_ij;
}

template<bool pbc>
__device__ INT isoverlap( DOUBLE_4 ci, DOUBLE_4 cj )
{
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x*r_ij.x + r_ij.y*r_ij.y + r_ij.z*r_ij.z;

    DOUBLE sigma = (ci.w + cj.w)*(ci.w + cj.w);

    return r_sq < sigma;
}

template<bool pbc>
__device__ INT isoverlap( DOUBLE_4 ci, DOUBLE x, DOUBLE y, DOUBLE z, DOUBLE w )
{
    DOUBLE_4 cj = {x,y,z,w};
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x*r_ij.x + r_ij.y*r_ij.y + r_ij.z*r_ij.z;

    DOUBLE sigma = (ci.w + cj.w)*(ci.w + cj.w);

    return r_sq < sigma;
}

template<bool pbc>
__global__ void check_beads_overlap( void* ptr, void* ptr_lj, INT* overlap )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    extern __shared__ DOUBLE_4 shPosition[];
    DOUBLE_4* tab = (DOUBLE_4*) ptr;
    DOUBLE_4 curr = tab[id];
    DOUBLE_2* ljtab = (DOUBLE_2*) ptr_lj;
    curr.w = ljtab[id].x;
    char rep = 0;
    for ( INT i = 0; i < blockIdx.x; ++i )
    {
        shPosition[ threadIdx.x ] = tab[ i * blockDim.x + threadIdx.x ];
        DOUBLE_2 lw = ljtab[i * blockDim.x + threadIdx.x];
        shPosition[ threadIdx.x ].w = lw.x;
        __syncthreads( );
        for ( INT j = 0; j < blockDim.x; ++j )
        {
            rep = rep | isoverlap<pbc > (curr, shPosition[j]);
        }
        __syncthreads( );
    }
    shPosition[ threadIdx.x ] = curr;
    __syncthreads( );
    for ( INT j = 0; j < threadIdx.x; ++j )
    {
        rep = rep | isoverlap<pbc > (curr, shPosition[j]);
    }
    if( rep ) overlap[ 0 ] = 1;
}

#ifdef USE_FLOAT
template<bool pbc, INT uncount>
__global__ void check_beads_overlap_tex( INT* overlap )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    extern __shared__ DOUBLE_4 shPosition[];
    DOUBLE_4 curr = tex1Dfetch( tr_coords, id );
    INT rep = 0;
    for ( INT i = 0; i < blockIdx.x; ++i )
    {
        shPosition[ threadIdx.x ] = tex1Dfetch( tr_coords, i * blockDim.x + threadIdx.x );
        __syncthreads( );
#pragma unroll
        for ( INT j = 0; j < uncount; ++j )
        {
            rep = rep | isoverlap<pbc > (curr, shPosition[j]);
        }
        __syncthreads( );
    }
    shPosition[ threadIdx.x ] = curr;
    __syncthreads( );
    for ( INT j = 0; j < threadIdx.x; ++j )
    {
        rep = rep | isoverlap<pbc > (curr, shPosition[j]);
    }
    if( rep ) overlap[ 0 ] = 1;
}
#endif

__global__ void make_coords( DOUBLE_4* coords, DOUBLE_4* src_coords, DOUBLE_2* src_lj )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x;
    coords[id] = src_coords[id];
    coords[id].w = src_lj[id].x;
}

__device__ void energy_reduction( DOUBLE* out, DOUBLE e_curr )
{
    extern __shared__ DOUBLE sh[];
    int tid = threadIdx.x;
    volatile DOUBLE* redE = sh;

    __syncthreads( );
    redE[tid] = e_curr;
    __syncthreads( );
    if ( blockDim.x >= 1024 )
    {
        if ( tid < 512 ) redE[tid] += redE[tid + 512];
        __syncthreads( );
    }
    if ( blockDim.x >= 512 )
    {
        if ( tid < 256 ) redE[tid] += redE[tid + 256];
        __syncthreads( );
    }
    if ( blockDim.x >= 256 )
    {
        if ( tid < 128 ) redE[tid] += redE[tid + 128];
        __syncthreads( );
    }
    if ( blockDim.x >= 128 )
    {
        if ( tid < 64 ) redE[tid] += redE[tid + 64];
        __syncthreads( );
    }
    if ( tid < 32 )
    {
        if ( blockDim.x >= 64 ) redE[tid] += redE[tid + 32];
        if ( blockDim.x > 32 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
        else if ( tid < 16 )
        {
            redE[tid] += redE[tid + 16];
            redE[tid] += redE[tid + 8];
            redE[tid] += redE[tid + 4];
            redE[tid] += redE[tid + 2];
            redE[tid] += redE[tid + 1];
        }
    }
    if ( tid == 0 )
    {
        *out = redE[0];
    }
}

template<bool checkOverlap>
__device__ DOUBLE_4 electro( DOUBLE_4 r_ij, DOUBLE qj, DOUBLE r_sq, DOUBLE li, DOUBLE lj )
{
    DOUBLE r_norm = sqrtf( r_sq );
    DOUBLE E = qj * expf( -dev_kappa * r_norm ) / r_norm;
    DOUBLE f = E * (1 + dev_kappa * r_norm) / r_sq;

    r_ij.x *= f;
    r_ij.y *= f;
    r_ij.z *= f;
    if ( checkOverlap ) r_ij.w = (li + lj) > r_norm ? nanf( "" ) : E;
    else r_ij.w = E;
    return r_ij;
}

template<bool pbc,bool checkOverlap>
__device__ DOUBLE_4 electro_s( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE li, DOUBLE lj )
{
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    if ( r_sq <= dev_cutoffc )
    {
        return electro<checkOverlap>( r_ij, cj.w, r_sq, li, lj );
    }
    DOUBLE_4 ret = {0,0,0,0};
    return ret;
}

template<INT cg>
__device__ bool gconnected( INT id, INT jd, INT* gconns )
{
    bool conn = 0;
    if ( cg >= 9 )
    {
	INT* gcurr = gconns + id * dev_bonds;
        for ( INT i = 0; i < dev_bonds; ++i )
            conn = conn || (jd == gcurr[i]);
    }
    else
    {
	INT* gcurr = gconns + id * cg;
        if ( cg >= 1 ) conn = conn || (jd == gcurr[0]);
        if ( cg >= 2 ) conn = conn || (jd == gcurr[1]);
        if ( cg >= 3 ) conn = conn || (jd == gcurr[2]);
        if ( cg >= 4 ) conn = conn || (jd == gcurr[3]);
        if ( cg >= 5 ) conn = conn || (jd == gcurr[4]);
        if ( cg >= 6 ) conn = conn || (jd == gcurr[5]);
        if ( cg >= 7 ) conn = conn || (jd == gcurr[6]);
        if ( cg >= 8 ) conn = conn || (jd == gcurr[7]);
    }
    return conn;
}

template<bool pbc, INT cg,bool checkOverlap>
__device__ DOUBLE_4 electro_g( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE li, DOUBLE lj, INT id, INT jd, INT* gconns )
{
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    if ( r_sq <= dev_cutoffc )
    {
        r_ij = electro<checkOverlap>( r_ij, cj.w, r_sq, li, lj );
        bool conn = gconnected<cg>( id, jd, gconns );
        if ( conn )
        {
            r_ij.x *= dev_bond_c_scale;
            r_ij.y *= dev_bond_c_scale;
            r_ij.z *= dev_bond_c_scale;
            r_ij.w *= dev_bond_c_scale;
            if ( dev_bond_c_scale != 0 ) return r_ij;
        }
        else
        {
            return r_ij;
        }
    }
    DOUBLE_4 ret = {0,0,0,0};
    return ret;
}

template<bool pbc, INT cg,bool checkOverlap>
__global__ void comp_electro( DOUBLE_4* tab, DOUBLE* qtab, DOUBLE_2* ljtab, DOUBLE_3* forces )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x; /*id*/
    extern __shared__ DOUBLE sh[];
    DOUBLE_4* shPosition = (DOUBLE_4*)sh;
    DOUBLE* shLJ = sh + 4*blockDim.x;
    DOUBLE_3* ftab = (DOUBLE_3*) forces;
    
    DOUBLE_4 curr = tab[id];
    curr.w = qtab[id];
    DOUBLE currRadiiLJ = ljtab[id].x;
    
    DOUBLE_3 f_curr = {0.0f, 0.0f, 0.0f};
    DOUBLE e_curr = 0.0f;

    for ( INT i = 0; i < gridDim.x; ++i )
    {
        DOUBLE_2 lw = ljtab[i * blockDim.x + threadIdx.x];
        shLJ[ threadIdx.x ] = lw.x;
        shPosition[ threadIdx.x ] = tab[ i * blockDim.x + threadIdx.x ];
        shPosition[ threadIdx.x ].w = qtab[i * blockDim.x + threadIdx.x];
        __syncthreads( );
        for ( INT j = 0; j < blockDim.x; ++j )
        {
            DOUBLE_4 f;
            INT jd = i * blockDim.x + j;
            if ( cg == 0 )
                f = electro_s<pbc,checkOverlap> (curr, shPosition[j],currRadiiLJ,shLJ[j]);
            else
                f = electro_g<pbc, cg, checkOverlap> (curr, shPosition[j],currRadiiLJ,shLJ[j], id, jd, dev_gconns );
            f_curr.x += (id!=jd) ? f.x : 0.0f;
            f_curr.y += (id!=jd) ? f.y : 0.0f;
            f_curr.z += (id!=jd) ? f.z : 0.0f;
            e_curr +=   (id!=jd) ? f.w : 0.0f;
        }
        __syncthreads( );
    }
    f_curr.x *= curr.w * dev_gamma_eps; /* curr.w = Qi*/
    f_curr.y *= curr.w * dev_gamma_eps;
    f_curr.z *= curr.w * dev_gamma_eps;
    e_curr *= curr.w * dev_gamma_eps;
    ftab[id] = f_curr;
    if ( id >= dev_size ) e_curr = 0;
    energy_reduction( (DOUBLE*)forces + DIMS0 * blockDim.x * gridDim.x + blockIdx.x, e_curr );
}

typedef struct
{
    DOUBLE_4 a;
    DOUBLE_4 b;
} DOUBLE_8;

template<bool chOverlap>
__device__ DOUBLE_4 LJ_s( DOUBLE_4 r_ij, DOUBLE r_sq , DOUBLE_2 li, DOUBLE_2 lj, DOUBLE flag )
{
    DOUBLE r_norm = sqrtf( r_sq );
    DOUBLE epsilon_ij = sqrtf( li.y * lj.y );
    DOUBLE sig_r_6 = (li.x + lj.x) / r_norm;

    sig_r_6 *= sig_r_6;
    sig_r_6 *= sig_r_6 * sig_r_6;

    DOUBLE El = epsilon_ij * (sig_r_6 * (sig_r_6 - dev_lj_6_term));
    DOUBLE fl = epsilon_ij * 6 * (sig_r_6 * (2 * sig_r_6 - dev_lj_6_term)) / r_sq;

    r_ij.x = r_ij.x * fl;
    r_ij.y = r_ij.y * fl;
    r_ij.z = r_ij.z * fl;
    if ( chOverlap ) r_ij.w = flag + El;
    else  r_ij.w = El;

    return r_ij;
}

template<bool pbc,bool dyn_cutoff,bool chOverlap>
__device__ DOUBLE_8 electroLJ( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE_2 li, DOUBLE_2 lj )
{
    DOUBLE_8 r_ij;
    r_ij.a = get_dist<pbc>( ci, cj );
    r_ij.b.x = r_ij.b.y = r_ij.b.z = r_ij.b.w = 0;
    
    DOUBLE r_sq = r_ij.a.x * r_ij.a.x + r_ij.a.y * r_ij.a.y + r_ij.a.z * r_ij.a.z;
    
    DOUBLE flag = ((li.x + lj.x)*(li.x + lj.x) > r_sq && chOverlap) ? nanf( "" ) : 0;
    if ( r_sq <= (dyn_cutoff ? (1.259920941444f*(li.x + lj.x)*(li.x + lj.x)) : dev_cutoffLJ) )
    {
        r_ij.b = LJ_s<chOverlap>( r_ij.a, r_sq , li, lj, flag );
    }
    if ( r_sq <= dev_cutoffc )
    {
        r_ij.a = electro<chOverlap>( r_ij.a, cj.w, r_sq, li.x, lj.x );
    }
    else
    {
        r_ij.a.x = r_ij.a.y = r_ij.a.z = r_ij.a.w = 0;
    }
    if (chOverlap) r_ij.a.w += flag;
    return r_ij;
}

template<bool pbc,bool dyn_cutoff,INT cg,bool chOverlap>
__device__ DOUBLE_8 electroLJ_g( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE_2 li, DOUBLE_2 lj, INT id, INT jd, INT* gconns )
{
    DOUBLE_8 r_ij;
    r_ij.a = get_dist<pbc>( ci, cj );
    r_ij.b.x = r_ij.b.y = r_ij.b.z = r_ij.b.w = 0;

    DOUBLE r_sq = r_ij.a.x * r_ij.a.x + r_ij.a.y * r_ij.a.y + r_ij.a.z * r_ij.a.z;

    DOUBLE flag = ((li.x + lj.x)*(li.x + lj.x) > r_sq && chOverlap) ? nanf( "" ) : 0;
    INT conn = -1;
    if ( r_sq <= (dyn_cutoff ? (1.259920941444f*(li.x + lj.x)*(li.x + lj.x)) : dev_cutoffLJ) )
    {
        r_ij.b = LJ_s<chOverlap>( r_ij.a, r_sq , li, lj, flag );
        conn = gconnected<cg>( id, jd, gconns );
        if ( conn )
        {
            r_ij.b.x *= dev_bond_lj_scale;
            r_ij.b.y *= dev_bond_lj_scale;
            r_ij.b.z *= dev_bond_lj_scale;
            r_ij.b.w *= dev_bond_lj_scale;
            if ( dev_bond_lj_scale == 0 )
                r_ij.b.x = r_ij.b.y = r_ij.b.z = r_ij.b.w = 0;
        }
    }
    if ( r_sq <= dev_cutoffc )
    {
        r_ij.a = electro<chOverlap>( r_ij.a, cj.w, r_sq, li.x, lj.x );
        if (conn==-1) conn = gconnected<cg>( id, jd, gconns );
        if ( conn )
        {
            r_ij.a.x *= dev_bond_c_scale;
            r_ij.a.y *= dev_bond_c_scale;
            r_ij.a.z *= dev_bond_c_scale;
            r_ij.a.w *= dev_bond_c_scale;
            if ( dev_bond_c_scale == 0 )
                r_ij.a.x = r_ij.a.y = r_ij.a.z = r_ij.a.w = 0;
        }
    }
    else
    {
        r_ij.a.x = r_ij.a.y = r_ij.a.z = r_ij.a.w = 0;
    }
    if (chOverlap) r_ij.a.w += flag;
    return r_ij;
}

template<bool pbc, bool dyn_cutoff, INT cg, bool chOverlap>
__global__ void comp_electroLJ( DOUBLE_4* tab, DOUBLE* qtab, DOUBLE_2* ljtab, DOUBLE_3* forces )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x; /*id*/
    extern __shared__ DOUBLE sh[];
    DOUBLE_4* shPosition = (DOUBLE_4*)sh;
    DOUBLE_2* shLJ = (DOUBLE_2*)(sh + 4*blockDim.x);
    DOUBLE_3* ftab = (DOUBLE_3*) forces;

    DOUBLE_4 curr = tab[id];
    curr.w = qtab[id];
    DOUBLE_2 currLJ = ljtab[id];

    DOUBLE_8 f_curr = {0,0,0,0,0,0,0,0};
    
    for ( INT i = 0; i < gridDim.x; ++i )
    {
        shLJ[ threadIdx.x ] = ljtab[ i * blockDim.x + threadIdx.x];
        shPosition[ threadIdx.x ] = tab[ i * blockDim.x + threadIdx.x ];
        shPosition[ threadIdx.x ].w = qtab[ i * blockDim.x + threadIdx.x];
        __syncthreads( );
        for ( INT j = 0; j < blockDim.x; ++j )
        {
            DOUBLE_8 f;
            INT jd = i * blockDim.x + j;
            if ( cg == 0 ) f = electroLJ<pbc,dyn_cutoff,chOverlap> (curr, shPosition[j],currLJ,shLJ[j]);
            else f = electroLJ_g<pbc,dyn_cutoff,cg,chOverlap> (curr, shPosition[j],currLJ,shLJ[j],id,jd,dev_gconns);
            f_curr.a.x += (id!=jd) ? f.a.x : 0.0f;
            f_curr.a.y += (id!=jd) ? f.a.y : 0.0f;
            f_curr.a.z += (id!=jd) ? f.a.z : 0.0f;
            f_curr.a.w += (id!=jd) ? f.a.w : 0.0f;

            f_curr.b.x += (id!=jd) ? f.b.x : 0.0f;
            f_curr.b.y += (id!=jd) ? f.b.y : 0.0f;
            f_curr.b.z += (id!=jd) ? f.b.z : 0.0f;
            f_curr.b.w += (id!=jd) ? f.b.w : 0.0f;
        }
        __syncthreads( );
    }
    f_curr.a.x *= curr.w * dev_gamma_eps; /* curr.w = Qi*/
    f_curr.a.y *= curr.w * dev_gamma_eps;
    f_curr.a.z *= curr.w * dev_gamma_eps;
    f_curr.a.w *= curr.w * dev_gamma_eps;

    f_curr.b.x *= dev_alpha_lj;
    f_curr.b.y *= dev_alpha_lj;
    f_curr.b.z *= dev_alpha_lj;
    f_curr.b.w *= dev_alpha_lj;

    DOUBLE e_cu = f_curr.a.w;
    ftab[id].x = f_curr.a.x + f_curr.b.x;
    ftab[id].y = f_curr.a.y + f_curr.b.y;
    ftab[id].z = f_curr.a.z + f_curr.b.z;
    
    if ( id >= dev_size ) e_cu = 0;
    energy_reduction( (DOUBLE*)forces + 3 * blockDim.x * gridDim.x + blockIdx.x, e_cu );
    __syncthreads();
    e_cu = f_curr.b.w;
    if ( id >= dev_size ) e_cu = 0;
    energy_reduction( (DOUBLE*)forces + 3 * blockDim.x * gridDim.x + blockIdx.x + gridDim.x, e_cu );
}

template<bool pbc, bool dyn_cutoff,bool chOverlap>
__device__ DOUBLE_4 sLJ( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE eps1, DOUBLE eps2 )
{
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    DOUBLE flag = ((ci.w + cj.w)*(ci.w + cj.w) > r_sq && chOverlap) ? nanf( "" ) : 0;
    
    if ( r_sq <= (dyn_cutoff ? (1.259920941444f * (ci.w + cj.w)*(ci.w + cj.w)) : dev_cutoffLJ ) )
    {
        DOUBLE_2 li = {ci.w, eps1};
        DOUBLE_2 lj = {cj.w, eps2};
        return LJ_s<chOverlap>( r_ij, r_sq , li, lj, flag );
    }
    DOUBLE_4 ret = {0.0f, 0.0f, 0.0f, flag};
    return ret;
}

template<bool pbc, bool dyn_cutoff, bool chOverlap, INT cg>
__device__ DOUBLE_4 sLJ_g( DOUBLE_4 ci, DOUBLE_4 cj, DOUBLE eps1, DOUBLE eps2, INT id, INT jd, INT* gconns )
{
    DOUBLE_4 r_ij = get_dist<pbc>( ci, cj );

    DOUBLE r_sq = r_ij.x * r_ij.x + r_ij.y * r_ij.y + r_ij.z * r_ij.z;
    DOUBLE flag = ((ci.w + cj.w)*(ci.w + cj.w) > r_sq && chOverlap) ? nanf( "" ) : 0;

    if ( r_sq <= (dyn_cutoff ? (1.259920941444f * (ci.w + cj.w)*(ci.w + cj.w)) : dev_cutoffLJ ) )
    {
        DOUBLE_2 li = {ci.w, eps1};
        DOUBLE_2 lj = {cj.w, eps2};
        r_ij = LJ_s<chOverlap>( r_ij, r_sq , li, lj, flag );
        bool conn = gconnected<cg>( id, jd, gconns );
        if ( conn )
        {
            r_ij.x *= dev_bond_lj_scale;
            r_ij.y *= dev_bond_lj_scale;
            r_ij.z *= dev_bond_lj_scale;
            r_ij.w *= dev_bond_lj_scale;
            if ( dev_bond_lj_scale != 0 ) return r_ij;
        }
        else
        {
            return r_ij;
        }
    }
    DOUBLE_4 ret = {0.0f, 0.0f, 0.0f, flag};
    return ret;
}

template<bool pbc, bool dyn_cutoff, INT cg, bool chOverlap>
__global__ void comp_LJ( DOUBLE_4* ptr, DOUBLE_2* ptr_lj, DOUBLE_3* forces )
{
    INT id = blockIdx.x * blockDim.x + threadIdx.x; /*id*/
    extern __shared__ DOUBLE sh[];
    DOUBLE_4* shPosition = (DOUBLE_4*) sh;
    DOUBLE* shEps = sh + 4 * blockDim.x;
    DOUBLE_4* tab = (DOUBLE_4*) ptr;
    DOUBLE_3* ftab = (DOUBLE_3*) forces;
    DOUBLE_4 curr = tab[id];
    DOUBLE_3 f_curr = {0,0,0};
    DOUBLE_2* vLJ = (DOUBLE_2*) ptr_lj;
    DOUBLE eps_curr = 0;
    DOUBLE e_curr = 0;
    DOUBLE_2 currLJ = vLJ[id];
    curr.w = currLJ.x;
    eps_curr = currLJ.y;
    for ( INT i = 0; i < gridDim.x; ++i )
    {
        DOUBLE_4 posi = tab[ i * blockDim.x + threadIdx.x ];
        DOUBLE_2 LJi = vLJ[ i * blockDim.x + threadIdx.x ];
        posi.w = LJi.x;
        shPosition[ threadIdx.x ] = posi;
        shEps[ threadIdx.x ] = LJi.y;
        __syncthreads( );
        for ( INT j = 0; j < blockDim.x; ++j )
        {
            DOUBLE_4 f;
            INT jd = i * blockDim.x + j;
            if ( cg == 0 ) f = sLJ<pbc, dyn_cutoff,chOverlap> ( curr, shPosition[j], shEps[j], eps_curr );
            else f = sLJ_g<pbc, dyn_cutoff,chOverlap,cg> ( curr, shPosition[j], shEps[j], eps_curr, id, jd, dev_gconns );
            f_curr.x += (id!=jd) ? f.x : 0.0f;
            f_curr.y += (id!=jd) ? f.y : 0.0f;
            f_curr.z += (id!=jd) ? f.z : 0.0f;
            e_curr += (id!=jd) ? f.w : 0.0f;
        }
        __syncthreads( );
    }
    f_curr.x *= dev_alpha_lj;
    f_curr.y *= dev_alpha_lj;
    f_curr.z *= dev_alpha_lj;
    e_curr   *= dev_alpha_lj;

    ftab[id] = f_curr;
    if ( id >= dev_size ) e_curr = 0;
    energy_reduction( (DOUBLE*)forces + 3 * blockDim.x * gridDim.x + blockIdx.x, e_curr );
}

INT check_E( DOUBLE* _E, INT energy_type, INT iters, DOUBLE* es = h_forces )
{
    INT rep = 0;
    INT i;
    if ( _E )
    {
        _E[ energy_type ] = 0;
        for ( i = 0; i < iters; ++i )
        {
            _E[ energy_type ] += es[ i ];
        }
        _E[ energy_type ] *= 0.5;
        rep = isnan( _E[ energy_type ] );
    }
    else
    {
        for ( i = 0; i < iters; ++i )
        {
            rep |= isnan( es[ i ] );
        }
    }
    return rep;
}

template<bool pbc, INT cg, bool checkOverlap>
void run_elec_bs()
{
    INT shSize = cuda_block * 5 * sizeof(DOUBLE);
    comp_electro<pbc,cg,checkOverlap> <<<cblocks, cuda_block, shSize >>>((DOUBLE_4*)d_coord, d_Q, (DOUBLE_2*)d_lj, (DOUBLE_3*)d_forces );
}

template<bool pbc,bool checkOverlap>
void run_elec_g()
{
    switch(max_gs)
    {
        case(8): run_elec_bs<pbc,8,checkOverlap>(); break;
	case(7): run_elec_bs<pbc,7,checkOverlap>(); break;
	case(6): run_elec_bs<pbc,6,checkOverlap>(); break;
	case(5): run_elec_bs<pbc,5,checkOverlap>(); break;
	case(4): run_elec_bs<pbc,4,checkOverlap>(); break;
	case(3): run_elec_bs<pbc,3,checkOverlap>(); break;
	case(2): run_elec_bs<pbc,2,checkOverlap>(); break;
	case(1): run_elec_bs<pbc,1,checkOverlap>(); break;
	case(0): run_elec_bs<pbc,0,checkOverlap>(); break;
	default: if ( max_gs > 8 ) run_elec_bs<pbc,9,checkOverlap>();
                 else UNERR("Unimplemented")
    }
}

template<bool checkOverlap>
INT calc_elec( DOUBLE* _E )
{
    if ( bc == BC_BOX ) run_elec_g<true, checkOverlap>();
    else                run_elec_g<false,checkOverlap>();
    cudaThreadSynchronize( ); CCERR
    cudaMemcpy( h_forces, d_forces + DIMS0*csize, cblocks * sizeof (DOUBLE), cudaMemcpyDeviceToHost ); CCERR
    return check_E( _E, ENERGY_COULOMB, cblocks );
}

template <bool pbc, bool dyn_cutoff,INT cg,bool chOverlap>
void run_LJ_g()
{
    INT shSize = cuda_block * 5 * sizeof(DOUBLE);
    comp_LJ < pbc, dyn_cutoff,cg,chOverlap> <<<cblocks, cuda_block, shSize>>>((DOUBLE_4*)d_coord,(DOUBLE_2*)d_lj,(DOUBLE_3*)d_forces);
}

template <bool pbc, bool dyn_cutoff,bool chOverlap>
void run_LJ()
{
    switch(max_gs)
    {
        case(8): run_LJ_g<pbc,dyn_cutoff,8,chOverlap>(); break;
        case(7): run_LJ_g<pbc,dyn_cutoff,7,chOverlap>(); break;
        case(6): run_LJ_g<pbc,dyn_cutoff,6,chOverlap>(); break;
        case(5): run_LJ_g<pbc,dyn_cutoff,5,chOverlap>(); break;
        case(4): run_LJ_g<pbc,dyn_cutoff,4,chOverlap>(); break;
        case(3): run_LJ_g<pbc,dyn_cutoff,3,chOverlap>(); break;
        case(2): run_LJ_g<pbc,dyn_cutoff,2,chOverlap>(); break;
        case(1): run_LJ_g<pbc,dyn_cutoff,1,chOverlap>(); break;
        case(0): run_LJ_g<pbc,dyn_cutoff,0,chOverlap>(); break;
        default: if ( max_gs > 8 ) run_LJ_g<pbc,dyn_cutoff,9,chOverlap>();
                 else UNERR("Unimplemented")
    }
}

template<bool chOverlap>
INT calc_LJ( DOUBLE* _E )
{
    if ( bc == BC_BOX )
    {
        if ( cutoff_lj <= 0.0f )
        {
            run_LJ<true,true,chOverlap>();
        }
        else
        {
            run_LJ<true,false,chOverlap>();
        }
    }
    else
    {
        if ( cutoff_lj <= 0.0f )
        {
            run_LJ<false,true,chOverlap>();
        }
        else
        {
            run_LJ<false,false,chOverlap>();
        }
    }
    cudaThreadSynchronize( ); CCERR

    cudaMemcpy( h_forces, d_forces + DIMS0*csize, cblocks * sizeof (DOUBLE), cudaMemcpyDeviceToHost ); CCERR

    return check_E( _E, ENERGY_LJ, cblocks );
}

template <bool pbc, bool dyn_cutoff,INT cg,bool chOverlap>
void run_electro_LJ_g()
{
    INT shSize = cuda_block * 6 * sizeof(DOUBLE);
    comp_electroLJ<pbc,dyn_cutoff,cg,chOverlap> <<<cblocks,cuda_block,shSize>>>((DOUBLE_4*)d_coord, d_Q, (DOUBLE_2*)d_lj, (DOUBLE_3*)d_forces);
}

template <bool pbc, bool dyn_cutoff, bool chOverlap>
void run_electro_LJ()
{
    switch(max_gs)
    {
	case(8): run_electro_LJ_g<pbc,dyn_cutoff,8,chOverlap>(); break;
	case(7): run_electro_LJ_g<pbc,dyn_cutoff,7,chOverlap>(); break;
	case(6): run_electro_LJ_g<pbc,dyn_cutoff,6,chOverlap>(); break;
	case(5): run_electro_LJ_g<pbc,dyn_cutoff,5,chOverlap>(); break;
	case(4): run_electro_LJ_g<pbc,dyn_cutoff,4,chOverlap>(); break;
	case(3): run_electro_LJ_g<pbc,dyn_cutoff,3,chOverlap>(); break;
	case(2): run_electro_LJ_g<pbc,dyn_cutoff,2,chOverlap>(); break;
	case(1): run_electro_LJ_g<pbc,dyn_cutoff,1,chOverlap>(); break;
	case(0): run_electro_LJ_g<pbc,dyn_cutoff,0,chOverlap>(); break;
	default: if ( max_gs > 8 ) run_electro_LJ_g<pbc,dyn_cutoff,9,chOverlap>();
	else UNERR("Unimplemented")		
    }
}

template<bool chOverlap>
INT calc_electro_LJ( DOUBLE* _E )
{
    if ( bc == BC_BOX )
    {
        if ( cutoff_lj <= 0.0f )
        {
            run_electro_LJ<true,true,chOverlap>();
        }
        else
        {
            run_electro_LJ<true,false,chOverlap>();
        }
    }
    else
    {
        if ( cutoff_lj <= 0.0f )
        {
            run_electro_LJ<false,true,chOverlap>();
        }
        else
        {
            run_electro_LJ<false,false,chOverlap>();
        }
    }
    cudaThreadSynchronize( ); CCERR

    cudaMemcpy( h_forces, d_forces + DIMS0*csize, 2*cblocks*sizeof(DOUBLE), cudaMemcpyDeviceToHost ); CCERR

    INT rep = check_E( _E, ENERGY_COULOMB, cblocks );
    rep |= check_E( _E, ENERGY_LJ, cblocks, h_forces + cblocks );
    return rep;
}

INT calc_overlap()
{
    INT rep;
    INT shMemSize = cuda_block * 4 * sizeof(DOUBLE);
    rep = 0;
    cudaMemset( d_overlap, 0, sizeof (INT) ); CCERR

    if ( bc == BC_BOX )
    {
        check_beads_overlap<true> <<<cblocks, cuda_block, shMemSize >>>(d_coord, d_lj, d_overlap);
    }
    else
    {
#if USE_FLOAT
        make_coords<<<cblocks,cuda_block>>>( (DOUBLE_4*)d_ljcoord, (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj );
        cudaThreadSynchronize( ); CCERR
        switch( cuda_block )
        {
            case(512): check_beads_overlap_tex<false,512><<<cblocks, cuda_block, shMemSize >>>( d_overlap ); break;
            case(256): check_beads_overlap_tex<false,256><<<cblocks, cuda_block, shMemSize >>>( d_overlap ); break;
            case(128): check_beads_overlap_tex<false,128><<<cblocks, cuda_block, shMemSize >>>( d_overlap ); break;
            case( 64): check_beads_overlap_tex<false, 64><<<cblocks, cuda_block, shMemSize >>>( d_overlap ); break;
            case( 32): check_beads_overlap_tex<false, 32><<<cblocks, cuda_block, shMemSize >>>( d_overlap ); break;
            default: UNERR("Wrong cuda_block value")
        }
#else
        check_beads_overlap<false> <<<cblocks, cuda_block, shMemSize >>>(d_coord, d_lj, d_overlap);
#endif
    }
    cudaThreadSynchronize( ); CCERR
    INT overlap;
    cudaMemcpy( &overlap, d_overlap, sizeof (INT), cudaMemcpyDeviceToHost ); CCERR
    rep = rep || (INT)(overlap);
    return rep;
}

template<bool chOverlap>
INT brute_force( DOUBLE* _E )
{
    INT rep = 0;
    if ( elec || alpha_lj != 0.0 ) /* if elec or lj turn on*/
    {
        if ( elec && alpha_lj == 0.0 ) /* only electro static */
        {
            rep = calc_elec<chOverlap>( _E );
        }
        else if ( !elec && alpha_lj != 0.0 ) /* only LJ */
        {
            rep = calc_LJ<chOverlap>( _E );
        }
        else /* both */
        {
            rep = calc_electro_LJ<chOverlap>( _E );
        }
    }
    else
    {
        if ( chOverlap ) rep = calc_overlap();
    }
    return rep;
}

__device__ void red_single( INT tid, INT curr_size, DOUBLE* sh )
{
    sh[tid             ] = fmax( sh[tid             ], sh[tid+curr_size             ] );
    sh[tid+  blockDim.x] = fmax( sh[tid+  blockDim.x], sh[tid+curr_size+  blockDim.x] );
    sh[tid+2*blockDim.x] = fmax( sh[tid+2*blockDim.x], sh[tid+curr_size+2*blockDim.x] );
    sh[tid+3*blockDim.x] = fmin( sh[tid+3*blockDim.x], sh[tid+curr_size+3*blockDim.x] );
    sh[tid+4*blockDim.x] = fmin( sh[tid+4*blockDim.x], sh[tid+curr_size+4*blockDim.x] );
    sh[tid+5*blockDim.x] = fmin( sh[tid+5*blockDim.x], sh[tid+curr_size+5*blockDim.x] );
}

template<INT cuda_block,INT curr_size>
__device__ void red6( INT tid )
{
    extern __shared__ DOUBLE sh[];
    if ( cuda_block >= 2*curr_size )
    {
        if ( tid < curr_size )
        {
            red_single( tid, curr_size, sh );
        }
    }
}

template<INT cuda_block>
__global__ void comp_h_l( void* coord, DOUBLE* fout )
{
    INT tid = threadIdx.x;
    INT id = blockDim.x*blockIdx.x + tid;
    DOUBLE_4* cds = (DOUBLE_4*)coord;
    extern __shared__ DOUBLE sh[];
    DOUBLE hx, hy, hz, lx, ly, lz;
    hx = hy = hz = -FLT_MAX;
    lx = ly = lz = FLT_MAX;
    while ( id < dev_size )
    {
        DOUBLE x, y, z;
        x = cds[id].x;
        y = cds[id].y;
        z = cds[id].z;
        hx = fmax(x,hx); hy = fmax(y,hy); hz = fmax(z,hz);
        lx = fmin(x,lx); ly = fmin(y,ly); lz = fmin(z,lz);
        id += blockDim.x*gridDim.x;
    }
    sh[tid             ] = hx;
    sh[tid+  blockDim.x] = hy;
    sh[tid+2*blockDim.x] = hz;
    sh[tid+3*blockDim.x] = lx;
    sh[tid+4*blockDim.x] = ly;
    sh[tid+5*blockDim.x] = lz;

    __syncthreads();
    red6<cuda_block,512>( tid ); __syncthreads();
    red6<cuda_block,256>( tid ); __syncthreads();
    red6<cuda_block,128>( tid ); __syncthreads();
    red6<cuda_block, 64>( tid ); __syncthreads();

    if ( cuda_block >= 64 )
    {
        if( tid < 32 )
        {
            red_single( tid, 32, sh );
            red_single( tid, 16, sh );
            red_single( tid,  8, sh );
            red_single( tid,  4, sh );
            red_single( tid,  2, sh );
            red_single( tid,  1, sh );
        }
    }
    else if( cuda_block == 32 )
    {
        if( tid < 16 )
        {
            red_single( tid, 16, sh );
            red_single( tid,  8, sh );
            red_single( tid,  4, sh );
            red_single( tid,  2, sh );
            red_single( tid,  1, sh );
        }
    }// no else
    if ( tid == 0 )
    {
#pragma unroll 6
        for ( INT i = 0; i < 6; ++i )
            fout[6*blockIdx.x+i]=sh[i*blockDim.x];
    }
}

template<INT last>
__device__ INT correct_indx( INT indx )
{
    return ( indx < 0 ) ? (last + indx) : (indx)%last;
}

__device__
INT make_key( INT x, INT y, INT z )
{
    INT cx = correct_indx<BUCKET_SPATIAL_M + 1>( x );
    INT cy = correct_indx<BUCKET_SPATIAL_M + 1>( y );
    INT cz = correct_indx<BUCKET_SPATIAL_M + 1>( z );
    INT key = ( cx << BUCKET_SPATIAL_X) |
              ( cy << BUCKET_SPATIAL_Y) |
              ( cz << BUCKET_SPATIAL_Z);
    return key;
}

template<int i>
__device__ void add_key( INT* kkeys, INT* kvalues, INT id, INT kx, INT ky, INT kz )
{
    INT key = make_key( kx+(i&1), ky+((i>>1)&1), kz+(i>>2) );
    kkeys[ PH_COUNT*id + i ] = key;
    kvalues[ PH_COUNT*id + i ] = (id<<PATTERN_BITS) + i;
}

__global__ void compute_keys_value( DOUBLE_4* coords, INT* kkeys, INT* kvals )
{
    INT id = blockDim.x*blockIdx.x + threadIdx.x;

    if ( id < dev_size )
    {
        INT kx, ky, kz;
        kx = (INT)floor( (coords[id].x + dev_offset[0]) / dev_blen[0] );
        ky = (INT)floor( (coords[id].y + dev_offset[1]) / dev_blen[1] );
        kz = (INT)floor( (coords[id].z + dev_offset[2]) / dev_blen[2] );
        add_key<0>( kkeys, kvals, id, kx, ky, kz);
        add_key<1>( kkeys, kvals, id, kx, ky, kz);
        add_key<2>( kkeys, kvals, id, kx, ky, kz);
        add_key<3>( kkeys, kvals, id, kx, ky, kz);
        add_key<4>( kkeys, kvals, id, kx, ky, kz);
        add_key<5>( kkeys, kvals, id, kx, ky, kz);
        add_key<6>( kkeys, kvals, id, kx, ky, kz);
        add_key<7>( kkeys, kvals, id, kx, ky, kz);
    }
}

__device__ void add_key_periodic_ph( INT* kkeys, INT* kvals, INT key, INT id, INT i )
{
    unsigned old = atomicInc( &d_keys_size, 1<<30 );
    kkeys[ PH_COUNT*dev_size + old ] = key;
    kvals[ PH_COUNT*dev_size + old ] = (id<<PATTERN_BITS) | i | PATTERN_PH;
}

__device__ INT in_phan_reg_down( DOUBLE x, INT dim )
{
    switch(dim)
    {
        case 0: return x <= -dev_box0/2 + dev_max_cutoff + EPS;
        case 1: return x <= -dev_box1/2 + dev_max_cutoff + EPS;
        case 2: return x <= -dev_box2/2 + dev_max_cutoff + EPS;
    }
    return 0;
}

__device__ INT in_phan_reg_up( DOUBLE x, INT dim )
{
    switch(dim)
    {
        case 0: return x >= dev_box0/2 - dev_max_cutoff - EPS;
        case 1: return x >= dev_box1/2 - dev_max_cutoff - EPS;
        case 2: return x >= dev_box2/2 - dev_max_cutoff - EPS;
    }
    return 0;
}

__global__ void compute_keys_value_periodic( DOUBLE* coords, INT* kkeys, INT* kvals )
{
    INT id = blockDim.x*blockIdx.x + threadIdx.x;
    if ( id < dev_size )
    {
        INT kx, ky, kz;
        INT k, n, m;
        DOUBLE xx = coords[id*DIMS1];
        DOUBLE yy = coords[id*DIMS1+1];
        DOUBLE zz = coords[id*DIMS1+2];
        kx = (INT)floor( (xx + dev_offset[0]) / dev_blen[0] );
        ky = (INT)floor( (yy + dev_offset[1]) / dev_blen[1] );
        kz = (INT)floor( (zz + dev_offset[2]) / dev_blen[2] );
        add_key<0>( kkeys, kvals, id, kx, ky, kz);
        add_key<1>( kkeys, kvals, id, kx, ky, kz);
        add_key<2>( kkeys, kvals, id, kx, ky, kz);
        add_key<3>( kkeys, kvals, id, kx, ky, kz);
        add_key<4>( kkeys, kvals, id, kx, ky, kz);
        add_key<5>( kkeys, kvals, id, kx, ky, kz);
        add_key<6>( kkeys, kvals, id, kx, ky, kz);
        add_key<7>( kkeys, kvals, id, kx, ky, kz);
        for ( k = 0; k <= in_phan_reg_down(xx,0); ++k )
                for ( n = 0; n <= in_phan_reg_down(yy,1); ++n )
                    for ( m = 0; m <= in_phan_reg_down(zz,2); ++m )
                    {
                        if( k | n | m )
                        {
                            INT ix,iy,iz;
                            INT a,b,c;
                            INT d,e,f;
                            DOUBLE dx,dy,dz;
                            dx = xx + k * dev_box0;
                            dy = yy + n * dev_box1;
                            dz = zz + m * dev_box2;

                            ix = (INT)FLOORD((dx+dev_offset[0])/dev_blen[0]);
                            iy = (INT)FLOORD((dy+dev_offset[1])/dev_blen[1]);
                            iz = (INT)FLOORD((dz+dev_offset[2])/dev_blen[2]);

                            for ( a = 0; a <= 1-k; ++a )
                                for ( b = 0; b <= 1-n; ++b )
                                    for ( c = 0; c <= 1-m; ++c )
                                    {
                                        INT nkey = make_key( ix+a, iy+b, iz+c );
                                        add_key_periodic_ph( kkeys, kvals, nkey, id, (a)|(b<<1)|(c<<2) );
                                    }
                            for ( d = 0; d <= in_phan_reg_up(xx,0); ++d )
                                for ( e = 0; e <= in_phan_reg_up(yy,0); ++e )
                                    for ( f = 0; f <= in_phan_reg_up(zz,0); ++f )
                                    {
                                        if( (d | e | f) && (( (!k & !d) + 3*(!n & !e) + 9*(!m & !f) + d*2 + e*6 + f*18 ) <= 13) )
                                        {
                                            INT cx, cy, cz;
                                            cx = (INT)FLOORD((dx-d*dev_box0+dev_offset[0])/dev_blen[0]);
                                            cy = (INT)FLOORD((dy-e*dev_box1+dev_offset[1])/dev_blen[1]);
                                            cz = (INT)FLOORD((dz-f*dev_box2+dev_offset[2])/dev_blen[2]);
                                            if ( cx>=-1 && cy >=-1 && cz>=-1)
                                            {
                                                cx = d ? 0 : ix;
                                                cy = e ? 0 : iy;
                                                cz = f ? 0 : iz;

                                                for ( a = 0; a <= (1-k)*(1-d); ++a )
                                                    for ( b = 0; b <= (1-n)*(1-e); ++b )
                                                        for ( c = 0; c <= (1-m)*(1-f); ++c )
                                                        {
                                                            INT nkey = make_key(cx+a,cy+b,cz+c);
                                                            add_key_periodic_ph( kkeys, kvals, nkey, id, (a)|(b<<1)|(c<<2) );
                                                        }
                                            }
                                        }
                                    }
                        }
                    }
    }
}

#define MASK (~15)
#define COUNT 16

template<bool pbc,INT op>
__device__ INT check_s_overlap( DOUBLE_4 curr, volatile DOUBLE_4* cop, INT true_id, volatile INT* iop)
{
    INT res = isoverlap<pbc>( curr, cop[op].x, cop[op].y, cop[op].z, cop[op].w );
    res = (true_id != iop[ op ]) * res;
    return res;
}

template<bool pbc>
__global__ void checking_overlap( DOUBLE_4* coords, DOUBLE_2* lj, INT* kkeys, INT* kvals, INT* out )
{
    INT tid = threadIdx.x;
    INT id = blockDim.x*blockIdx.x + tid;
    INT mid = tid&(COUNT-1);
    extern __shared__ DOUBLE sh[];
    volatile DOUBLE_4* sh_coord = (DOUBLE_4*)sh  + (tid&MASK);
    volatile INT* sh_ids = (INT*)(sh + 4*blockDim.x)  + (tid&MASK);
    INT last_warp_key = kkeys[ (id & MASK) + COUNT - 1 ];//mask of last in wrap
    INT curr_id = kvals[id]>>PATTERN_BITS;
    INT rep = 0;
    DOUBLE_4 curr = coords[curr_id];
    curr.w = lj[curr_id].x;
    sh_coord[mid].x = curr.x;
    sh_coord[mid].y = curr.y;
    sh_coord[mid].z = curr.z;
    sh_coord[mid].w = curr.w;
    sh_ids[mid] = curr_id;

    while(1)
    {
        rep = rep || check_s_overlap<pbc, 0>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 1>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 2>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 3>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 4>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 5>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 6>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 7>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 8>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc, 9>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,10>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,11>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,12>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,13>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,14>( curr, sh_coord, curr_id, sh_ids );
        rep = rep || check_s_overlap<pbc,15>( curr, sh_coord, curr_id, sh_ids );

        id -= COUNT;
        if ( id > 0 )
        {
            INT ckey = kkeys[ id & MASK ];
            INT  cid = kvals[ id ]>>PATTERN_BITS;
            sh_coord[mid].x = coords[cid].x;
            sh_coord[mid].y = coords[cid].y;
            sh_coord[mid].z = coords[cid].z;
            sh_coord[mid].w = lj[cid].x;
            sh_ids[mid] = cid;

            if ( ckey != last_warp_key ) break;
        }
        else
        {
            break;
        }
    }
    if ( rep ) *out = rep;
}

template<bool pbc, bool checkOverlap, INT cg>
__device__ DOUBLE_4 compute_electro( DOUBLE_4 curr, DOUBLE c_lj_r , INT c_key,
                                   volatile DOUBLE_4* cop, volatile DOUBLE* lj_r, volatile INT* kop,
                                   INT ckid, volatile INT* iop,
                                   INT op )
{
    DOUBLE_4 res = {0,0,0,0};
    INT p1 = ckid & PATTERN_MASK;
    INT p2 = iop[ op ] & PATTERN_MASK;
    INT pred = ((ckid>>PATTERN_BITS) != (iop[ op ]>>PATTERN_BITS)) && (c_key == kop[op]) && !(p1 & p2);
    if ( pred )
    {
        DOUBLE_4 r_op = { cop[ op ].x, cop[ op ].y, cop[ op ].z, cop[ op ].w };
        if ( cg == 0 ) res = electro_s<pbc,checkOverlap>( curr, r_op, c_lj_r, lj_r[op] );
        else res = electro_g<pbc,cg,checkOverlap>( curr, r_op, c_lj_r, lj_r[op],
                ckid>>PATTERN_BITS, iop[ op ]>>PATTERN_BITS, dev_gconns );
    }
    return res;
}

__device__ DOUBLE_4 add4( DOUBLE_4 a, DOUBLE_4 b)
{
    a.x += b.x;
    a.y += b.y;
    a.z += b.z;
    a.w += b.w;
    return a;
}

__device__ DOUBLE_8 add8( DOUBLE_8 a, DOUBLE_8 b)
{
    a.a = add4( a.a, b.a );
    a.b = add4( a.b, b.b );
    return a;
}

#if USE_FLOAT
template<bool pbc,bool checkOverlap, INT cg>
__global__ void compute_electro_spatial_indxs( DOUBLE* forces, INT* kkeys, INT* kvals, INT* indxs, UINT* pos_q )
#else
template<bool pbc,bool checkOverlap, INT cg>
__global__ void compute_electro_spatial_indxs( DOUBLE_4* coords, DOUBLE_2* LJ, DOUBLE* Q,
                                               DOUBLE* forces, INT* kkeys, INT* kvals,
                                               INT* indxs, UINT* pos_q )
#endif
{
    INT tid = threadIdx.x;
    INT mid = tid & (COUNT-1);
    INT id = blockDim.x*blockIdx.x + tid;
    extern __shared__ DOUBLE sh[];
    volatile DOUBLE_4* sh_coord = (DOUBLE_4*)sh + (tid&MASK);//begin of wrap memory
    volatile DOUBLE* sh_lj_r = sh + 4*blockDim.x  + (tid&MASK);
    volatile INT* sh_kids = (INT*)(sh + 5*blockDim.x)  + (tid&MASK);
    volatile INT* sh_keys = sh_kids + blockDim.x;
    INT curr_key = kkeys[id];
    INT curr_kid = kvals[id];
#if USE_FLOAT
    DOUBLE curr_lj_r = (tex1Dfetch( tr_lj, curr_kid>>PATTERN_BITS )).x;
    DOUBLE_4 curr = tex1Dfetch( tr_coords, curr_kid>>PATTERN_BITS );
    curr.w = tex1Dfetch( tr_q, curr_kid>>PATTERN_BITS );
#else
    DOUBLE curr_lj_r = LJ[curr_kid>>PATTERN_BITS].x;
    DOUBLE_4 curr = coords[curr_kid>>PATTERN_BITS];
    curr.w = Q[curr_kid>>PATTERN_BITS];
#endif
    DOUBLE_4 res = {0,0,0,0};
    INT pos = id/COUNT;
    INT beg = indxs[ 2*pos   ];
    INT end = indxs[ 2*pos +1];

    for ( INT id = beg; id < end; id += COUNT )
    {
        INT cid = kvals[ id+mid ];
#if USE_FLOAT
        DOUBLE_4 sc = tex1Dfetch( tr_coords, cid>>PATTERN_BITS );
        DOUBLE_2 slj = tex1Dfetch( tr_lj, cid>>PATTERN_BITS );
#else
        DOUBLE_4 sc = coords[cid>>PATTERN_BITS];
        DOUBLE_2 slj = LJ[cid>>PATTERN_BITS];
#endif
        sh_coord[mid].x = sc.x;
        sh_coord[mid].y = sc.y;
        sh_coord[mid].z = sc.z;
#if USE_FLOAT
        sh_coord[mid].w = tex1Dfetch( tr_q, cid>>PATTERN_BITS );
#else
        sh_coord[mid].w = Q[cid>>PATTERN_BITS];
#endif
        sh_lj_r[mid] = slj.x;
        sh_kids[mid] = cid;
        sh_keys[mid] = kkeys[ id+mid ];

        for( INT i = 0; i < COUNT; ++i )
        {
            res = add4(res,compute_electro<pbc,checkOverlap,cg>( curr, curr_lj_r, curr_key, sh_coord,
                sh_lj_r, sh_keys, curr_kid, sh_kids, i ) );
        }
    }

    id = blockDim.x*blockIdx.x + tid;
    if ( id < PH_COUNT*dev_size + d_keys_size )// ommit dummy beads writes
    {
        INT out_id =  curr_kid & PATTERN_MASK;
        UINT curr_id = curr_kid >> PATTERN_BITS;
        if ( pbc )
        {
            if ( out_id & PATTERN_PH )
                out_id = atomicInc( pos_q + curr_id, 1<<30 ) + PH_COUNT;
        }
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id  ] = res.x * curr.w * dev_gamma_eps;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+1] = res.y * curr.w * dev_gamma_eps;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+2] = res.z * curr.w * dev_gamma_eps;

        if( pbc ) forces[ curr_id + DIMS0*dev_size*MAX_MULT + dev_size*out_id ] = res.w * curr.w * dev_gamma_eps;
        else forces[ curr_id + DIMS0*dev_size*PH_COUNT + dev_size*out_id ] = res.w * curr.w * dev_gamma_eps;
    }
}

template<bool pbc,bool dyn_cutoff,bool chOverlap,INT cg>
__device__ DOUBLE_4 compute_LJ( DOUBLE_4 curr, DOUBLE curr_eps, INT c_key,
                              volatile DOUBLE_4* cop, volatile DOUBLE* lj_eps, volatile INT* kop,
                              INT ckid, volatile INT* iop, INT op )
{
    DOUBLE_4 res = {0,0,0,0};
    INT p1 = ckid & PATTERN_MASK;
    INT p2 = iop[ op ] & PATTERN_MASK;
    INT pred = ((ckid>>PATTERN_BITS) != (iop[ op ]>>PATTERN_BITS)) && (c_key == kop[op]) && !(p1 & p2);
    if(pred)
    {
        DOUBLE_4 r_op = { cop[ op ].x, cop[ op ].y, cop[ op ].z, cop[ op ].w };
        if ( cg == 0 ) res = sLJ<pbc,dyn_cutoff,chOverlap>( curr, r_op, curr_eps, lj_eps[op] );
        else res = sLJ_g<pbc,dyn_cutoff,chOverlap,cg>( curr, r_op, curr_eps, lj_eps[op],
                ckid>>PATTERN_BITS, iop[ op ]>>PATTERN_BITS, dev_gconns );
    }
    return res;
}

#if USE_FLOAT
template<bool pbc, bool dyn_cutoff, bool chOverlap, INT cg>
__global__ void compute_LJ_spatial_indxs( DOUBLE* forces, INT* kkeys, INT* kvals, INT* indxs, UINT* pos_lj )
#else
template<bool pbc, bool dyn_cutoff, bool chOverlap, INT cg>
__global__ void compute_LJ_spatial_indxs( DOUBLE_4* coords, DOUBLE_2* LJ,
                                          DOUBLE* forces, INT* kkeys, INT* kvals, 
                                          INT* indxs, UINT* pos_lj )
#endif
{
    INT tid = threadIdx.x;
    INT mid = tid & (COUNT-1);
    INT id = blockDim.x*blockIdx.x + tid;
    extern __shared__ DOUBLE sh[];
    volatile DOUBLE_4* sh_coord = (DOUBLE_4*)sh + (tid&MASK);//begin of wrap memory
    volatile DOUBLE* sh_lj_eps = sh + 4*blockDim.x  + (tid&MASK);
    volatile INT* sh_kids = (INT*)(sh + 5*blockDim.x)  + (tid&MASK);
    volatile INT* sh_keys = sh_kids + blockDim.x;
    INT curr_key = kkeys[id];
    INT curr_kid = kvals[id];
    DOUBLE_4 res = {0,0,0,0};
#if USE_FLOAT
    DOUBLE_4 curr = tex1Dfetch( tr_coords, curr_kid>>PATTERN_BITS );
    DOUBLE_2 curr_slj = tex1Dfetch( tr_lj, curr_kid>>PATTERN_BITS );
#else
    DOUBLE_4 curr = coords[curr_kid>>PATTERN_BITS];
    DOUBLE_2 curr_slj = LJ[curr_kid>>PATTERN_BITS];
#endif
    DOUBLE curr_eps = curr_slj.y;
    curr.w = curr_slj.x;
    INT pos = id/COUNT;
    INT beg = indxs[ 2*pos   ];
    INT end = indxs[ 2*pos +1];
    for ( INT id = beg; id < end; id += COUNT )
    {
        INT  cid = kvals[ id+mid ];
#if USE_FLOAT
        DOUBLE_4 sc = tex1Dfetch( tr_coords, cid>>PATTERN_BITS );
        DOUBLE_2 slj = tex1Dfetch( tr_lj, cid>>PATTERN_BITS );
#else
        DOUBLE_4 sc = coords[cid>>PATTERN_BITS];
        DOUBLE_2 slj = LJ[cid>>PATTERN_BITS];
#endif
        sh_coord[mid].x = sc.x;
        sh_coord[mid].y = sc.y;
        sh_coord[mid].z = sc.z;
        sh_coord[mid].w = slj.x;
        sh_lj_eps[mid] = slj.y;
        sh_kids[mid] = cid;
        sh_keys[mid] = kkeys[ id+mid ];

        for( INT i = 0; i < COUNT; ++i )
        {
            res = add4( res, compute_LJ<pbc,dyn_cutoff,chOverlap,cg>( curr, curr_eps,
                                                   curr_key, sh_coord, sh_lj_eps,
                                                   sh_keys, curr_kid, sh_kids, i) );
        }
    }
    id = blockDim.x*blockIdx.x + tid;
    if ( id < PH_COUNT*dev_size + d_keys_size )// ommit dummy beads
    {
        INT out_id =  curr_kid & PATTERN_MASK;
        UINT curr_id = curr_kid >> PATTERN_BITS;
        if ( pbc )
        {
            if ( out_id & PATTERN_PH )
            out_id = atomicInc( pos_lj + curr_id, 1<<30 ) + PH_COUNT;
        }
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id  ] = res.x * dev_alpha_lj;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+1] = res.y * dev_alpha_lj;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+2] = res.z * dev_alpha_lj;

        if( pbc )
            forces[ curr_id + DIMS0*dev_size*MAX_MULT + dev_size*out_id ] = res.w * dev_alpha_lj;
        else
            forces[ curr_id + DIMS0*dev_size*PH_COUNT + dev_size*out_id ] = res.w * dev_alpha_lj;
    }
}

template<bool pbc,bool dyn_cutoff,bool chOverlap, INT cg>
__device__ DOUBLE_8 compute_electroLJ( DOUBLE_4 curr, DOUBLE_2 c_lj , INT c_key, 
                                       INT ckid, volatile DOUBLE_4* cop, volatile DOUBLE_2* lj,
                                       volatile INT* kop, volatile INT* iop, INT op )
{
    DOUBLE_8 res = {0,0,0,0,0,0,0,0};
    DOUBLE_2 r_lj = { lj[op].x, lj[op].y };
    INT p1 = ckid & PATTERN_MASK;
    INT p2 = iop[op] & PATTERN_MASK;
    INT pred = ((ckid>>PATTERN_BITS) != (iop[ op ]>>PATTERN_BITS)) && (c_key == kop[op]) && !(p1&p2);
    if(pred)
    {
        DOUBLE_4 r_op = { cop[ op ].x, cop[ op ].y, cop[ op ].z, cop[ op ].w };
        if ( cg == 0 ) res = electroLJ<pbc,dyn_cutoff,chOverlap>( curr, r_op, c_lj, r_lj );
        else  res = electroLJ_g<pbc,dyn_cutoff,cg,chOverlap>( curr, r_op, c_lj, r_lj,
                ckid>>PATTERN_BITS, iop[ op ]>>PATTERN_BITS, dev_gconns );
    }
    return res;
}

#if USE_FLOAT
template<bool pbc,bool dyn_cutoff,bool chOverlap, INT cg>
__global__ void compute_electroLJ_spatial_indxs( DOUBLE* forces, INT* kkeys, INT* kvals,
                                                 INT* indxs, UINT* pos_qlj )
#else
template<bool pbc,bool dyn_cutoff,bool chOverlap, INT cg>
__global__ void compute_electroLJ_spatial_indxs( DOUBLE_4* coords, DOUBLE_2* LJ, DOUBLE* Q,
                                                 DOUBLE* forces, INT* kkeys, INT* kvals,
                                                 INT* indxs, UINT* pos_qlj )
#endif
{
    INT tid = threadIdx.x;
    INT mid = tid & (COUNT-1);
    INT id = blockDim.x*blockIdx.x + tid;
    extern __shared__ DOUBLE sh[];
    volatile DOUBLE_4* sh_coord = (DOUBLE_4*)sh + (tid&MASK);//begin of wrap memory
    volatile DOUBLE_2* sh_lj = (DOUBLE_2*)(sh + 4*blockDim.x) + (tid&MASK);
    volatile INT* sh_ids = (INT*)(sh + 6*blockDim.x) + (tid&MASK);
    volatile INT* sh_keys = sh_ids + blockDim.x;
    INT curr_key = kkeys[id];
    INT curr_kid = kvals[id];
#if USE_FLOAT
    DOUBLE_2 curr_lj = tex1Dfetch( tr_lj, curr_kid>>PATTERN_BITS );
    DOUBLE_4 curr = tex1Dfetch( tr_coords, curr_kid>>PATTERN_BITS );
    curr.w = tex1Dfetch( tr_q, curr_kid>>PATTERN_BITS );
#else
    DOUBLE_2 curr_lj = LJ[curr_kid>>PATTERN_BITS];
    DOUBLE_4 curr = coords[curr_kid>>PATTERN_BITS];
    curr.w = Q[curr_kid>>PATTERN_BITS];
#endif
    INT pos = id/COUNT;
    INT beg = indxs[ 2*pos   ];
    INT end = indxs[ 2*pos +1];
    DOUBLE_8 res = {0,0,0,0,0,0,0,0};
    for ( INT id = beg; id < end; id += COUNT )
    {
        INT  cid = kvals[ id+mid ];
#if USE_FLOAT
        DOUBLE_4 sc = tex1Dfetch( tr_coords, cid>>PATTERN_BITS );
        DOUBLE_2 slj = tex1Dfetch( tr_lj, cid>>PATTERN_BITS );
#else
        DOUBLE_4 sc = coords[ cid>>PATTERN_BITS ];
        DOUBLE_2 slj = LJ[cid>>PATTERN_BITS];
#endif
        sh_coord[mid].x = sc.x;
        sh_coord[mid].y = sc.y;
        sh_coord[mid].z = sc.z;
#if USE_FLOAT
        sh_coord[mid].w = tex1Dfetch( tr_q, cid>>PATTERN_BITS );
#else
        sh_coord[mid].w = Q[cid>>PATTERN_BITS];
#endif
        sh_lj[mid].x = slj.x;
        sh_lj[mid].y = slj.y;
        sh_ids[mid] = cid;
        sh_keys[mid] = kkeys[ id+mid ];

        for( INT i = 0; i < COUNT; ++i )
        {
            res = add8( res,compute_electroLJ<pbc,dyn_cutoff,chOverlap,cg>( curr, curr_lj,
                        curr_key, curr_kid, sh_coord, sh_lj, sh_keys, sh_ids, i ) );
        }
    }
    
    id = blockDim.x*blockIdx.x + tid;
    if ( id < PH_COUNT*dev_size + d_keys_size )// ommit dummy beads writes
    {
        INT out_id =  curr_kid & PATTERN_MASK;
        UINT curr_id = curr_kid >> PATTERN_BITS;
        if ( pbc )
        {
            if ( out_id & PATTERN_PH )
            out_id = atomicInc( pos_qlj + curr_id, 1<<30 ) + PH_COUNT;
        }
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id  ] = res.a.x * curr.w * dev_gamma_eps + res.b.x * dev_alpha_lj;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+1] = res.a.y * curr.w * dev_gamma_eps + res.b.y * dev_alpha_lj;
        forces[ DIMS0*curr_id + DIMS0*dev_size*out_id+2] = res.a.z * curr.w * dev_gamma_eps + res.b.z * dev_alpha_lj;

        if( pbc )
            forces[ curr_id + DIMS0*dev_size*MAX_MULT + dev_size*out_id ] = res.a.w * curr.w * dev_gamma_eps;
        else
            forces[ curr_id + DIMS0*dev_size*PH_COUNT + dev_size*out_id ] = res.a.w * curr.w * dev_gamma_eps;
        INT off = 0;
        if(pbc) off = dev_size*MAX_MULT;
        else off = dev_size*PH_COUNT;
        if( pbc )
            forces[ curr_id + DIMS0*dev_size*MAX_MULT + dev_size*out_id + off ] = res.b.w * dev_alpha_lj;
        else
            forces[ curr_id + DIMS0*dev_size*PH_COUNT + dev_size*out_id + off ] = res.b.w * dev_alpha_lj;
    }
}

__global__ void get_indx_spatial( INT* kkeys, INT* indxs )
{
    INT tid = threadIdx.x;
    INT last_indx = PH_COUNT*dev_size;
    last_indx += d_keys_size;
    if( (blockDim.x*blockIdx.x + tid)*COUNT < last_indx )
    {
        INT id_end = (blockDim.x*blockIdx.x + tid) * COUNT;/*first in warp*/
        INT last_warp_key = kkeys[ id_end + COUNT - 1 ];/*last in wrap*/
        while(1)
        {
            id_end += COUNT;
            if ( id_end < last_indx )
            {
                INT ckey = kkeys[ id_end ];
                if ( ckey != last_warp_key ) break;
            }
            else
            {
                break;
            }
        }
        INT id_beg = (blockDim.x*blockIdx.x + tid) * COUNT;
        INT first_warp_key = kkeys[ id_beg ];//first in wrap
        id_beg += COUNT-1; //move to last in warp
        while(1)
        {
            id_beg -= COUNT;
            if ( id_beg >= 0 )
            {
                INT ckey = kkeys[ id_beg ];
                if ( ckey != first_warp_key ) break;
            }
            else
            {
                break;
            }
        }
        INT id = blockDim.x*blockIdx.x + tid;
        indxs[ id*2   ] = (id_beg & MASK)+COUNT;// >=
        indxs[ id*2 +1] = id_end;// <
    }
    else
    {
        INT id = blockDim.x*blockIdx.x + tid;
        //no computation
        indxs[ id*2   ] = 0;// >=
        indxs[ id*2 +1] = 0;// <
    }
}

template<INT count, bool second>
__global__ void reduce_forces( DOUBLE* forces, DOUBLE* help )
{
    INT tid = threadIdx.x;
    INT id = blockDim.x*blockIdx.x + tid;
    DOUBLE e_curr = 0;
    DOUBLE es_curr = 0;
    if ( id < dev_size )
    {
        DOUBLE_3 f = {0, 0, 0};

        for ( INT i = 0; i < count; ++i )
        {
            f.x += forces[ DIMS0*id + DIMS0*dev_size*i  ];
            f.y += forces[ DIMS0*id + DIMS0*dev_size*i+1];
            f.z += forces[ DIMS0*id + DIMS0*dev_size*i+2];
        }
        forces[ DIMS0*id  ] = f.x;
        forces[ DIMS0*id+1] = f.y;
        forces[ DIMS0*id+2] = f.z;

        INT offset = DIMS0*dev_size*count;
        for ( INT i = 0; i < count; ++i )
        {
            e_curr += forces[ id + dev_size*i + offset ];
        }

        if ( second ) 
        {
            offset = DIMS0*dev_size*count + dev_size*count;
            for ( INT i = 0; i < count; ++i )
            {
                es_curr += forces[ id + dev_size*i + offset ];
            }
        }
    }
    energy_reduction( help + blockIdx.x, e_curr );
    if ( second )
    {
        __syncthreads();
        energy_reduction( help + blockIdx.x + gridDim.x, es_curr );
    }
}

__global__ void add_dummy_beads( INT* kkeys, INT* kvals )
{
    INT tid = threadIdx.x;
    INT last_indx = PH_COUNT * dev_size + d_keys_size;
    kkeys[ last_indx + tid ]  = (BUCKET_SPATIAL_M + 1 + (tid & 0xF) ) << BUCKET_SPATIAL_Z;
    kkeys[ last_indx + tid ] |= (BUCKET_SPATIAL_M + 1 + ((tid & 0xF0)>>4) ) << BUCKET_SPATIAL_Y;
    kkeys[ last_indx + tid ] |= (BUCKET_SPATIAL_M + 1 + ((tid & 0xF00)>>8) ) << BUCKET_SPATIAL_Y;
    kvals[ last_indx + tid ] = PATTERN_PH;
}

template<bool chOverlap, INT cg>
INT spatial_force( DOUBLE* _E )
{
    INT rep = 0;
    DOUBLE* hl = new DOUBLE[sm_red*6];
    INT i;
    INT shMem = cuda_block*sizeof(DOUBLE)*6;
    switch(cuda_block)
    {
        case(1024): comp_h_l<1024><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        case( 512): comp_h_l< 512><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        case( 256): comp_h_l< 256><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        case( 128): comp_h_l< 128><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        case(  64): comp_h_l<  64><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        case(  32): comp_h_l<  32><<<sm_red,cuda_block,shMem>>>( d_coord, d_out ); break;
        default: UNERR("Unimplemented")
    }
    cudaThreadSynchronize(); CCERR
    cudaMemcpy( hl, d_out, sm_red*sizeof(DOUBLE)*6, cudaMemcpyDeviceToHost ); CCERR
    for ( i = 1; i < sm_red; ++i )
    {
        hl[0]=MAX(hl[6*i  ],hl[0]);
        hl[1]=MAX(hl[6*i+1],hl[1]);
        hl[2]=MAX(hl[6*i+2],hl[2]);
        hl[3]=MIN(hl[6*i+3],hl[3]);
        hl[4]=MIN(hl[6*i+4],hl[4]);
        hl[5]=MIN(hl[6*i+5],hl[5]);
    }
    if ( bc == BC_BOX )
    {
        hl[0] = MAXD(   max_force_cutoff + xbox/2, hl[0] );
        hl[1] = MAXD(   max_force_cutoff + ybox/2, hl[1] );
        hl[2] = MAXD(   max_force_cutoff + zbox/2, hl[2] );
        hl[3] = MIND( xbox/2 - 3*max_force_cutoff, hl[3] );
        hl[4] = MIND( ybox/2 - 3*max_force_cutoff, hl[4] );
        hl[5] = MIND( zbox/2 - 3*max_force_cutoff, hl[5] );
    }
    hl[0] = MAXD( (hl[0]-hl[3])/BUCKET_SPATIAL_M, max_force_cutoff );
    hl[1] = MAXD( (hl[1]-hl[4])/BUCKET_SPATIAL_M, max_force_cutoff );
    hl[2] = MAXD( (hl[2]-hl[5])/BUCKET_SPATIAL_M, max_force_cutoff );
    hl[3]=-hl[3];
    hl[4]=-hl[4];
    hl[5]=-hl[5];
    if ( bc == BC_BOX )
    {
        if ( xbox / hl[0] <= 3.01 ||
            ybox / hl[1] <= 3.01 || 
            zbox / hl[2] <= 3.01 ||
            hl[3] > xbox/2 || hl[4] > ybox/2 || hl[5] > zbox/2 )
        {
            if ( hl[3] > xbox/2 || hl[4] > ybox/2 || hl[5] > zbox/2 )
            {
                printf("%g %g %g , %g %g %g\n",xbox/2,ybox/2,zbox/2,hl[3],hl[4],hl[5]);
                UNERR("Wrong correction")
            }
            forceBrute = true;
        }
    }
    cudaMemcpyToSymbol( dev_offset, hl+3, sizeof(DOUBLE)*3 ); CCERR
    cudaMemcpyToSymbol( dev_blen, hl, sizeof(DOUBLE)*3 ); CCERR
    delete[] hl;
    if ( forceBrute ) return 0;

    UINT keys_size = 0;
    cudaMemcpyToSymbol( d_keys_size, &keys_size, sizeof(UINT) ); CCERR
    if( bc == BC_BOX )compute_keys_value_periodic<<<cblocks,cuda_block>>>( d_coord, pkkeys, pkvals );
    else compute_keys_value<<<cblocks,cuda_block>>>( (DOUBLE_4*) d_coord, pkkeys, pkvals );
    cudaThreadSynchronize(); CCERR
    keys_size = 0;
    if ( bc == BC_BOX ) { cudaMemcpyFromSymbol( &keys_size, d_keys_size, sizeof(UINT) ); CCERR }
    sorter->sort( size*PH_COUNT + keys_size );
    cudaThreadSynchronize(); CCERR
    add_dummy_beads<<<1,cuda_block>>>( pkkeys, pkvals );
    cudaThreadSynchronize(); CCERR
    if ( alpha_lj != 0.0 || elec )
    {
        INT indx_blocks = ((PH_COUNT*size+keys_size+COUNT-1)/COUNT + cuda_block - 1)/cuda_block;
        get_indx_spatial<<<indx_blocks,cuda_block>>>( pkkeys, d_indxs );
        cudaThreadSynchronize(); CCERR
        INT shMem = cuda_block * (sizeof(DOUBLE)*5 + sizeof(INT)*2);
        if ( bc == BC_BOX ) cudaMemset( pkpos, 0, size*sizeof(UINT) ); CCERR
        cudaMemset( d_forces, 0, forces_size*sizeof(DOUBLE) );
        INT nblocks = (PH_COUNT*size + keys_size + cuda_block-1)/cuda_block;
        if ( elec && alpha_lj == 0.0 )
        {
#if USE_FLOAT
            if ( bc == BC_BOX ) compute_electro_spatial_indxs<true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, pkpos );
            else compute_electro_spatial_indxs<false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, NULL );
#else
            if ( bc == BC_BOX ) compute_electro_spatial_indxs<true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, pkpos );
            else  compute_electro_spatial_indxs<false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, NULL );
#endif
        }
        else if( !elec )
        {
            if ( cutoff_lj <= 0.0f )
            {
#if USE_FLOAT
                if ( bc == BC_BOX ) compute_LJ_spatial_indxs<true,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>> ( d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_LJ_spatial_indxs<false,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>> ( d_forces, pkkeys, pkvals, d_indxs, NULL );
#else
                if ( bc == BC_BOX ) compute_LJ_spatial_indxs<true,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>> ( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_LJ_spatial_indxs<false,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>> ( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_forces, pkkeys, pkvals, d_indxs, NULL );
#endif
            }
            else
            {
#if USE_FLOAT
                if ( bc == BC_BOX ) compute_LJ_spatial_indxs<true,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_LJ_spatial_indxs<false,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, NULL );
#else
                if ( bc == BC_BOX ) compute_LJ_spatial_indxs<true,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_LJ_spatial_indxs<false,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_forces, pkkeys, pkvals, d_indxs, NULL );
#endif
            }
        }
        else
        {
            INT shMem = cuda_block * (sizeof(DOUBLE)*6 + sizeof(INT)*2);
            if ( cutoff_lj <= 0.0f )
            {
#if USE_FLOAT
                if ( bc == BC_BOX ) compute_electroLJ_spatial_indxs<true,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, pkpos);
                else compute_electroLJ_spatial_indxs<false,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, NULL );
#else
                if ( bc == BC_BOX ) compute_electroLJ_spatial_indxs<true,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_electroLJ_spatial_indxs<false,true,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, pkpos );
#endif
            }
            else
            {
#if USE_FLOAT
                if ( bc == BC_BOX ) compute_electroLJ_spatial_indxs<true,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, pkpos);
                else compute_electroLJ_spatial_indxs<false,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( d_forces, pkkeys, pkvals, d_indxs, NULL );
#else
                if ( bc == BC_BOX ) compute_electroLJ_spatial_indxs<true,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, pkpos );
                else compute_electroLJ_spatial_indxs<false,false,chOverlap,cg><<<nblocks,cuda_block,shMem>>>( (DOUBLE_4*)d_coord, (DOUBLE_2*)d_lj, d_Q, d_forces, pkkeys, pkvals, d_indxs, NULL );
#endif
            }
        }
        cudaThreadSynchronize(); CCERR
                
        if ( alpha_lj != 0.0 && elec )
        {
            INT shMem = cuda_block*sizeof(DOUBLE);
            if ( bc == BC_BOX )
            {
                reduce_forces<MAX_MULT,true><<<cblocks,cuda_block,shMem>>>( d_forces, d_help );
            }
            else
            {
                reduce_forces<PH_COUNT,true><<<cblocks,cuda_block,shMem>>>( d_forces, d_help );
            }
            cudaThreadSynchronize(); CCERR
            cudaMemcpy( h_forces, d_help, 2*cblocks * sizeof (DOUBLE), cudaMemcpyDeviceToHost ); CCERR
            rep = rep || check_E( _E, ENERGY_COULOMB, cblocks );
            rep = rep || check_E( _E,      ENERGY_LJ, cblocks, h_forces + cblocks );
        }
        else
        {
            INT shMem = cuda_block*sizeof(DOUBLE);
            if ( bc == BC_BOX )
            {
                reduce_forces<MAX_MULT,false><<<cblocks,cuda_block,shMem>>>( d_forces, d_help );
            }
            else
            {
                reduce_forces<PH_COUNT,false><<<cblocks,cuda_block,shMem>>>( d_forces, d_help );
            }
            cudaThreadSynchronize(); CCERR
            cudaMemcpy( h_forces, d_help, cblocks * sizeof (DOUBLE), cudaMemcpyDeviceToHost ); CCERR
            rep = rep || check_E( _E, elec ? ENERGY_COULOMB: ENERGY_LJ, cblocks );
        }
    }
    else
    {
        //checking overlap
        cudaMemset( d_overlap, 0, sizeof(INT) ); CCERR
        INT cblocks = (PH_COUNT*size + keys_size + cuda_block - 1)/cuda_block;
        INT shMem = cuda_block*(4*sizeof(DOUBLE)+sizeof(INT));
        if ( bc == BC_BOX ) checking_overlap<true><<<cblocks,cuda_block,shMem>>>( (DOUBLE_4*) d_coord, (DOUBLE_2*)d_lj, pkkeys, pkvals, d_overlap);
        else checking_overlap<false><<<cblocks,cuda_block,shMem>>>( (DOUBLE_4*) d_coord, (DOUBLE_2*)d_lj, pkkeys, pkvals, d_overlap);
        cudaThreadSynchronize(); CCERR

        INT overlap;
        cudaMemcpy( &overlap, d_overlap, sizeof (INT), cudaMemcpyDeviceToHost ); CCERR
        rep = rep || overlap;
        return rep;
    }
    return rep;
}

template<INT cg>
INT spatial_force_overlap( DOUBLE* _E )
{
    if ( check_overlap ) return spatial_force<true,cg>( _E );
    else  return spatial_force<false,cg>( _E );
}

INT spatial_force_g( DOUBLE* _E )
{
    switch(max_gs)
    {
	case(8): return spatial_force_overlap<8>(_E);
	case(7): return spatial_force_overlap<7>(_E);
	case(6): return spatial_force_overlap<6>(_E);
	case(5): return spatial_force_overlap<5>(_E);
	case(4): return spatial_force_overlap<4>(_E);
	case(3): return spatial_force_overlap<3>(_E);
	case(2): return spatial_force_overlap<2>(_E);
	case(1): return spatial_force_overlap<1>(_E);
	case(0): return spatial_force_overlap<0>(_E);
	default: if ( max_gs > 8 ) return spatial_force_overlap<9>(_E);
	else UNERR("Unimplemented")
    }
    return 0; 
}

extern "C"
INT compute_forces( DOUBLE* _E, DOUBLE curr_time )
{
    INT rep, tid;
    if( d_bforces || bc == BC_SPHERE )
    {
        cudaMemcpy( coord, d_coord, sizeof(DOUBLE)*size*DIMS1, cudaMemcpyDeviceToHost ); CCERR
        init_iter_forces( _E );
    }
    else
    {
        if ( _E ) memset( _E, 0, sizeof (DOUBLE) * (ENERGY_LAST + 1) );
    }
    tid = rep = 0;
#if _OPENMP
#pragma omp parallel private(tid) reduction(||:rep)
{
    rep = 0;
    tid = omp_get_thread_num();
#pragma omp master
{
#endif
    forceBrute = false;
    if ( nb_list == BUCKET_SPATIAL )
    {
#if !USE_CUDPP && !USE_THRUST
        UNERR("Compile with CUDPP or Thrust")
#endif
        if ( check_overlap ) rep = rep || spatial_force_g( _E );
        else rep = rep || spatial_force_g( _E );
    }
    if ( nb_list == BUCKET_NONE || forceBrute )
    {
        if ( check_overlap ) rep = rep || brute_force<true>( _E );
        else rep = rep || brute_force<false>( _E );
    }
    if ( d_bforces ) repeat[tid] = rep;
#if _OPENMP
}
#endif
    if( d_bforces ) rep |= bonded_inter_dynamic( _E, tid, curr_time );
#if _OPENMP
}
#endif
    if( d_bforces )
    {
        INT j;
        rep = bonded_inter_collect(_E);
        if ( ! rep )
        {
            cudaMemcpy( d_bforces, F[0], sizeof(DOUBLE)*size*DIMS0, cudaMemcpyHostToDevice ); CCERR
            CUBLAS_AXPYD( size*DIMS0, 1.0f, d_bforces, 1, d_forces, 1 ); cublasCheckError(__FILE__,__LINE__);
            if( _E )
            {
                for ( j = 0; j < ENERGY_LAST; ++j )
                {
                    _E[ j ] += ES[0][j];
                }
            }
        }
    }
    return rep;
}

extern "C"
void init_forces( )
{
    INT fbonded = bond_size || angle_cos_size || angle_size ||
                  dihe_size || dihe_angle_size || E_ext;
    E = (DOUBLE*) malloc( sizeof (DOUBLE) * (ENERGY_LAST + 1) ); CHMEM(E)
    cudaMalloc( (void**) &d_overlap, sizeof (INT) ); CCERR
    forces_size = 0;
    if ( !elec && alpha_lj == 0 && fbonded ) forces_size = csize * DIMS1 * sizeof (DOUBLE);
    else if ( elec || alpha_lj != 0 )
    {
        if( nb_list == BUCKET_SPATIAL )
        {
            INT count = (elec!=0)+(alpha_lj!=0);
            INT mult = (bc==BC_BOX) ? MAX_MULT : PH_COUNT;
            forces_size = mult*csize*sizeof(DOUBLE)*(DIMS0 + count);
        }
        else
        {
            INT count = (elec!=0)+(alpha_lj!=0);
            forces_size = DIMS0*csize + count*cblocks;
        }
    }
    if ( elec || alpha_lj != 0 || fbonded )
    {
        h_forces = (DOUBLE*) malloc( sizeof (DOUBLE) * csize * DIMS1 );
        LOGV(("sizes: %d %d",forces_size,csize*DIMS1));
        cudaMalloc( (void**) &d_forces, forces_size * sizeof (DOUBLE) ); CCERR
        cudaMalloc( (void**) &d_sforces, csize * DIMS1 * sizeof (DOUBLE) ); CCERR
        if ( elec )
        {
            DOUBLE src_kappa = kappa_c;
            DOUBLE src_gamma_eps = gamma_c / epsilon_c;
            DOUBLE src_cutoff = cutoff_c*cutoff_c;
            DOUBLE src_bondQs = bond_c_scale;
            cudaMemcpyToSymbol( dev_kappa, &src_kappa, sizeof(DOUBLE) ); CCERR
            cudaMemcpyToSymbol( dev_gamma_eps, &src_gamma_eps, sizeof(DOUBLE) ); CCERR
            cudaMemcpyToSymbol( dev_cutoffc, &src_cutoff, sizeof(DOUBLE) ); CCERR
            cudaMemcpyToSymbol( dev_bond_c_scale, &src_bondQs, sizeof(DOUBLE) ); CCERR
            DOUBLE* h_Q = (DOUBLE*) malloc( sizeof(DOUBLE)*csize ); CHMEM(h_Q)
            memset( h_Q, 0, sizeof(DOUBLE)*csize );
            for( INT i = 0; i < size; ++ i )
            {
                h_Q[ i ] = Q[ i ];
            }
            cudaMalloc( (void**)&d_Q, sizeof(DOUBLE)*csize ); CCERR
            cudaMemcpy( d_Q,h_Q,sizeof(DOUBLE)*size,cudaMemcpyHostToDevice); CCERR
            free(h_Q);
        }
        if ( alpha_lj != 0.0 )
        {
            DOUBLE src_cutoff = cutoff_lj*cutoff_lj;
            DOUBLE src_alpha_lj = alpha_lj;
            DOUBLE src_bondLJs = bond_lj_scale;
            cudaMemcpyToSymbol( dev_cutoffLJ, &src_cutoff, sizeof(DOUBLE) ); CCERR
            cudaMemcpyToSymbol( dev_alpha_lj, &src_alpha_lj, sizeof(DOUBLE) ); CCERR
            DOUBLE src_6term = (DOUBLE)lj_6_term;
            cudaMemcpyToSymbol( dev_lj_6_term, &src_6term, sizeof(DOUBLE) ); CCERR
            cudaMemcpyToSymbol( dev_bond_lj_scale, &src_bondLJs, sizeof(DOUBLE) ); CCERR
        }
    }
    /* for all we check overlap*/
    DOUBLE* lj_data = (DOUBLE*) malloc( csize*sizeof(DOUBLE)*2 ); CHMEM(lj_data)
    memset( lj_data, 0, csize*sizeof(DOUBLE)*2 );
    for ( INT i = 0; i < 2 * size; ++i )
    {
        lj_data[i] = LJ[i];
    }
    cudaMalloc( (void**) &d_lj, sizeof(DOUBLE)*csize*2 ); CCERR
    cudaMemcpy( d_lj, lj_data, sizeof(DOUBLE)*csize*2, cudaMemcpyHostToDevice ); CCERR
    free( lj_data );

    if ( bc == BC_BOX )
    {
        DOUBLE bx[3];
        DOUBLE iL[3];
        bx[0] = box[0], bx[1] = box[1], bx[2] = box[2];
        iL[0] = inv_L[0], iL[1] = inv_L[1], iL[2] = inv_L[2];
        cudaMemcpyToSymbol( dev_box0, bx, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_box1, bx + 1, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_box2, bx + 2, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL0, iL, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL1, iL + 1, sizeof (DOUBLE) ); CCERR
        cudaMemcpyToSymbol( dev_invL2, iL + 2, sizeof (DOUBLE) ); CCERR
    }
    max_gs = 0;

    if( bond_size > 0 && (( elec && bond_c_scale != 1 ) || ( alpha_lj != 0.0 && bond_lj_scale != 1 )) )
    {
        graph = get_connection_graph( &max_gs, csize );
        LOGV(("max connections: %d", max_gs));
        cudaMemcpyToSymbol( dev_bonds, &max_gs, sizeof (INT) ); CCERR
        cudaMalloc( (void**) &d_gconns, sizeof(INT)*max_gs*csize ); CCERR
        cudaMemcpy( d_gconns, graph, sizeof(INT)*max_gs*size, cudaMemcpyHostToDevice ); CCERR
        cudaMemcpyToSymbol( dev_gconns, &d_gconns, sizeof(INT*) ); CCERR
    }
    init_cutoff();
    if( nb_list == BUCKET_SPATIAL )
    {
#if !USE_CUDPP && !USE_THRUST
        UNERR( "Only brute without THRUST or CUDPP" )
#endif
        INT mult = ( bc == BC_BOX ) ? MAX_MULT : PH_COUNT;
        INT alloc_size = mult*size + cuda_block;
        cudaMalloc( (void**) &pkkeys, sizeof(INT)*alloc_size ); CCERR
        cudaMalloc( (void**) &pkvals, sizeof(INT)*alloc_size ); CCERR
        if ( alpha_lj != 0 || elec )
        {
            cudaMalloc( (void**) &d_help, sizeof(DOUBLE)*2*cblocks ); CCERR
            cudaMalloc( (void**) &pkpos, sizeof(UINT)*size ); CCERR
        }
        cudaThreadSynchronize(); CCERR
        sorter = new StaticSorter( pkkeys, pkvals, mult*size );
                
        INT indxs_size = (mult*size+COUNT)/COUNT;
        indxs_size = ((indxs_size + cuda_block - 1)/cuda_block)*cuda_block;
        cudaMalloc( (void**) &d_indxs, sizeof(INT)*2*indxs_size ); CCERR

#if USE_FLOAT
        //bind textures
        tr_coords.addressMode[0] = cudaAddressModeClamp;
        tr_coords.addressMode[1] = cudaAddressModeClamp;
        tr_coords.addressMode[2] = cudaAddressModeClamp;
        tr_coords.filterMode = cudaFilterModePoint;
        tr_coords.normalized = false;
        cudaChannelFormatDesc channel1 = cudaCreateChannelDesc<DOUBLE_4>(); CCERR
        cudaBindTexture( 0, tr_coords, d_coord, channel1, size*DIMS1*sizeof(DOUBLE) ); CCERR
        if ( d_Q )
        {
            tr_q.addressMode[0] = cudaAddressModeClamp;
            tr_q.addressMode[1] = cudaAddressModeClamp;
            tr_q.addressMode[2] = cudaAddressModeClamp;
            tr_q.filterMode = cudaFilterModePoint;
            tr_q.normalized = false;
            cudaChannelFormatDesc channel2 = cudaCreateChannelDesc<DOUBLE>(); CCERR
            cudaBindTexture( 0, tr_q, d_Q, channel2, size*sizeof(DOUBLE) ); CCERR
        }

        if ( d_lj )
        {
            tr_lj.addressMode[0] = cudaAddressModeClamp;
            tr_lj.addressMode[1] = cudaAddressModeClamp;
            tr_lj.addressMode[2] = cudaAddressModeClamp;
            tr_lj.filterMode = cudaFilterModePoint;
            tr_lj.normalized = false;
            cudaChannelFormatDesc channel3 = cudaCreateChannelDesc<DOUBLE_2>(); CCERR
            cudaBindTexture( 0, tr_lj, d_lj, channel3, size*sizeof(DOUBLE_2) ); CCERR
        }
#endif
        cudaMalloc( (void**) &d_out, sm_red*sizeof(DOUBLE)*6 ); CCERR
        cudaMemcpyToSymbol( dev_max_cutoff, &max_force_cutoff, sizeof(DOUBLE) );
    }
    else if ( nb_list != BUCKET_NONE )
    {
        UNERR("Only nb_list=brute and nb_list=spatial are supported.")
    }
    else
    {
#if USE_FLOAT
        cudaMalloc( (void**) &d_ljcoord, csize * DIMS1 * sizeof (DOUBLE) ); CCERR
        tr_coords.addressMode[0] = cudaAddressModeClamp;
        tr_coords.addressMode[1] = cudaAddressModeClamp;
        tr_coords.addressMode[2] = cudaAddressModeClamp;
        tr_coords.filterMode = cudaFilterModePoint;
        tr_coords.normalized = false;
        cudaChannelFormatDesc channel1 = cudaCreateChannelDesc<DOUBLE_4>(); CCERR
        cudaBindTexture( 0, tr_coords, d_ljcoord, channel1, csize*DIMS1*sizeof(DOUBLE) ); CCERR
#endif
    }
    if( fbonded )
    {
        free(E);
        init_Fs();
        cudaMalloc( (void**) &d_bforces, sizeof(DOUBLE)*size*DIMS0 ); CCERR
    }
    INT h_size = size;
    cudaMemcpyToSymbol( dev_size, &h_size, sizeof(INT) ); CCERR
}

extern "C"
void free_forces( )
{
    free( E );
    if ( h_forces ) free( h_forces );
    if ( d_forces ) { cudaFree( d_forces ); CCERR }
    if ( d_ljcoord ) { cudaFree( d_ljcoord ); CCERR }
    free( graph );
    if( d_gconns ) { cudaFree( d_gconns ); CCERR }
    if( pkkeys ) { cudaFree( pkkeys ); CCERR }
    if( pkvals ) { cudaFree( pkvals ); CCERR }
    cudaFree( d_overlap ); CCERR
    if( d_bforces )
    {
        free_Fs();
        cudaFree( d_bforces ); CCERR
    }
    if ( d_indxs ) { cudaFree( d_indxs ); CCERR }
#if USE_CUDPP  || USE_THRUST
    if ( sorter ) delete sorter;
#endif
}
